# -*- coding: utf-8 -*-
#
# BigDFT-suite documentation build configuration file, created by
# sphinx-quickstart on Sun Sep 30 14:58:47 2018.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sourcepath = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, sourcepath)
suitepath = os.path.abspath(os.path.join(sourcepath, '..'))
pythonfutilepath = os.path.join(sourcepath, 'futile', 'python')
sys.path.insert(0, pythonfutilepath)
sys.path.insert(0, suitepath)
print (sys.path)
# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
#needs_sphinx = '1.6.7'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
			  'sphinx.ext.doctest',
			  'sphinx.ext.intersphinx',
			  'sphinx.ext.todo',
			  'sphinx.ext.mathjax',
			  'sphinx.ext.ifconfig',
			  'sphinx.ext.viewcode',
			  'sphinx.ext.githubpages',
			  'sphinxcontrib.programoutput',
			  'sphinxfortran.fortran_domain',
			  'nbsphinx']

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['sphinxdoctemplates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'BigDFT-suite'
copyright = u'2018-%s, BigDFT developers'
author = u'BigDFT developers'

absolute_intersphinx_path = os.environ.get(
	'ABSOLUTE_INTERSPHINX_PATH') == 'True'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = u'1.9'
# The full version, including alpha/beta/rc tags.
release = u'1.9.2'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['sphinxdocbuild', 'cmake*', 'ipython', 'Thumbs.db',
					'.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'
html_logo = os.path.join('static', 'logo-bigdft-white.svg')
html_favicon = os.path.join('static', 'bigdft-logo-reduced.svg')

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['static']

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = [
	'custom.css',
]

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'BigDFT-suitedoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
	# The paper size ('letterpaper' or 'a4paper').
	#
	# 'papersize': 'letterpaper',

	# The font size ('10pt', '11pt' or '12pt').
	#
	# 'pointsize': '10pt',

	# Additional stuff for the LaTeX preamble.
	#
	# 'preamble': '',

	# Latex figure (float) alignment
	#
	# 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
	(master_doc, 'BigDFT-suite.tex', u'BigDFT-suite Documentation',
	 u'BigDFT developers', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
	(master_doc, 'bigdft-suite', u'BigDFT-suite Documentation',
	 [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
	(master_doc, 'BigDFT-suite', u'BigDFT-suite Documentation',
	 author, 'BigDFT-suite', 'A DFT suite based on Wavelets.',
	 'Miscellaneous'),
]

# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']

intersphinx_mapping = {'https://docs.python.org/3/': None}

builddir = None


def get_project_tuple(package):
	projectdir = os.path.abspath(os.path.join(suitepath, package, 'build',
											  'html'))
	if not absolute_intersphinx_path:
		projectdir = os.path.relpath(projectdir, os.getcwd())
	inventory_path = os.path.join(projectdir, 'objects.inv')
	return projectdir, inventory_path

on_rtd = os.environ.get('READTHEDOCS') == 'True'

if on_rtd:
	intersphinx_mapping['futile'] = ('https://bigdft-suite.readthedocs.io/projects/futile/en/latest/', None)
	intersphinx_mapping['pybigdft'] = ('https://bigdft-suite.readthedocs.io/projects/PyBigDFT/en/latest/', None)
else:
	#build help pages from template
	# try to run the build for the various subprojects
	# apidoc command
	# sphinx-apidoc -o ./PyBigDFT/source ./PyBigDFT -fe
	import sphinxprebuild as spb
	spb.project_builder('futile', builddir, 'doc', '..').build()
	spb.project_builder('PyBigDFT', builddir, sourcepath='..', extraargs='-a -E').build()
	print (get_project_tuple('futile')) #spb.project_tuple('futile'), builddir)
	intersphinx_mapping['futile'] = get_project_tuple('futile')
	print (get_project_tuple('PyBigDFT'))
	intersphinx_mapping['pybigdft'] = get_project_tuple('PyBigDFT')
	print(sys.path)
