
module mod_R_1_norm

  implicit none

contains

!> compute the 1 norm of a combination of cubic harmonics, whose decomposition coefficients are given in the R basis
!!  
recursive function R_1_norm(a,c,lmax)
 
  implicit none
 
  ! input
  real(kind=8), intent(in):: a               !< exponent of the cubic harmonic 
  integer     , intent(in):: lmax            !< maximum angular momentum of the deconposition
  real(kind=8), intent(in),dimension(*):: c  !< coefficients of the decomposition
  
  ! return value
  real(kind=8):: R_1_norm
  
  ! local variables
  real(kind=8), parameter :: PI=3.1415926535897932384626433832795d0
  
  ! selection on lmax
  R_1_norm=0.0d0
  select case(lmax)
    case (0)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
    case (1)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
    case (2)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
    case (3)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
    case (4)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
    case (5)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
    case (6)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
    case (7)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
    case (8)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(121)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(124)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(126)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(131)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(133)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(135)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(142)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(144)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(146)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(148)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(157)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(159)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(161)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(163)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(165)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
    case (9)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(121)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(124)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(126)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(131)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(133)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(135)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(142)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(144)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(146)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(148)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(157)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(159)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(161)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(163)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(165)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
    case (10)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(121)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(124)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(126)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(131)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(133)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(135)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(142)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(144)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(146)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(148)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(157)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(159)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(161)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(163)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(165)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(221)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(224)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(226)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(231)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(233)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(235)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(242)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(244)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(246)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(248)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(257)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(259)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(261)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(263)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(265)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(276)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(278)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(280)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(282)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(284)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(286)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
    case (11)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(121)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(124)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(126)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(131)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(133)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(135)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(142)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(144)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(146)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(148)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(157)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(159)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(161)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(163)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(165)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(221)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(224)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(226)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(231)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(233)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(235)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(242)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(244)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(246)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(248)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(257)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(259)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(261)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(263)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(265)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(276)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(278)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(280)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(282)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(284)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(286)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
    case (12)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(121)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(124)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(126)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(131)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(133)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(135)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(142)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(144)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(146)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(148)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(157)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(159)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(161)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(163)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(165)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(221)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(224)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(226)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(231)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(233)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(235)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(242)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(244)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(246)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(248)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(257)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(259)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(261)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(263)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(265)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(276)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(278)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(280)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(282)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(284)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(286)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(365)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(368)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(370)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(375)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(377)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(379)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(386)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(388)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(390)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(392)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(401)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(403)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(405)*27.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(407)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(409)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(420)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(422)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(424)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(426)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(428)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(430)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(443)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(445)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(447)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(449)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(451)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(453)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(455)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
    case (13)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(121)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(124)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(126)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(131)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(133)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(135)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(142)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(144)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(146)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(148)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(157)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(159)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(161)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(163)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(165)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(221)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(224)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(226)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(231)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(233)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(235)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(242)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(244)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(246)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(248)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(257)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(259)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(261)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(263)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(265)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(276)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(278)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(280)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(282)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(284)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(286)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(365)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(368)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(370)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(375)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(377)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(379)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(386)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(388)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(390)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(392)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(401)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(403)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(405)*27.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(407)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(409)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(420)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(422)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(424)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(426)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(428)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(430)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(443)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(445)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(447)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(449)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(451)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(453)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(455)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
    case (14)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(121)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(124)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(126)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(131)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(133)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(135)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(142)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(144)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(146)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(148)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(157)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(159)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(161)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(163)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(165)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(221)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(224)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(226)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(231)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(233)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(235)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(242)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(244)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(246)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(248)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(257)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(259)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(261)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(263)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(265)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(276)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(278)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(280)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(282)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(284)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(286)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(365)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(368)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(370)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(375)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(377)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(379)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(386)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(388)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(390)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(392)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(401)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(403)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(405)*27.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(407)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(409)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(420)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(422)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(424)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(426)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(428)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(430)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(443)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(445)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(447)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(449)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(451)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(453)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(455)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(561)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(564)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(566)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(571)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(573)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(575)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(582)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(584)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(586)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(588)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(597)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(599)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(601)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(603)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(605)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(616)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(618)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(620)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(622)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(624)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(626)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(639)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(641)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(643)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(645)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(647)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(649)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(651)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(666)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(668)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(670)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(672)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(674)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(676)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(678)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(680)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
    case (15)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(121)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(124)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(126)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(131)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(133)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(135)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(142)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(144)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(146)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(148)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(157)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(159)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(161)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(163)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(165)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(221)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(224)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(226)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(231)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(233)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(235)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(242)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(244)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(246)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(248)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(257)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(259)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(261)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(263)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(265)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(276)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(278)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(280)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(282)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(284)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(286)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(365)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(368)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(370)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(375)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(377)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(379)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(386)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(388)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(390)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(392)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(401)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(403)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(405)*27.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(407)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(409)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(420)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(422)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(424)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(426)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(428)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(430)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(443)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(445)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(447)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(449)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(451)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(453)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(455)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(561)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(564)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(566)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(571)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(573)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(575)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(582)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(584)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(586)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(588)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(597)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(599)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(601)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(603)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(605)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(616)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(618)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(620)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(622)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(624)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(626)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(639)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(641)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(643)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(645)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(647)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(649)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(651)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(666)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(668)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(670)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(672)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(674)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(676)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(678)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(680)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
    case (16)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(121)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(124)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(126)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(131)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(133)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(135)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(142)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(144)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(146)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(148)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(157)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(159)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(161)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(163)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(165)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(221)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(224)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(226)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(231)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(233)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(235)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(242)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(244)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(246)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(248)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(257)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(259)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(261)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(263)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(265)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(276)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(278)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(280)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(282)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(284)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(286)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(365)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(368)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(370)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(375)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(377)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(379)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(386)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(388)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(390)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(392)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(401)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(403)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(405)*27.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(407)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(409)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(420)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(422)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(424)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(426)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(428)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(430)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(443)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(445)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(447)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(449)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(451)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(453)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(455)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(561)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(564)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(566)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(571)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(573)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(575)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(582)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(584)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(586)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(588)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(597)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(599)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(601)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(603)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(605)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(616)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(618)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(620)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(622)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(624)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(626)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(639)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(641)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(643)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(645)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(647)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(649)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(651)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(666)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(668)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(670)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(672)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(674)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(676)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(678)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(680)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(817)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(820)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(822)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(827)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(829)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(831)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(838)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(840)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(842)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(844)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(853)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(855)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(857)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(859)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(861)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(872)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(874)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(876)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(878)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(880)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(882)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(895)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(897)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(899)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(901)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(903)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(905)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(907)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(922)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(924)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(926)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(928)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(930)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(932)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(934)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(936)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(953)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(955)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(957)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(959)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(961)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(963)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(965)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(967)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(969)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
    case (17)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(121)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(124)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(126)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(131)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(133)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(135)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(142)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(144)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(146)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(148)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(157)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(159)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(161)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(163)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(165)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(221)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(224)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(226)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(231)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(233)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(235)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(242)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(244)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(246)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(248)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(257)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(259)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(261)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(263)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(265)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(276)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(278)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(280)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(282)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(284)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(286)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(365)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(368)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(370)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(375)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(377)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(379)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(386)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(388)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(390)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(392)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(401)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(403)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(405)*27.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(407)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(409)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(420)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(422)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(424)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(426)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(428)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(430)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(443)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(445)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(447)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(449)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(451)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(453)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(455)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(561)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(564)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(566)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(571)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(573)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(575)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(582)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(584)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(586)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(588)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(597)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(599)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(601)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(603)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(605)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(616)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(618)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(620)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(622)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(624)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(626)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(639)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(641)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(643)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(645)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(647)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(649)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(651)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(666)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(668)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(670)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(672)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(674)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(676)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(678)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(680)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(817)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(820)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(822)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(827)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(829)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(831)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(838)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(840)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(842)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(844)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(853)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(855)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(857)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(859)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(861)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(872)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(874)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(876)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(878)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(880)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(882)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(895)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(897)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(899)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(901)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(903)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(905)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(907)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(922)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(924)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(926)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(928)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(930)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(932)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(934)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(936)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(953)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(955)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(957)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(959)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(961)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(963)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(965)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(967)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(969)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
    case (18)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(121)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(124)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(126)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(131)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(133)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(135)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(142)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(144)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(146)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(148)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(157)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(159)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(161)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(163)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(165)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(221)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(224)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(226)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(231)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(233)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(235)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(242)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(244)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(246)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(248)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(257)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(259)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(261)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(263)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(265)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(276)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(278)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(280)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(282)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(284)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(286)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(365)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(368)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(370)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(375)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(377)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(379)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(386)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(388)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(390)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(392)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(401)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(403)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(405)*27.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(407)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(409)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(420)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(422)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(424)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(426)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(428)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(430)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(443)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(445)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(447)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(449)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(451)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(453)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(455)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(561)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(564)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(566)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(571)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(573)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(575)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(582)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(584)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(586)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(588)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(597)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(599)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(601)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(603)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(605)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(616)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(618)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(620)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(622)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(624)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(626)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(639)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(641)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(643)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(645)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(647)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(649)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(651)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(666)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(668)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(670)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(672)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(674)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(676)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(678)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(680)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(817)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(820)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(822)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(827)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(829)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(831)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(838)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(840)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(842)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(844)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(853)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(855)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(857)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(859)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(861)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(872)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(874)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(876)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(878)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(880)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(882)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(895)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(897)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(899)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(901)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(903)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(905)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(907)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(922)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(924)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(926)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(928)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(930)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(932)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(934)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(936)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(953)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(955)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(957)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(959)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(961)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(963)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(965)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(967)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(969)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(1141)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1144)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1146)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1151)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1153)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1155)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1162)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1164)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1166)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1168)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1177)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1179)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1181)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1183)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1185)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1196)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1198)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1200)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1202)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1204)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1206)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1219)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1221)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1223)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1225)*3375.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1227)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1229)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1231)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1246)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1248)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1250)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1252)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1254)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1256)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1258)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1260)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1277)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1279)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1281)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1283)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1285)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1287)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1289)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1291)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1293)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1312)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1314)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1316)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1318)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1320)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1322)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1324)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1326)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1328)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1330)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
    case (19)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(121)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(124)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(126)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(131)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(133)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(135)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(142)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(144)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(146)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(148)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(157)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(159)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(161)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(163)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(165)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(221)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(224)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(226)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(231)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(233)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(235)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(242)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(244)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(246)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(248)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(257)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(259)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(261)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(263)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(265)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(276)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(278)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(280)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(282)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(284)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(286)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(365)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(368)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(370)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(375)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(377)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(379)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(386)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(388)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(390)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(392)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(401)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(403)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(405)*27.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(407)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(409)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(420)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(422)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(424)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(426)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(428)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(430)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(443)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(445)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(447)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(449)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(451)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(453)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(455)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(561)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(564)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(566)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(571)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(573)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(575)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(582)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(584)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(586)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(588)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(597)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(599)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(601)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(603)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(605)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(616)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(618)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(620)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(622)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(624)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(626)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(639)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(641)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(643)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(645)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(647)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(649)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(651)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(666)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(668)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(670)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(672)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(674)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(676)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(678)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(680)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(817)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(820)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(822)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(827)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(829)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(831)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(838)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(840)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(842)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(844)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(853)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(855)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(857)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(859)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(861)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(872)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(874)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(876)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(878)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(880)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(882)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(895)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(897)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(899)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(901)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(903)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(905)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(907)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(922)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(924)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(926)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(928)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(930)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(932)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(934)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(936)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(953)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(955)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(957)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(959)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(961)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(963)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(965)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(967)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(969)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(1141)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1144)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1146)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1151)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1153)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1155)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1162)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1164)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1166)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1168)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1177)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1179)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1181)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1183)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1185)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1196)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1198)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1200)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1202)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1204)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1206)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1219)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1221)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1223)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1225)*3375.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1227)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1229)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1231)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1246)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1248)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1250)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1252)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1254)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1256)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1258)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1260)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1277)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1279)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1281)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1283)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1285)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1287)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1289)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1291)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1293)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1312)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1314)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1316)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1318)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1320)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1322)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1324)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1326)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1328)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1330)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
    case (20)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(121)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(124)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(126)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(131)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(133)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(135)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(142)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(144)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(146)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(148)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(157)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(159)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(161)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(163)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(165)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(221)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(224)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(226)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(231)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(233)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(235)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(242)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(244)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(246)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(248)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(257)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(259)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(261)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(263)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(265)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(276)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(278)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(280)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(282)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(284)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(286)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(365)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(368)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(370)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(375)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(377)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(379)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(386)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(388)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(390)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(392)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(401)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(403)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(405)*27.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(407)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(409)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(420)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(422)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(424)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(426)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(428)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(430)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(443)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(445)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(447)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(449)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(451)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(453)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(455)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(561)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(564)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(566)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(571)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(573)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(575)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(582)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(584)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(586)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(588)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(597)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(599)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(601)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(603)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(605)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(616)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(618)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(620)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(622)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(624)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(626)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(639)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(641)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(643)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(645)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(647)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(649)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(651)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(666)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(668)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(670)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(672)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(674)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(676)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(678)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(680)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(817)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(820)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(822)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(827)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(829)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(831)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(838)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(840)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(842)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(844)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(853)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(855)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(857)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(859)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(861)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(872)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(874)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(876)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(878)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(880)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(882)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(895)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(897)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(899)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(901)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(903)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(905)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(907)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(922)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(924)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(926)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(928)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(930)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(932)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(934)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(936)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(953)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(955)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(957)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(959)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(961)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(963)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(965)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(967)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(969)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(1141)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1144)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1146)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1151)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1153)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1155)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1162)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1164)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1166)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1168)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1177)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1179)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1181)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1183)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1185)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1196)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1198)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1200)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1202)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1204)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1206)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1219)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1221)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1223)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1225)*3375.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1227)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1229)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1231)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1246)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1248)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1250)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1252)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1254)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1256)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1258)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1260)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1277)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1279)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1281)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1283)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1285)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1287)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1289)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1291)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1293)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1312)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1314)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1316)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1318)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1320)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1322)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1324)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1326)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1328)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1330)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1541)*654729075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1544)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1546)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1551)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1553)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1555)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1562)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1564)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1566)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1568)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1577)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1579)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1581)*93555.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1583)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1585)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1596)*893025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1598)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1600)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1602)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1604)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1606)*893025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1619)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1621)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1623)*33075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1625)*23625.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1627)*33075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1629)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1631)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1646)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1648)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1650)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1652)*23625.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1654)*23625.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1656)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1658)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1660)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1677)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1679)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1681)*93555.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1683)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1685)*33075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1687)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1689)*93555.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1691)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1693)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1712)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1714)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1716)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1718)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1720)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1722)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1724)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1726)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1728)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1730)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1751)*654729075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1753)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1755)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1757)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1759)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1761)*893025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1763)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1765)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1767)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1769)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1771)*654729075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
    case (21)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(121)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(124)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(126)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(131)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(133)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(135)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(142)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(144)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(146)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(148)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(157)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(159)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(161)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(163)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(165)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(221)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(224)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(226)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(231)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(233)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(235)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(242)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(244)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(246)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(248)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(257)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(259)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(261)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(263)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(265)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(276)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(278)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(280)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(282)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(284)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(286)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(365)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(368)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(370)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(375)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(377)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(379)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(386)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(388)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(390)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(392)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(401)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(403)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(405)*27.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(407)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(409)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(420)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(422)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(424)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(426)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(428)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(430)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(443)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(445)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(447)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(449)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(451)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(453)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(455)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(561)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(564)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(566)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(571)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(573)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(575)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(582)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(584)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(586)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(588)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(597)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(599)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(601)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(603)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(605)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(616)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(618)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(620)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(622)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(624)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(626)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(639)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(641)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(643)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(645)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(647)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(649)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(651)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(666)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(668)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(670)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(672)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(674)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(676)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(678)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(680)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(817)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(820)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(822)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(827)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(829)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(831)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(838)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(840)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(842)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(844)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(853)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(855)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(857)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(859)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(861)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(872)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(874)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(876)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(878)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(880)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(882)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(895)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(897)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(899)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(901)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(903)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(905)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(907)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(922)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(924)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(926)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(928)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(930)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(932)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(934)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(936)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(953)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(955)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(957)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(959)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(961)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(963)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(965)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(967)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(969)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(1141)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1144)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1146)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1151)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1153)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1155)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1162)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1164)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1166)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1168)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1177)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1179)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1181)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1183)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1185)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1196)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1198)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1200)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1202)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1204)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1206)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1219)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1221)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1223)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1225)*3375.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1227)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1229)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1231)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1246)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1248)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1250)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1252)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1254)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1256)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1258)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1260)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1277)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1279)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1281)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1283)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1285)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1287)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1289)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1291)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1293)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1312)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1314)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1316)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1318)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1320)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1322)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1324)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1326)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1328)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1330)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1541)*654729075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1544)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1546)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1551)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1553)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1555)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1562)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1564)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1566)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1568)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1577)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1579)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1581)*93555.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1583)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1585)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1596)*893025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1598)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1600)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1602)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1604)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1606)*893025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1619)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1621)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1623)*33075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1625)*23625.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1627)*33075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1629)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1631)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1646)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1648)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1650)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1652)*23625.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1654)*23625.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1656)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1658)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1660)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1677)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1679)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1681)*93555.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1683)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1685)*33075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1687)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1689)*93555.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1691)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1693)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1712)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1714)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1716)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1718)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1720)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1722)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1724)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1726)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1728)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1730)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1751)*654729075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1753)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1755)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1757)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1759)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1761)*893025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1763)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1765)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1767)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1769)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1771)*654729075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
    case (22)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(121)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(124)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(126)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(131)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(133)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(135)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(142)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(144)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(146)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(148)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(157)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(159)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(161)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(163)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(165)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(221)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(224)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(226)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(231)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(233)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(235)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(242)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(244)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(246)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(248)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(257)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(259)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(261)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(263)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(265)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(276)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(278)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(280)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(282)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(284)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(286)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(365)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(368)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(370)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(375)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(377)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(379)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(386)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(388)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(390)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(392)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(401)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(403)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(405)*27.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(407)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(409)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(420)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(422)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(424)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(426)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(428)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(430)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(443)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(445)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(447)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(449)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(451)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(453)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(455)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(561)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(564)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(566)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(571)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(573)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(575)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(582)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(584)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(586)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(588)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(597)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(599)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(601)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(603)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(605)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(616)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(618)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(620)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(622)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(624)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(626)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(639)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(641)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(643)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(645)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(647)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(649)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(651)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(666)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(668)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(670)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(672)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(674)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(676)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(678)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(680)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(817)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(820)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(822)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(827)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(829)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(831)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(838)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(840)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(842)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(844)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(853)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(855)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(857)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(859)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(861)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(872)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(874)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(876)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(878)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(880)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(882)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(895)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(897)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(899)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(901)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(903)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(905)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(907)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(922)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(924)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(926)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(928)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(930)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(932)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(934)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(936)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(953)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(955)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(957)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(959)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(961)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(963)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(965)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(967)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(969)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(1141)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1144)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1146)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1151)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1153)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1155)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1162)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1164)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1166)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1168)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1177)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1179)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1181)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1183)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1185)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1196)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1198)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1200)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1202)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1204)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1206)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1219)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1221)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1223)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1225)*3375.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1227)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1229)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1231)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1246)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1248)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1250)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1252)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1254)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1256)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1258)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1260)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1277)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1279)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1281)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1283)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1285)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1287)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1289)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1291)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1293)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1312)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1314)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1316)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1318)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1320)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1322)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1324)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1326)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1328)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1330)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1541)*654729075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1544)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1546)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1551)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1553)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1555)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1562)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1564)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1566)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1568)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1577)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1579)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1581)*93555.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1583)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1585)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1596)*893025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1598)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1600)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1602)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1604)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1606)*893025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1619)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1621)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1623)*33075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1625)*23625.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1627)*33075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1629)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1631)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1646)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1648)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1650)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1652)*23625.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1654)*23625.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1656)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1658)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1660)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1677)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1679)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1681)*93555.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1683)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1685)*33075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1687)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1689)*93555.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1691)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1693)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1712)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1714)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1716)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1718)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1720)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1722)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1724)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1726)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1728)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1730)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1751)*654729075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1753)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1755)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1757)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1759)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1761)*893025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1763)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1765)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1767)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1769)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1771)*654729075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(2025)*13749310575.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2028)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2030)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2035)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2037)*34459425.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2039)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2046)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2048)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2050)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2052)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2061)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2063)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2065)*1216215.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2067)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2069)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2080)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2082)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2084)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2086)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2088)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2090)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2103)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2105)*893025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2107)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2109)*212625.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2111)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2113)*893025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2115)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2130)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2132)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2134)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2136)*165375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2138)*165375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2140)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2142)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2144)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2161)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2163)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2165)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2167)*212625.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2169)*165375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2171)*212625.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2173)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2175)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2177)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2196)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2198)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2200)*1216215.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2202)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2204)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2206)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2208)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2210)*1216215.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2212)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2214)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2235)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2237)*34459425.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2239)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2241)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2243)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2245)*893025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2247)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2249)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2251)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2253)*34459425.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2255)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2278)*13749310575.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2280)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2282)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2284)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2286)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2288)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2290)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2292)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2294)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2296)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2298)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2300)*13749310575.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
    case (23)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(121)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(124)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(126)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(131)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(133)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(135)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(142)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(144)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(146)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(148)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(157)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(159)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(161)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(163)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(165)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(221)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(224)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(226)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(231)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(233)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(235)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(242)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(244)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(246)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(248)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(257)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(259)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(261)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(263)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(265)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(276)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(278)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(280)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(282)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(284)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(286)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(365)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(368)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(370)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(375)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(377)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(379)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(386)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(388)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(390)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(392)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(401)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(403)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(405)*27.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(407)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(409)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(420)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(422)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(424)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(426)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(428)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(430)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(443)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(445)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(447)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(449)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(451)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(453)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(455)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(561)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(564)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(566)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(571)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(573)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(575)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(582)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(584)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(586)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(588)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(597)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(599)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(601)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(603)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(605)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(616)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(618)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(620)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(622)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(624)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(626)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(639)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(641)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(643)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(645)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(647)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(649)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(651)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(666)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(668)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(670)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(672)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(674)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(676)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(678)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(680)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(817)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(820)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(822)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(827)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(829)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(831)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(838)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(840)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(842)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(844)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(853)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(855)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(857)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(859)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(861)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(872)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(874)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(876)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(878)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(880)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(882)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(895)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(897)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(899)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(901)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(903)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(905)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(907)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(922)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(924)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(926)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(928)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(930)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(932)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(934)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(936)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(953)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(955)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(957)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(959)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(961)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(963)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(965)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(967)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(969)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(1141)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1144)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1146)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1151)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1153)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1155)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1162)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1164)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1166)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1168)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1177)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1179)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1181)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1183)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1185)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1196)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1198)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1200)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1202)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1204)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1206)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1219)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1221)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1223)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1225)*3375.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1227)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1229)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1231)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1246)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1248)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1250)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1252)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1254)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1256)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1258)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1260)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1277)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1279)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1281)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1283)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1285)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1287)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1289)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1291)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1293)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1312)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1314)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1316)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1318)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1320)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1322)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1324)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1326)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1328)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1330)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1541)*654729075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1544)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1546)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1551)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1553)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1555)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1562)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1564)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1566)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1568)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1577)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1579)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1581)*93555.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1583)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1585)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1596)*893025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1598)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1600)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1602)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1604)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1606)*893025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1619)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1621)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1623)*33075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1625)*23625.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1627)*33075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1629)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1631)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1646)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1648)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1650)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1652)*23625.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1654)*23625.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1656)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1658)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1660)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1677)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1679)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1681)*93555.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1683)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1685)*33075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1687)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1689)*93555.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1691)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1693)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1712)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1714)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1716)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1718)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1720)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1722)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1724)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1726)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1728)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1730)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1751)*654729075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1753)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1755)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1757)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1759)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1761)*893025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1763)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1765)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1767)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1769)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1771)*654729075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(2025)*13749310575.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2028)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2030)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2035)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2037)*34459425.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2039)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2046)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2048)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2050)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2052)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2061)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2063)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2065)*1216215.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2067)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2069)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2080)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2082)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2084)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2086)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2088)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2090)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2103)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2105)*893025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2107)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2109)*212625.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2111)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2113)*893025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2115)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2130)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2132)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2134)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2136)*165375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2138)*165375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2140)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2142)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2144)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2161)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2163)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2165)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2167)*212625.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2169)*165375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2171)*212625.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2173)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2175)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2177)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2196)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2198)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2200)*1216215.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2202)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2204)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2206)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2208)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2210)*1216215.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2212)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2214)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2235)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2237)*34459425.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2239)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2241)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2243)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2245)*893025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2247)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2249)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2251)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2253)*34459425.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2255)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2278)*13749310575.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2280)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2282)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2284)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2286)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2288)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2290)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2292)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2294)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2296)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2298)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2300)*13749310575.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
    case (24)
      R_1_norm=R_1_norm+c(1)*PI**1*sqrt(PI)/(a*sqrt(a))
      R_1_norm=R_1_norm+c(5)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(8)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(10)*1.0d0/2.0d0*PI**1*sqrt(PI)/(a**2*sqrt(a))
      R_1_norm=R_1_norm+c(21)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(24)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(26)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(31)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(33)*1.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(35)*3.0d0/4.0d0*PI**1*sqrt(PI)/(a**3*sqrt(a))
      R_1_norm=R_1_norm+c(57)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(60)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(62)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(67)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(69)*1.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(71)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(78)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(80)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(82)*3.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(84)*15.0d0/8.0d0*PI**1*sqrt(PI)/(a**4*sqrt(a))
      R_1_norm=R_1_norm+c(121)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(124)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(126)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(131)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(133)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(135)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(142)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(144)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(146)*3.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(148)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(157)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(159)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(161)*9.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(163)*15.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(165)*105.0d0/16.0d0*PI**1*sqrt(PI)/(a**5*sqrt(a))
      R_1_norm=R_1_norm+c(221)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(224)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(226)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(231)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(233)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(235)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(242)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(244)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(246)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(248)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(257)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(259)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(261)*9.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(263)*15.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(265)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(276)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(278)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(280)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(282)*45.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(284)*105.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(286)*945.0d0/32.0d0*PI**1*sqrt(PI)/(a**6*sqrt(a))
      R_1_norm=R_1_norm+c(365)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(368)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(370)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(375)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(377)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(379)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(386)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(388)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(390)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(392)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(401)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(403)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(405)*27.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(407)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(409)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(420)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(422)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(424)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(426)*45.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(428)*105.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(430)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(443)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(445)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(447)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(449)*225.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(451)*315.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(453)*945.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(455)*10395.0d0/64.0d0*PI**1*sqrt(PI)/(a**7*sqrt(a))
      R_1_norm=R_1_norm+c(561)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(564)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(566)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(571)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(573)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(575)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(582)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(584)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(586)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(588)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(597)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(599)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(601)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(603)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(605)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(616)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(618)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(620)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(622)*135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(624)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(626)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(639)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(641)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(643)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(645)*225.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(647)*315.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(649)*945.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(651)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(666)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(668)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(670)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(672)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(674)*1575.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(676)*2835.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(678)*10395.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(680)*135135.0d0/128.0d0*PI**1*sqrt(PI)/(a**8*sqrt(a))
      R_1_norm=R_1_norm+c(817)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(820)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(822)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(827)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(829)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(831)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(838)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(840)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(842)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(844)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(853)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(855)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(857)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(859)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(861)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(872)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(874)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(876)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(878)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(880)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(882)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(895)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(897)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(899)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(901)*675.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(903)*945.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(905)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(907)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(922)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(924)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(926)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(928)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(930)*1575.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(932)*2835.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(934)*10395.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(936)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(953)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(955)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(957)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(959)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(961)*11025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(963)*14175.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(965)*31185.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(967)*135135.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(969)*2027025.0d0/256.0d0*PI**1*sqrt(PI)/(a**9*sqrt(a))
      R_1_norm=R_1_norm+c(1141)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1144)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1146)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1151)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1153)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1155)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1162)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1164)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1166)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1168)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1177)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1179)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1181)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1183)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1185)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1196)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1198)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1200)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1202)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1204)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1206)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1219)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1221)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1223)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1225)*3375.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1227)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1229)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1231)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1246)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1248)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1250)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1252)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1254)*4725.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1256)*8505.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1258)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1260)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1277)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1279)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1281)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1283)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1285)*11025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1287)*14175.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1289)*31185.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1291)*135135.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1293)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1312)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1314)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1316)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1318)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1320)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1322)*99225.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1324)*155925.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1326)*405405.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1328)*2027025.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1330)*34459425.0d0/512.0d0*PI**1*sqrt(PI)/(a**10*sqrt(a))
      R_1_norm=R_1_norm+c(1541)*654729075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1544)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1546)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1551)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1553)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1555)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1562)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1564)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1566)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1568)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1577)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1579)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1581)*93555.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1583)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1585)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1596)*893025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1598)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1600)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1602)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1604)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1606)*893025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1619)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1621)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1623)*33075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1625)*23625.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1627)*33075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1629)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1631)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1646)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1648)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1650)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1652)*23625.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1654)*23625.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1656)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1658)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1660)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1677)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1679)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1681)*93555.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1683)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1685)*33075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1687)*42525.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1689)*93555.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1691)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1693)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1712)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1714)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1716)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1718)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1720)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1722)*99225.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1724)*155925.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1726)*405405.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1728)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1730)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1751)*654729075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1753)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1755)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1757)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1759)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1761)*893025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1763)*1091475.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1765)*2027025.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1767)*6081075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1769)*34459425.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(1771)*654729075.0d0/1024.0d0*PI**1*sqrt(PI)/(a**11*sqrt(a))
      R_1_norm=R_1_norm+c(2025)*13749310575.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2028)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2030)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2035)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2037)*34459425.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2039)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2046)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2048)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2050)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2052)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2061)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2063)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2065)*1216215.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2067)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2069)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2080)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2082)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2084)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2086)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2088)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2090)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2103)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2105)*893025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2107)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2109)*212625.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2111)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2113)*893025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2115)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2130)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2132)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2134)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2136)*165375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2138)*165375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2140)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2142)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2144)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2161)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2163)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2165)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2167)*212625.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2169)*165375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2171)*212625.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2173)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2175)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2177)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2196)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2198)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2200)*1216215.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2202)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2204)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2206)*297675.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2208)*467775.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2210)*1216215.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2212)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2214)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2235)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2237)*34459425.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2239)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2241)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2243)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2245)*893025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2247)*1091475.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2249)*2027025.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2251)*6081075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2253)*34459425.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2255)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2278)*13749310575.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2280)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2282)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2284)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2286)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2288)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2290)*9823275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2292)*14189175.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2294)*30405375.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2296)*103378275.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2298)*654729075.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2300)*13749310575.0d0/2048.0d0*PI**1*sqrt(PI)/(a**12*sqrt(a))
      R_1_norm=R_1_norm+c(2601)*316234143225.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2604)*13749310575.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2606)*13749310575.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2611)*1964187225.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2613)*654729075.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2615)*1964187225.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2622)*516891375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2624)*103378275.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2626)*103378275.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2628)*516891375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2637)*212837625.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2639)*30405375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2641)*18243225.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2643)*30405375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2645)*212837625.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2656)*127702575.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2658)*14189175.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2660)*6081075.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2662)*6081075.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2664)*14189175.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2666)*127702575.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2679)*108056025.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2681)*9823275.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2683)*3274425.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2685)*2338875.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2687)*3274425.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2689)*9823275.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2691)*108056025.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2706)*127702575.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2708)*9823275.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2710)*2679075.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2712)*1488375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2714)*1488375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2716)*2679075.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2718)*9823275.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2720)*127702575.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2737)*212837625.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2739)*14189175.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2741)*3274425.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2743)*1488375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2745)*1157625.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2747)*1488375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2749)*3274425.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2751)*14189175.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2753)*212837625.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2772)*516891375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2774)*30405375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2776)*6081075.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2778)*2338875.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2780)*1488375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2782)*1488375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2784)*2338875.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2786)*6081075.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2788)*30405375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2790)*516891375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2811)*1964187225.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2813)*103378275.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2815)*18243225.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2817)*6081075.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2819)*3274425.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2821)*2679075.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2823)*3274425.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2825)*6081075.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2827)*18243225.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2829)*103378275.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2831)*1964187225.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2854)*13749310575.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2856)*654729075.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2858)*103378275.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2860)*30405375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2862)*14189175.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2864)*9823275.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2866)*9823275.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2868)*14189175.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2870)*30405375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2872)*103378275.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2874)*654729075.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2876)*13749310575.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2901)*316234143225.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2903)*13749310575.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2905)*1964187225.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2907)*516891375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2909)*212837625.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2911)*127702575.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2913)*108056025.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2915)*127702575.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2917)*212837625.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2919)*516891375.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2921)*1964187225.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2923)*13749310575.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
      R_1_norm=R_1_norm+c(2925)*316234143225.0d0/4096.0d0*PI**1*sqrt(PI)/(a**13*sqrt(a))
    case default
      print*,'Error: R_1_norm not implemented for lmax=',lmax
      stop
  end select
 
end function

end module
