
!> This module defines utility routines for converting 
!! x^i*y^j*z^k*exp(-a*r^2) functions and their products
!! in terms of derivatives of the exp(-a*r^2) function
!!
!! These routines are generated automatically through 
!! maxima scripting.
!!
!!
!! Author: I. Duchemin 10 Oct 2012
!!
module mod_R_to_D_Conversion

  implicit none

  private
  public :: R_to_D

contains

!> expansion of x^i*y^j*z^k*exp(-a*r^2) in terms of derivatives of exp(-a*r^2)
!!
recursive subroutine R_to_D(a,ir,c)
 
  implicit none
 
  integer     , intent(in):: ir!< index of the element in the r basis
  real(kind=8), intent(in):: a!< gaussian exponant of the cubic harmonic
  real(kind=8), intent(out),dimension(455) :: c!< coefficient of decomposition in the basis {1,x,y,z,x^2,...}
 
  real(kind=8), parameter :: PI=3.1415926535897932384626433832795d0
 
  ! set c to 0
  c=0.0d0
 
  ! selection on ir
  select case(ir)
    case (1)
      c(1)=1.0d0
    case (2)
      c(2)=1.0d0/2.0d0/a
    case (3)
      c(3)=1.0d0/2.0d0/a
    case (4)
      c(4)=1.0d0/2.0d0/a
    case (5)
      c(1)=1.0d0/2.0d0/a
      c(5)=1.0d0/4.0d0/(a**2)
    case (6)
      c(6)=1.0d0/4.0d0/(a**2)
    case (7)
      c(7)=1.0d0/4.0d0/(a**2)
    case (8)
      c(1)=1.0d0/2.0d0/a
      c(8)=1.0d0/4.0d0/(a**2)
    case (9)
      c(9)=1.0d0/4.0d0/(a**2)
    case (10)
      c(1)=1.0d0/2.0d0/a
      c(10)=1.0d0/4.0d0/(a**2)
    case (11)
      c(2)=3.0d0/4.0d0/(a**2)
      c(11)=1.0d0/8.0d0/(a**3)
    case (12)
      c(3)=1.0d0/4.0d0/(a**2)
      c(12)=1.0d0/8.0d0/(a**3)
    case (13)
      c(4)=1.0d0/4.0d0/(a**2)
      c(13)=1.0d0/8.0d0/(a**3)
    case (14)
      c(2)=1.0d0/4.0d0/(a**2)
      c(14)=1.0d0/8.0d0/(a**3)
    case (15)
      c(15)=1.0d0/8.0d0/(a**3)
    case (16)
      c(2)=1.0d0/4.0d0/(a**2)
      c(16)=1.0d0/8.0d0/(a**3)
    case (17)
      c(3)=3.0d0/4.0d0/(a**2)
      c(17)=1.0d0/8.0d0/(a**3)
    case (18)
      c(4)=1.0d0/4.0d0/(a**2)
      c(18)=1.0d0/8.0d0/(a**3)
    case (19)
      c(3)=1.0d0/4.0d0/(a**2)
      c(19)=1.0d0/8.0d0/(a**3)
    case (20)
      c(4)=3.0d0/4.0d0/(a**2)
      c(20)=1.0d0/8.0d0/(a**3)
    case (21)
      c(1)=3.0d0/4.0d0/(a**2)
      c(5)=3.0d0/4.0d0/(a**3)
      c(21)=1.0d0/16.0d0/(a**4)
    case (22)
      c(6)=3.0d0/8.0d0/(a**3)
      c(22)=1.0d0/16.0d0/(a**4)
    case (23)
      c(7)=3.0d0/8.0d0/(a**3)
      c(23)=1.0d0/16.0d0/(a**4)
    case (24)
      c(1)=1.0d0/4.0d0/(a**2)
      c(5)=1.0d0/8.0d0/(a**3)
      c(8)=1.0d0/8.0d0/(a**3)
      c(24)=1.0d0/16.0d0/(a**4)
    case (25)
      c(9)=1.0d0/8.0d0/(a**3)
      c(25)=1.0d0/16.0d0/(a**4)
    case (26)
      c(1)=1.0d0/4.0d0/(a**2)
      c(5)=1.0d0/8.0d0/(a**3)
      c(10)=1.0d0/8.0d0/(a**3)
      c(26)=1.0d0/16.0d0/(a**4)
    case (27)
      c(6)=3.0d0/8.0d0/(a**3)
      c(27)=1.0d0/16.0d0/(a**4)
    case (28)
      c(7)=1.0d0/8.0d0/(a**3)
      c(28)=1.0d0/16.0d0/(a**4)
    case (29)
      c(6)=1.0d0/8.0d0/(a**3)
      c(29)=1.0d0/16.0d0/(a**4)
    case (30)
      c(7)=3.0d0/8.0d0/(a**3)
      c(30)=1.0d0/16.0d0/(a**4)
    case (31)
      c(1)=3.0d0/4.0d0/(a**2)
      c(8)=3.0d0/4.0d0/(a**3)
      c(31)=1.0d0/16.0d0/(a**4)
    case (32)
      c(9)=3.0d0/8.0d0/(a**3)
      c(32)=1.0d0/16.0d0/(a**4)
    case (33)
      c(1)=1.0d0/4.0d0/(a**2)
      c(8)=1.0d0/8.0d0/(a**3)
      c(10)=1.0d0/8.0d0/(a**3)
      c(33)=1.0d0/16.0d0/(a**4)
    case (34)
      c(9)=3.0d0/8.0d0/(a**3)
      c(34)=1.0d0/16.0d0/(a**4)
    case (35)
      c(1)=3.0d0/4.0d0/(a**2)
      c(10)=3.0d0/4.0d0/(a**3)
      c(35)=1.0d0/16.0d0/(a**4)
    case (36)
      c(2)=15.0d0/8.0d0/(a**3)
      c(11)=5.0d0/8.0d0/(a**4)
      c(36)=1.0d0/32.0d0/(a**5)
    case (37)
      c(3)=3.0d0/8.0d0/(a**3)
      c(12)=3.0d0/8.0d0/(a**4)
      c(37)=1.0d0/32.0d0/(a**5)
    case (38)
      c(4)=3.0d0/8.0d0/(a**3)
      c(13)=3.0d0/8.0d0/(a**4)
      c(38)=1.0d0/32.0d0/(a**5)
    case (39)
      c(2)=3.0d0/8.0d0/(a**3)
      c(11)=1.0d0/16.0d0/(a**4)
      c(14)=3.0d0/16.0d0/(a**4)
      c(39)=1.0d0/32.0d0/(a**5)
    case (40)
      c(15)=3.0d0/16.0d0/(a**4)
      c(40)=1.0d0/32.0d0/(a**5)
    case (41)
      c(2)=3.0d0/8.0d0/(a**3)
      c(11)=1.0d0/16.0d0/(a**4)
      c(16)=3.0d0/16.0d0/(a**4)
      c(41)=1.0d0/32.0d0/(a**5)
    case (42)
      c(3)=3.0d0/8.0d0/(a**3)
      c(12)=3.0d0/16.0d0/(a**4)
      c(17)=1.0d0/16.0d0/(a**4)
      c(42)=1.0d0/32.0d0/(a**5)
    case (43)
      c(4)=1.0d0/8.0d0/(a**3)
      c(13)=1.0d0/16.0d0/(a**4)
      c(18)=1.0d0/16.0d0/(a**4)
      c(43)=1.0d0/32.0d0/(a**5)
    case (44)
      c(3)=1.0d0/8.0d0/(a**3)
      c(12)=1.0d0/16.0d0/(a**4)
      c(19)=1.0d0/16.0d0/(a**4)
      c(44)=1.0d0/32.0d0/(a**5)
    case (45)
      c(4)=3.0d0/8.0d0/(a**3)
      c(13)=3.0d0/16.0d0/(a**4)
      c(20)=1.0d0/16.0d0/(a**4)
      c(45)=1.0d0/32.0d0/(a**5)
    case (46)
      c(2)=3.0d0/8.0d0/(a**3)
      c(14)=3.0d0/8.0d0/(a**4)
      c(46)=1.0d0/32.0d0/(a**5)
    case (47)
      c(15)=3.0d0/16.0d0/(a**4)
      c(47)=1.0d0/32.0d0/(a**5)
    case (48)
      c(2)=1.0d0/8.0d0/(a**3)
      c(14)=1.0d0/16.0d0/(a**4)
      c(16)=1.0d0/16.0d0/(a**4)
      c(48)=1.0d0/32.0d0/(a**5)
    case (49)
      c(15)=3.0d0/16.0d0/(a**4)
      c(49)=1.0d0/32.0d0/(a**5)
    case (50)
      c(2)=3.0d0/8.0d0/(a**3)
      c(16)=3.0d0/8.0d0/(a**4)
      c(50)=1.0d0/32.0d0/(a**5)
    case (51)
      c(3)=15.0d0/8.0d0/(a**3)
      c(17)=5.0d0/8.0d0/(a**4)
      c(51)=1.0d0/32.0d0/(a**5)
    case (52)
      c(4)=3.0d0/8.0d0/(a**3)
      c(18)=3.0d0/8.0d0/(a**4)
      c(52)=1.0d0/32.0d0/(a**5)
    case (53)
      c(3)=3.0d0/8.0d0/(a**3)
      c(17)=1.0d0/16.0d0/(a**4)
      c(19)=3.0d0/16.0d0/(a**4)
      c(53)=1.0d0/32.0d0/(a**5)
    case (54)
      c(4)=3.0d0/8.0d0/(a**3)
      c(18)=3.0d0/16.0d0/(a**4)
      c(20)=1.0d0/16.0d0/(a**4)
      c(54)=1.0d0/32.0d0/(a**5)
    case (55)
      c(3)=3.0d0/8.0d0/(a**3)
      c(19)=3.0d0/8.0d0/(a**4)
      c(55)=1.0d0/32.0d0/(a**5)
    case (56)
      c(4)=15.0d0/8.0d0/(a**3)
      c(20)=5.0d0/8.0d0/(a**4)
      c(56)=1.0d0/32.0d0/(a**5)
    case (57)
      c(1)=15.0d0/8.0d0/(a**3)
      c(5)=45.0d0/16.0d0/(a**4)
      c(21)=15.0d0/32.0d0/(a**5)
      c(57)=1.0d0/64.0d0/(a**6)
    case (58)
      c(6)=15.0d0/16.0d0/(a**4)
      c(22)=5.0d0/16.0d0/(a**5)
      c(58)=1.0d0/64.0d0/(a**6)
    case (59)
      c(7)=15.0d0/16.0d0/(a**4)
      c(23)=5.0d0/16.0d0/(a**5)
      c(59)=1.0d0/64.0d0/(a**6)
    case (60)
      c(1)=3.0d0/8.0d0/(a**3)
      c(5)=3.0d0/8.0d0/(a**4)
      c(8)=3.0d0/16.0d0/(a**4)
      c(21)=1.0d0/32.0d0/(a**5)
      c(24)=3.0d0/16.0d0/(a**5)
      c(60)=1.0d0/64.0d0/(a**6)
    case (61)
      c(9)=3.0d0/16.0d0/(a**4)
      c(25)=3.0d0/16.0d0/(a**5)
      c(61)=1.0d0/64.0d0/(a**6)
    case (62)
      c(1)=3.0d0/8.0d0/(a**3)
      c(5)=3.0d0/8.0d0/(a**4)
      c(10)=3.0d0/16.0d0/(a**4)
      c(21)=1.0d0/32.0d0/(a**5)
      c(26)=3.0d0/16.0d0/(a**5)
      c(62)=1.0d0/64.0d0/(a**6)
    case (63)
      c(6)=9.0d0/16.0d0/(a**4)
      c(22)=3.0d0/32.0d0/(a**5)
      c(27)=3.0d0/32.0d0/(a**5)
      c(63)=1.0d0/64.0d0/(a**6)
    case (64)
      c(7)=3.0d0/16.0d0/(a**4)
      c(23)=1.0d0/32.0d0/(a**5)
      c(28)=3.0d0/32.0d0/(a**5)
      c(64)=1.0d0/64.0d0/(a**6)
    case (65)
      c(6)=3.0d0/16.0d0/(a**4)
      c(22)=1.0d0/32.0d0/(a**5)
      c(29)=3.0d0/32.0d0/(a**5)
      c(65)=1.0d0/64.0d0/(a**6)
    case (66)
      c(7)=9.0d0/16.0d0/(a**4)
      c(23)=3.0d0/32.0d0/(a**5)
      c(30)=3.0d0/32.0d0/(a**5)
      c(66)=1.0d0/64.0d0/(a**6)
    case (67)
      c(1)=3.0d0/8.0d0/(a**3)
      c(5)=3.0d0/16.0d0/(a**4)
      c(8)=3.0d0/8.0d0/(a**4)
      c(24)=3.0d0/16.0d0/(a**5)
      c(31)=1.0d0/32.0d0/(a**5)
      c(67)=1.0d0/64.0d0/(a**6)
    case (68)
      c(9)=3.0d0/16.0d0/(a**4)
      c(25)=3.0d0/32.0d0/(a**5)
      c(32)=1.0d0/32.0d0/(a**5)
      c(68)=1.0d0/64.0d0/(a**6)
    case (69)
      c(1)=1.0d0/8.0d0/(a**3)
      c(5)=1.0d0/16.0d0/(a**4)
      c(8)=1.0d0/16.0d0/(a**4)
      c(10)=1.0d0/16.0d0/(a**4)
      c(24)=1.0d0/32.0d0/(a**5)
      c(26)=1.0d0/32.0d0/(a**5)
      c(33)=1.0d0/32.0d0/(a**5)
      c(69)=1.0d0/64.0d0/(a**6)
    case (70)
      c(9)=3.0d0/16.0d0/(a**4)
      c(25)=3.0d0/32.0d0/(a**5)
      c(34)=1.0d0/32.0d0/(a**5)
      c(70)=1.0d0/64.0d0/(a**6)
    case (71)
      c(1)=3.0d0/8.0d0/(a**3)
      c(5)=3.0d0/16.0d0/(a**4)
      c(10)=3.0d0/8.0d0/(a**4)
      c(26)=3.0d0/16.0d0/(a**5)
      c(35)=1.0d0/32.0d0/(a**5)
      c(71)=1.0d0/64.0d0/(a**6)
    case (72)
      c(6)=15.0d0/16.0d0/(a**4)
      c(27)=5.0d0/16.0d0/(a**5)
      c(72)=1.0d0/64.0d0/(a**6)
    case (73)
      c(7)=3.0d0/16.0d0/(a**4)
      c(28)=3.0d0/16.0d0/(a**5)
      c(73)=1.0d0/64.0d0/(a**6)
    case (74)
      c(6)=3.0d0/16.0d0/(a**4)
      c(27)=1.0d0/32.0d0/(a**5)
      c(29)=3.0d0/32.0d0/(a**5)
      c(74)=1.0d0/64.0d0/(a**6)
    case (75)
      c(7)=3.0d0/16.0d0/(a**4)
      c(28)=3.0d0/32.0d0/(a**5)
      c(30)=1.0d0/32.0d0/(a**5)
      c(75)=1.0d0/64.0d0/(a**6)
    case (76)
      c(6)=3.0d0/16.0d0/(a**4)
      c(29)=3.0d0/16.0d0/(a**5)
      c(76)=1.0d0/64.0d0/(a**6)
    case (77)
      c(7)=15.0d0/16.0d0/(a**4)
      c(30)=5.0d0/16.0d0/(a**5)
      c(77)=1.0d0/64.0d0/(a**6)
    case (78)
      c(1)=15.0d0/8.0d0/(a**3)
      c(8)=45.0d0/16.0d0/(a**4)
      c(31)=15.0d0/32.0d0/(a**5)
      c(78)=1.0d0/64.0d0/(a**6)
    case (79)
      c(9)=15.0d0/16.0d0/(a**4)
      c(32)=5.0d0/16.0d0/(a**5)
      c(79)=1.0d0/64.0d0/(a**6)
    case (80)
      c(1)=3.0d0/8.0d0/(a**3)
      c(8)=3.0d0/8.0d0/(a**4)
      c(10)=3.0d0/16.0d0/(a**4)
      c(31)=1.0d0/32.0d0/(a**5)
      c(33)=3.0d0/16.0d0/(a**5)
      c(80)=1.0d0/64.0d0/(a**6)
    case (81)
      c(9)=9.0d0/16.0d0/(a**4)
      c(32)=3.0d0/32.0d0/(a**5)
      c(34)=3.0d0/32.0d0/(a**5)
      c(81)=1.0d0/64.0d0/(a**6)
    case (82)
      c(1)=3.0d0/8.0d0/(a**3)
      c(8)=3.0d0/16.0d0/(a**4)
      c(10)=3.0d0/8.0d0/(a**4)
      c(33)=3.0d0/16.0d0/(a**5)
      c(35)=1.0d0/32.0d0/(a**5)
      c(82)=1.0d0/64.0d0/(a**6)
    case (83)
      c(9)=15.0d0/16.0d0/(a**4)
      c(34)=5.0d0/16.0d0/(a**5)
      c(83)=1.0d0/64.0d0/(a**6)
    case (84)
      c(1)=15.0d0/8.0d0/(a**3)
      c(10)=45.0d0/16.0d0/(a**4)
      c(35)=15.0d0/32.0d0/(a**5)
      c(84)=1.0d0/64.0d0/(a**6)
    case (85)
      c(2)=105.0d0/16.0d0/(a**4)
      c(11)=105.0d0/32.0d0/(a**5)
      c(36)=21.0d0/64.0d0/(a**6)
      c(85)=1.0d0/128.0d0/(a**7)
    case (86)
      c(3)=15.0d0/16.0d0/(a**4)
      c(12)=45.0d0/32.0d0/(a**5)
      c(37)=15.0d0/64.0d0/(a**6)
      c(86)=1.0d0/128.0d0/(a**7)
    case (87)
      c(4)=15.0d0/16.0d0/(a**4)
      c(13)=45.0d0/32.0d0/(a**5)
      c(38)=15.0d0/64.0d0/(a**6)
      c(87)=1.0d0/128.0d0/(a**7)
    case (88)
      c(2)=15.0d0/16.0d0/(a**4)
      c(11)=5.0d0/16.0d0/(a**5)
      c(14)=15.0d0/32.0d0/(a**5)
      c(36)=1.0d0/64.0d0/(a**6)
      c(39)=5.0d0/32.0d0/(a**6)
      c(88)=1.0d0/128.0d0/(a**7)
    case (89)
      c(15)=15.0d0/32.0d0/(a**5)
      c(40)=5.0d0/32.0d0/(a**6)
      c(89)=1.0d0/128.0d0/(a**7)
    case (90)
      c(2)=15.0d0/16.0d0/(a**4)
      c(11)=5.0d0/16.0d0/(a**5)
      c(16)=15.0d0/32.0d0/(a**5)
      c(36)=1.0d0/64.0d0/(a**6)
      c(41)=5.0d0/32.0d0/(a**6)
      c(90)=1.0d0/128.0d0/(a**7)
    case (91)
      c(3)=9.0d0/16.0d0/(a**4)
      c(12)=9.0d0/16.0d0/(a**5)
      c(17)=3.0d0/32.0d0/(a**5)
      c(37)=3.0d0/64.0d0/(a**6)
      c(42)=3.0d0/32.0d0/(a**6)
      c(91)=1.0d0/128.0d0/(a**7)
    case (92)
      c(4)=3.0d0/16.0d0/(a**4)
      c(13)=3.0d0/16.0d0/(a**5)
      c(18)=3.0d0/32.0d0/(a**5)
      c(38)=1.0d0/64.0d0/(a**6)
      c(43)=3.0d0/32.0d0/(a**6)
      c(92)=1.0d0/128.0d0/(a**7)
    case (93)
      c(3)=3.0d0/16.0d0/(a**4)
      c(12)=3.0d0/16.0d0/(a**5)
      c(19)=3.0d0/32.0d0/(a**5)
      c(37)=1.0d0/64.0d0/(a**6)
      c(44)=3.0d0/32.0d0/(a**6)
      c(93)=1.0d0/128.0d0/(a**7)
    case (94)
      c(4)=9.0d0/16.0d0/(a**4)
      c(13)=9.0d0/16.0d0/(a**5)
      c(20)=3.0d0/32.0d0/(a**5)
      c(38)=3.0d0/64.0d0/(a**6)
      c(45)=3.0d0/32.0d0/(a**6)
      c(94)=1.0d0/128.0d0/(a**7)
    case (95)
      c(2)=9.0d0/16.0d0/(a**4)
      c(11)=3.0d0/32.0d0/(a**5)
      c(14)=9.0d0/16.0d0/(a**5)
      c(39)=3.0d0/32.0d0/(a**6)
      c(46)=3.0d0/64.0d0/(a**6)
      c(95)=1.0d0/128.0d0/(a**7)
    case (96)
      c(15)=9.0d0/32.0d0/(a**5)
      c(40)=3.0d0/64.0d0/(a**6)
      c(47)=3.0d0/64.0d0/(a**6)
      c(96)=1.0d0/128.0d0/(a**7)
    case (97)
      c(2)=3.0d0/16.0d0/(a**4)
      c(11)=1.0d0/32.0d0/(a**5)
      c(14)=3.0d0/32.0d0/(a**5)
      c(16)=3.0d0/32.0d0/(a**5)
      c(39)=1.0d0/64.0d0/(a**6)
      c(41)=1.0d0/64.0d0/(a**6)
      c(48)=3.0d0/64.0d0/(a**6)
      c(97)=1.0d0/128.0d0/(a**7)
    case (98)
      c(15)=9.0d0/32.0d0/(a**5)
      c(40)=3.0d0/64.0d0/(a**6)
      c(49)=3.0d0/64.0d0/(a**6)
      c(98)=1.0d0/128.0d0/(a**7)
    case (99)
      c(2)=9.0d0/16.0d0/(a**4)
      c(11)=3.0d0/32.0d0/(a**5)
      c(16)=9.0d0/16.0d0/(a**5)
      c(41)=3.0d0/32.0d0/(a**6)
      c(50)=3.0d0/64.0d0/(a**6)
      c(99)=1.0d0/128.0d0/(a**7)
    case (100)
      c(3)=15.0d0/16.0d0/(a**4)
      c(12)=15.0d0/32.0d0/(a**5)
      c(17)=5.0d0/16.0d0/(a**5)
      c(42)=5.0d0/32.0d0/(a**6)
      c(51)=1.0d0/64.0d0/(a**6)
      c(100)=1.0d0/128.0d0/(a**7)
    case (101)
      c(4)=3.0d0/16.0d0/(a**4)
      c(13)=3.0d0/32.0d0/(a**5)
      c(18)=3.0d0/16.0d0/(a**5)
      c(43)=3.0d0/32.0d0/(a**6)
      c(52)=1.0d0/64.0d0/(a**6)
      c(101)=1.0d0/128.0d0/(a**7)
    case (102)
      c(3)=3.0d0/16.0d0/(a**4)
      c(12)=3.0d0/32.0d0/(a**5)
      c(17)=1.0d0/32.0d0/(a**5)
      c(19)=3.0d0/32.0d0/(a**5)
      c(42)=1.0d0/64.0d0/(a**6)
      c(44)=3.0d0/64.0d0/(a**6)
      c(53)=1.0d0/64.0d0/(a**6)
      c(102)=1.0d0/128.0d0/(a**7)
    case (103)
      c(4)=3.0d0/16.0d0/(a**4)
      c(13)=3.0d0/32.0d0/(a**5)
      c(18)=3.0d0/32.0d0/(a**5)
      c(20)=1.0d0/32.0d0/(a**5)
      c(43)=3.0d0/64.0d0/(a**6)
      c(45)=1.0d0/64.0d0/(a**6)
      c(54)=1.0d0/64.0d0/(a**6)
      c(103)=1.0d0/128.0d0/(a**7)
    case (104)
      c(3)=3.0d0/16.0d0/(a**4)
      c(12)=3.0d0/32.0d0/(a**5)
      c(19)=3.0d0/16.0d0/(a**5)
      c(44)=3.0d0/32.0d0/(a**6)
      c(55)=1.0d0/64.0d0/(a**6)
      c(104)=1.0d0/128.0d0/(a**7)
    case (105)
      c(4)=15.0d0/16.0d0/(a**4)
      c(13)=15.0d0/32.0d0/(a**5)
      c(20)=5.0d0/16.0d0/(a**5)
      c(45)=5.0d0/32.0d0/(a**6)
      c(56)=1.0d0/64.0d0/(a**6)
      c(105)=1.0d0/128.0d0/(a**7)
    case (106)
      c(2)=15.0d0/16.0d0/(a**4)
      c(14)=45.0d0/32.0d0/(a**5)
      c(46)=15.0d0/64.0d0/(a**6)
      c(106)=1.0d0/128.0d0/(a**7)
    case (107)
      c(15)=15.0d0/32.0d0/(a**5)
      c(47)=5.0d0/32.0d0/(a**6)
      c(107)=1.0d0/128.0d0/(a**7)
    case (108)
      c(2)=3.0d0/16.0d0/(a**4)
      c(14)=3.0d0/16.0d0/(a**5)
      c(16)=3.0d0/32.0d0/(a**5)
      c(46)=1.0d0/64.0d0/(a**6)
      c(48)=3.0d0/32.0d0/(a**6)
      c(108)=1.0d0/128.0d0/(a**7)
    case (109)
      c(15)=9.0d0/32.0d0/(a**5)
      c(47)=3.0d0/64.0d0/(a**6)
      c(49)=3.0d0/64.0d0/(a**6)
      c(109)=1.0d0/128.0d0/(a**7)
    case (110)
      c(2)=3.0d0/16.0d0/(a**4)
      c(14)=3.0d0/32.0d0/(a**5)
      c(16)=3.0d0/16.0d0/(a**5)
      c(48)=3.0d0/32.0d0/(a**6)
      c(50)=1.0d0/64.0d0/(a**6)
      c(110)=1.0d0/128.0d0/(a**7)
    case (111)
      c(15)=15.0d0/32.0d0/(a**5)
      c(49)=5.0d0/32.0d0/(a**6)
      c(111)=1.0d0/128.0d0/(a**7)
    case (112)
      c(2)=15.0d0/16.0d0/(a**4)
      c(16)=45.0d0/32.0d0/(a**5)
      c(50)=15.0d0/64.0d0/(a**6)
      c(112)=1.0d0/128.0d0/(a**7)
    case (113)
      c(3)=105.0d0/16.0d0/(a**4)
      c(17)=105.0d0/32.0d0/(a**5)
      c(51)=21.0d0/64.0d0/(a**6)
      c(113)=1.0d0/128.0d0/(a**7)
    case (114)
      c(4)=15.0d0/16.0d0/(a**4)
      c(18)=45.0d0/32.0d0/(a**5)
      c(52)=15.0d0/64.0d0/(a**6)
      c(114)=1.0d0/128.0d0/(a**7)
    case (115)
      c(3)=15.0d0/16.0d0/(a**4)
      c(17)=5.0d0/16.0d0/(a**5)
      c(19)=15.0d0/32.0d0/(a**5)
      c(51)=1.0d0/64.0d0/(a**6)
      c(53)=5.0d0/32.0d0/(a**6)
      c(115)=1.0d0/128.0d0/(a**7)
    case (116)
      c(4)=9.0d0/16.0d0/(a**4)
      c(18)=9.0d0/16.0d0/(a**5)
      c(20)=3.0d0/32.0d0/(a**5)
      c(52)=3.0d0/64.0d0/(a**6)
      c(54)=3.0d0/32.0d0/(a**6)
      c(116)=1.0d0/128.0d0/(a**7)
    case (117)
      c(3)=9.0d0/16.0d0/(a**4)
      c(17)=3.0d0/32.0d0/(a**5)
      c(19)=9.0d0/16.0d0/(a**5)
      c(53)=3.0d0/32.0d0/(a**6)
      c(55)=3.0d0/64.0d0/(a**6)
      c(117)=1.0d0/128.0d0/(a**7)
    case (118)
      c(4)=15.0d0/16.0d0/(a**4)
      c(18)=15.0d0/32.0d0/(a**5)
      c(20)=5.0d0/16.0d0/(a**5)
      c(54)=5.0d0/32.0d0/(a**6)
      c(56)=1.0d0/64.0d0/(a**6)
      c(118)=1.0d0/128.0d0/(a**7)
    case (119)
      c(3)=15.0d0/16.0d0/(a**4)
      c(19)=45.0d0/32.0d0/(a**5)
      c(55)=15.0d0/64.0d0/(a**6)
      c(119)=1.0d0/128.0d0/(a**7)
    case (120)
      c(4)=105.0d0/16.0d0/(a**4)
      c(20)=105.0d0/32.0d0/(a**5)
      c(56)=21.0d0/64.0d0/(a**6)
      c(120)=1.0d0/128.0d0/(a**7)
    case (121)
      c(1)=105.0d0/16.0d0/(a**4)
      c(5)=105.0d0/8.0d0/(a**5)
      c(21)=105.0d0/32.0d0/(a**6)
      c(57)=7.0d0/32.0d0/(a**7)
      c(121)=1.0d0/256.0d0/(a**8)
    case (122)
      c(6)=105.0d0/32.0d0/(a**5)
      c(22)=105.0d0/64.0d0/(a**6)
      c(58)=21.0d0/128.0d0/(a**7)
      c(122)=1.0d0/256.0d0/(a**8)
    case (123)
      c(7)=105.0d0/32.0d0/(a**5)
      c(23)=105.0d0/64.0d0/(a**6)
      c(59)=21.0d0/128.0d0/(a**7)
      c(123)=1.0d0/256.0d0/(a**8)
    case (124)
      c(1)=15.0d0/16.0d0/(a**4)
      c(5)=45.0d0/32.0d0/(a**5)
      c(8)=15.0d0/32.0d0/(a**5)
      c(21)=15.0d0/64.0d0/(a**6)
      c(24)=45.0d0/64.0d0/(a**6)
      c(57)=1.0d0/128.0d0/(a**7)
      c(60)=15.0d0/128.0d0/(a**7)
      c(124)=1.0d0/256.0d0/(a**8)
    case (125)
      c(9)=15.0d0/32.0d0/(a**5)
      c(25)=45.0d0/64.0d0/(a**6)
      c(61)=15.0d0/128.0d0/(a**7)
      c(125)=1.0d0/256.0d0/(a**8)
    case (126)
      c(1)=15.0d0/16.0d0/(a**4)
      c(5)=45.0d0/32.0d0/(a**5)
      c(10)=15.0d0/32.0d0/(a**5)
      c(21)=15.0d0/64.0d0/(a**6)
      c(26)=45.0d0/64.0d0/(a**6)
      c(57)=1.0d0/128.0d0/(a**7)
      c(62)=15.0d0/128.0d0/(a**7)
      c(126)=1.0d0/256.0d0/(a**8)
    case (127)
      c(6)=45.0d0/32.0d0/(a**5)
      c(22)=15.0d0/32.0d0/(a**6)
      c(27)=15.0d0/64.0d0/(a**6)
      c(58)=3.0d0/128.0d0/(a**7)
      c(63)=5.0d0/64.0d0/(a**7)
      c(127)=1.0d0/256.0d0/(a**8)
    case (128)
      c(7)=15.0d0/32.0d0/(a**5)
      c(23)=5.0d0/32.0d0/(a**6)
      c(28)=15.0d0/64.0d0/(a**6)
      c(59)=1.0d0/128.0d0/(a**7)
      c(64)=5.0d0/64.0d0/(a**7)
      c(128)=1.0d0/256.0d0/(a**8)
    case (129)
      c(6)=15.0d0/32.0d0/(a**5)
      c(22)=5.0d0/32.0d0/(a**6)
      c(29)=15.0d0/64.0d0/(a**6)
      c(58)=1.0d0/128.0d0/(a**7)
      c(65)=5.0d0/64.0d0/(a**7)
      c(129)=1.0d0/256.0d0/(a**8)
    case (130)
      c(7)=45.0d0/32.0d0/(a**5)
      c(23)=15.0d0/32.0d0/(a**6)
      c(30)=15.0d0/64.0d0/(a**6)
      c(59)=3.0d0/128.0d0/(a**7)
      c(66)=5.0d0/64.0d0/(a**7)
      c(130)=1.0d0/256.0d0/(a**8)
    case (131)
      c(1)=9.0d0/16.0d0/(a**4)
      c(5)=9.0d0/16.0d0/(a**5)
      c(8)=9.0d0/16.0d0/(a**5)
      c(21)=3.0d0/64.0d0/(a**6)
      c(24)=9.0d0/16.0d0/(a**6)
      c(31)=3.0d0/64.0d0/(a**6)
      c(60)=3.0d0/64.0d0/(a**7)
      c(67)=3.0d0/64.0d0/(a**7)
      c(131)=1.0d0/256.0d0/(a**8)
    case (132)
      c(9)=9.0d0/32.0d0/(a**5)
      c(25)=9.0d0/32.0d0/(a**6)
      c(32)=3.0d0/64.0d0/(a**6)
      c(61)=3.0d0/128.0d0/(a**7)
      c(68)=3.0d0/64.0d0/(a**7)
      c(132)=1.0d0/256.0d0/(a**8)
    case (133)
      c(1)=3.0d0/16.0d0/(a**4)
      c(5)=3.0d0/16.0d0/(a**5)
      c(8)=3.0d0/32.0d0/(a**5)
      c(10)=3.0d0/32.0d0/(a**5)
      c(21)=1.0d0/64.0d0/(a**6)
      c(24)=3.0d0/32.0d0/(a**6)
      c(26)=3.0d0/32.0d0/(a**6)
      c(33)=3.0d0/64.0d0/(a**6)
      c(60)=1.0d0/128.0d0/(a**7)
      c(62)=1.0d0/128.0d0/(a**7)
      c(69)=3.0d0/64.0d0/(a**7)
      c(133)=1.0d0/256.0d0/(a**8)
    case (134)
      c(9)=9.0d0/32.0d0/(a**5)
      c(25)=9.0d0/32.0d0/(a**6)
      c(34)=3.0d0/64.0d0/(a**6)
      c(61)=3.0d0/128.0d0/(a**7)
      c(70)=3.0d0/64.0d0/(a**7)
      c(134)=1.0d0/256.0d0/(a**8)
    case (135)
      c(1)=9.0d0/16.0d0/(a**4)
      c(5)=9.0d0/16.0d0/(a**5)
      c(10)=9.0d0/16.0d0/(a**5)
      c(21)=3.0d0/64.0d0/(a**6)
      c(26)=9.0d0/16.0d0/(a**6)
      c(35)=3.0d0/64.0d0/(a**6)
      c(62)=3.0d0/64.0d0/(a**7)
      c(71)=3.0d0/64.0d0/(a**7)
      c(135)=1.0d0/256.0d0/(a**8)
    case (136)
      c(6)=45.0d0/32.0d0/(a**5)
      c(22)=15.0d0/64.0d0/(a**6)
      c(27)=15.0d0/32.0d0/(a**6)
      c(63)=5.0d0/64.0d0/(a**7)
      c(72)=3.0d0/128.0d0/(a**7)
      c(136)=1.0d0/256.0d0/(a**8)
    case (137)
      c(7)=9.0d0/32.0d0/(a**5)
      c(23)=3.0d0/64.0d0/(a**6)
      c(28)=9.0d0/32.0d0/(a**6)
      c(64)=3.0d0/64.0d0/(a**7)
      c(73)=3.0d0/128.0d0/(a**7)
      c(137)=1.0d0/256.0d0/(a**8)
    case (138)
      c(6)=9.0d0/32.0d0/(a**5)
      c(22)=3.0d0/64.0d0/(a**6)
      c(27)=3.0d0/64.0d0/(a**6)
      c(29)=9.0d0/64.0d0/(a**6)
      c(63)=1.0d0/128.0d0/(a**7)
      c(65)=3.0d0/128.0d0/(a**7)
      c(74)=3.0d0/128.0d0/(a**7)
      c(138)=1.0d0/256.0d0/(a**8)
    case (139)
      c(7)=9.0d0/32.0d0/(a**5)
      c(23)=3.0d0/64.0d0/(a**6)
      c(28)=9.0d0/64.0d0/(a**6)
      c(30)=3.0d0/64.0d0/(a**6)
      c(64)=3.0d0/128.0d0/(a**7)
      c(66)=1.0d0/128.0d0/(a**7)
      c(75)=3.0d0/128.0d0/(a**7)
      c(139)=1.0d0/256.0d0/(a**8)
    case (140)
      c(6)=9.0d0/32.0d0/(a**5)
      c(22)=3.0d0/64.0d0/(a**6)
      c(29)=9.0d0/32.0d0/(a**6)
      c(65)=3.0d0/64.0d0/(a**7)
      c(76)=3.0d0/128.0d0/(a**7)
      c(140)=1.0d0/256.0d0/(a**8)
    case (141)
      c(7)=45.0d0/32.0d0/(a**5)
      c(23)=15.0d0/64.0d0/(a**6)
      c(30)=15.0d0/32.0d0/(a**6)
      c(66)=5.0d0/64.0d0/(a**7)
      c(77)=3.0d0/128.0d0/(a**7)
      c(141)=1.0d0/256.0d0/(a**8)
    case (142)
      c(1)=15.0d0/16.0d0/(a**4)
      c(5)=15.0d0/32.0d0/(a**5)
      c(8)=45.0d0/32.0d0/(a**5)
      c(24)=45.0d0/64.0d0/(a**6)
      c(31)=15.0d0/64.0d0/(a**6)
      c(67)=15.0d0/128.0d0/(a**7)
      c(78)=1.0d0/128.0d0/(a**7)
      c(142)=1.0d0/256.0d0/(a**8)
    case (143)
      c(9)=15.0d0/32.0d0/(a**5)
      c(25)=15.0d0/64.0d0/(a**6)
      c(32)=5.0d0/32.0d0/(a**6)
      c(68)=5.0d0/64.0d0/(a**7)
      c(79)=1.0d0/128.0d0/(a**7)
      c(143)=1.0d0/256.0d0/(a**8)
    case (144)
      c(1)=3.0d0/16.0d0/(a**4)
      c(5)=3.0d0/32.0d0/(a**5)
      c(8)=3.0d0/16.0d0/(a**5)
      c(10)=3.0d0/32.0d0/(a**5)
      c(24)=3.0d0/32.0d0/(a**6)
      c(26)=3.0d0/64.0d0/(a**6)
      c(31)=1.0d0/64.0d0/(a**6)
      c(33)=3.0d0/32.0d0/(a**6)
      c(67)=1.0d0/128.0d0/(a**7)
      c(69)=3.0d0/64.0d0/(a**7)
      c(80)=1.0d0/128.0d0/(a**7)
      c(144)=1.0d0/256.0d0/(a**8)
    case (145)
      c(9)=9.0d0/32.0d0/(a**5)
      c(25)=9.0d0/64.0d0/(a**6)
      c(32)=3.0d0/64.0d0/(a**6)
      c(34)=3.0d0/64.0d0/(a**6)
      c(68)=3.0d0/128.0d0/(a**7)
      c(70)=3.0d0/128.0d0/(a**7)
      c(81)=1.0d0/128.0d0/(a**7)
      c(145)=1.0d0/256.0d0/(a**8)
    case (146)
      c(1)=3.0d0/16.0d0/(a**4)
      c(5)=3.0d0/32.0d0/(a**5)
      c(8)=3.0d0/32.0d0/(a**5)
      c(10)=3.0d0/16.0d0/(a**5)
      c(24)=3.0d0/64.0d0/(a**6)
      c(26)=3.0d0/32.0d0/(a**6)
      c(33)=3.0d0/32.0d0/(a**6)
      c(35)=1.0d0/64.0d0/(a**6)
      c(69)=3.0d0/64.0d0/(a**7)
      c(71)=1.0d0/128.0d0/(a**7)
      c(82)=1.0d0/128.0d0/(a**7)
      c(146)=1.0d0/256.0d0/(a**8)
    case (147)
      c(9)=15.0d0/32.0d0/(a**5)
      c(25)=15.0d0/64.0d0/(a**6)
      c(34)=5.0d0/32.0d0/(a**6)
      c(70)=5.0d0/64.0d0/(a**7)
      c(83)=1.0d0/128.0d0/(a**7)
      c(147)=1.0d0/256.0d0/(a**8)
    case (148)
      c(1)=15.0d0/16.0d0/(a**4)
      c(5)=15.0d0/32.0d0/(a**5)
      c(10)=45.0d0/32.0d0/(a**5)
      c(26)=45.0d0/64.0d0/(a**6)
      c(35)=15.0d0/64.0d0/(a**6)
      c(71)=15.0d0/128.0d0/(a**7)
      c(84)=1.0d0/128.0d0/(a**7)
      c(148)=1.0d0/256.0d0/(a**8)
    case (149)
      c(6)=105.0d0/32.0d0/(a**5)
      c(27)=105.0d0/64.0d0/(a**6)
      c(72)=21.0d0/128.0d0/(a**7)
      c(149)=1.0d0/256.0d0/(a**8)
    case (150)
      c(7)=15.0d0/32.0d0/(a**5)
      c(28)=45.0d0/64.0d0/(a**6)
      c(73)=15.0d0/128.0d0/(a**7)
      c(150)=1.0d0/256.0d0/(a**8)
    case (151)
      c(6)=15.0d0/32.0d0/(a**5)
      c(27)=5.0d0/32.0d0/(a**6)
      c(29)=15.0d0/64.0d0/(a**6)
      c(72)=1.0d0/128.0d0/(a**7)
      c(74)=5.0d0/64.0d0/(a**7)
      c(151)=1.0d0/256.0d0/(a**8)
    case (152)
      c(7)=9.0d0/32.0d0/(a**5)
      c(28)=9.0d0/32.0d0/(a**6)
      c(30)=3.0d0/64.0d0/(a**6)
      c(73)=3.0d0/128.0d0/(a**7)
      c(75)=3.0d0/64.0d0/(a**7)
      c(152)=1.0d0/256.0d0/(a**8)
    case (153)
      c(6)=9.0d0/32.0d0/(a**5)
      c(27)=3.0d0/64.0d0/(a**6)
      c(29)=9.0d0/32.0d0/(a**6)
      c(74)=3.0d0/64.0d0/(a**7)
      c(76)=3.0d0/128.0d0/(a**7)
      c(153)=1.0d0/256.0d0/(a**8)
    case (154)
      c(7)=15.0d0/32.0d0/(a**5)
      c(28)=15.0d0/64.0d0/(a**6)
      c(30)=5.0d0/32.0d0/(a**6)
      c(75)=5.0d0/64.0d0/(a**7)
      c(77)=1.0d0/128.0d0/(a**7)
      c(154)=1.0d0/256.0d0/(a**8)
    case (155)
      c(6)=15.0d0/32.0d0/(a**5)
      c(29)=45.0d0/64.0d0/(a**6)
      c(76)=15.0d0/128.0d0/(a**7)
      c(155)=1.0d0/256.0d0/(a**8)
    case (156)
      c(7)=105.0d0/32.0d0/(a**5)
      c(30)=105.0d0/64.0d0/(a**6)
      c(77)=21.0d0/128.0d0/(a**7)
      c(156)=1.0d0/256.0d0/(a**8)
    case (157)
      c(1)=105.0d0/16.0d0/(a**4)
      c(8)=105.0d0/8.0d0/(a**5)
      c(31)=105.0d0/32.0d0/(a**6)
      c(78)=7.0d0/32.0d0/(a**7)
      c(157)=1.0d0/256.0d0/(a**8)
    case (158)
      c(9)=105.0d0/32.0d0/(a**5)
      c(32)=105.0d0/64.0d0/(a**6)
      c(79)=21.0d0/128.0d0/(a**7)
      c(158)=1.0d0/256.0d0/(a**8)
    case (159)
      c(1)=15.0d0/16.0d0/(a**4)
      c(8)=45.0d0/32.0d0/(a**5)
      c(10)=15.0d0/32.0d0/(a**5)
      c(31)=15.0d0/64.0d0/(a**6)
      c(33)=45.0d0/64.0d0/(a**6)
      c(78)=1.0d0/128.0d0/(a**7)
      c(80)=15.0d0/128.0d0/(a**7)
      c(159)=1.0d0/256.0d0/(a**8)
    case (160)
      c(9)=45.0d0/32.0d0/(a**5)
      c(32)=15.0d0/32.0d0/(a**6)
      c(34)=15.0d0/64.0d0/(a**6)
      c(79)=3.0d0/128.0d0/(a**7)
      c(81)=5.0d0/64.0d0/(a**7)
      c(160)=1.0d0/256.0d0/(a**8)
    case (161)
      c(1)=9.0d0/16.0d0/(a**4)
      c(8)=9.0d0/16.0d0/(a**5)
      c(10)=9.0d0/16.0d0/(a**5)
      c(31)=3.0d0/64.0d0/(a**6)
      c(33)=9.0d0/16.0d0/(a**6)
      c(35)=3.0d0/64.0d0/(a**6)
      c(80)=3.0d0/64.0d0/(a**7)
      c(82)=3.0d0/64.0d0/(a**7)
      c(161)=1.0d0/256.0d0/(a**8)
    case (162)
      c(9)=45.0d0/32.0d0/(a**5)
      c(32)=15.0d0/64.0d0/(a**6)
      c(34)=15.0d0/32.0d0/(a**6)
      c(81)=5.0d0/64.0d0/(a**7)
      c(83)=3.0d0/128.0d0/(a**7)
      c(162)=1.0d0/256.0d0/(a**8)
    case (163)
      c(1)=15.0d0/16.0d0/(a**4)
      c(8)=15.0d0/32.0d0/(a**5)
      c(10)=45.0d0/32.0d0/(a**5)
      c(33)=45.0d0/64.0d0/(a**6)
      c(35)=15.0d0/64.0d0/(a**6)
      c(82)=15.0d0/128.0d0/(a**7)
      c(84)=1.0d0/128.0d0/(a**7)
      c(163)=1.0d0/256.0d0/(a**8)
    case (164)
      c(9)=105.0d0/32.0d0/(a**5)
      c(34)=105.0d0/64.0d0/(a**6)
      c(83)=21.0d0/128.0d0/(a**7)
      c(164)=1.0d0/256.0d0/(a**8)
    case (165)
      c(1)=105.0d0/16.0d0/(a**4)
      c(10)=105.0d0/8.0d0/(a**5)
      c(35)=105.0d0/32.0d0/(a**6)
      c(84)=7.0d0/32.0d0/(a**7)
      c(165)=1.0d0/256.0d0/(a**8)
    case (166)
      c(2)=945.0d0/32.0d0/(a**5)
      c(11)=315.0d0/16.0d0/(a**6)
      c(36)=189.0d0/64.0d0/(a**7)
      c(85)=9.0d0/64.0d0/(a**8)
      c(166)=1.0d0/512.0d0/(a**9)
    case (167)
      c(3)=105.0d0/32.0d0/(a**5)
      c(12)=105.0d0/16.0d0/(a**6)
      c(37)=105.0d0/64.0d0/(a**7)
      c(86)=7.0d0/64.0d0/(a**8)
      c(167)=1.0d0/512.0d0/(a**9)
    case (168)
      c(4)=105.0d0/32.0d0/(a**5)
      c(13)=105.0d0/16.0d0/(a**6)
      c(38)=105.0d0/64.0d0/(a**7)
      c(87)=7.0d0/64.0d0/(a**8)
      c(168)=1.0d0/512.0d0/(a**9)
    case (169)
      c(2)=105.0d0/32.0d0/(a**5)
      c(11)=105.0d0/64.0d0/(a**6)
      c(14)=105.0d0/64.0d0/(a**6)
      c(36)=21.0d0/128.0d0/(a**7)
      c(39)=105.0d0/128.0d0/(a**7)
      c(85)=1.0d0/256.0d0/(a**8)
      c(88)=21.0d0/256.0d0/(a**8)
      c(169)=1.0d0/512.0d0/(a**9)
    case (170)
      c(15)=105.0d0/64.0d0/(a**6)
      c(40)=105.0d0/128.0d0/(a**7)
      c(89)=21.0d0/256.0d0/(a**8)
      c(170)=1.0d0/512.0d0/(a**9)
    case (171)
      c(2)=105.0d0/32.0d0/(a**5)
      c(11)=105.0d0/64.0d0/(a**6)
      c(16)=105.0d0/64.0d0/(a**6)
      c(36)=21.0d0/128.0d0/(a**7)
      c(41)=105.0d0/128.0d0/(a**7)
      c(85)=1.0d0/256.0d0/(a**8)
      c(90)=21.0d0/256.0d0/(a**8)
      c(171)=1.0d0/512.0d0/(a**9)
    case (172)
      c(3)=45.0d0/32.0d0/(a**5)
      c(12)=135.0d0/64.0d0/(a**6)
      c(17)=15.0d0/64.0d0/(a**6)
      c(37)=45.0d0/128.0d0/(a**7)
      c(42)=45.0d0/128.0d0/(a**7)
      c(86)=3.0d0/256.0d0/(a**8)
      c(91)=15.0d0/256.0d0/(a**8)
      c(172)=1.0d0/512.0d0/(a**9)
    case (173)
      c(4)=15.0d0/32.0d0/(a**5)
      c(13)=45.0d0/64.0d0/(a**6)
      c(18)=15.0d0/64.0d0/(a**6)
      c(38)=15.0d0/128.0d0/(a**7)
      c(43)=45.0d0/128.0d0/(a**7)
      c(87)=1.0d0/256.0d0/(a**8)
      c(92)=15.0d0/256.0d0/(a**8)
      c(173)=1.0d0/512.0d0/(a**9)
    case (174)
      c(3)=15.0d0/32.0d0/(a**5)
      c(12)=45.0d0/64.0d0/(a**6)
      c(19)=15.0d0/64.0d0/(a**6)
      c(37)=15.0d0/128.0d0/(a**7)
      c(44)=45.0d0/128.0d0/(a**7)
      c(86)=1.0d0/256.0d0/(a**8)
      c(93)=15.0d0/256.0d0/(a**8)
      c(174)=1.0d0/512.0d0/(a**9)
    case (175)
      c(4)=45.0d0/32.0d0/(a**5)
      c(13)=135.0d0/64.0d0/(a**6)
      c(20)=15.0d0/64.0d0/(a**6)
      c(38)=45.0d0/128.0d0/(a**7)
      c(45)=45.0d0/128.0d0/(a**7)
      c(87)=3.0d0/256.0d0/(a**8)
      c(94)=15.0d0/256.0d0/(a**8)
      c(175)=1.0d0/512.0d0/(a**9)
    case (176)
      c(2)=45.0d0/32.0d0/(a**5)
      c(11)=15.0d0/32.0d0/(a**6)
      c(14)=45.0d0/32.0d0/(a**6)
      c(36)=3.0d0/128.0d0/(a**7)
      c(39)=15.0d0/32.0d0/(a**7)
      c(46)=15.0d0/128.0d0/(a**7)
      c(88)=3.0d0/128.0d0/(a**8)
      c(95)=5.0d0/128.0d0/(a**8)
      c(176)=1.0d0/512.0d0/(a**9)
    case (177)
      c(15)=45.0d0/64.0d0/(a**6)
      c(40)=15.0d0/64.0d0/(a**7)
      c(47)=15.0d0/128.0d0/(a**7)
      c(89)=3.0d0/256.0d0/(a**8)
      c(96)=5.0d0/128.0d0/(a**8)
      c(177)=1.0d0/512.0d0/(a**9)
    case (178)
      c(2)=15.0d0/32.0d0/(a**5)
      c(11)=5.0d0/32.0d0/(a**6)
      c(14)=15.0d0/64.0d0/(a**6)
      c(16)=15.0d0/64.0d0/(a**6)
      c(36)=1.0d0/128.0d0/(a**7)
      c(39)=5.0d0/64.0d0/(a**7)
      c(41)=5.0d0/64.0d0/(a**7)
      c(48)=15.0d0/128.0d0/(a**7)
      c(88)=1.0d0/256.0d0/(a**8)
      c(90)=1.0d0/256.0d0/(a**8)
      c(97)=5.0d0/128.0d0/(a**8)
      c(178)=1.0d0/512.0d0/(a**9)
    case (179)
      c(15)=45.0d0/64.0d0/(a**6)
      c(40)=15.0d0/64.0d0/(a**7)
      c(49)=15.0d0/128.0d0/(a**7)
      c(89)=3.0d0/256.0d0/(a**8)
      c(98)=5.0d0/128.0d0/(a**8)
      c(179)=1.0d0/512.0d0/(a**9)
    case (180)
      c(2)=45.0d0/32.0d0/(a**5)
      c(11)=15.0d0/32.0d0/(a**6)
      c(16)=45.0d0/32.0d0/(a**6)
      c(36)=3.0d0/128.0d0/(a**7)
      c(41)=15.0d0/32.0d0/(a**7)
      c(50)=15.0d0/128.0d0/(a**7)
      c(90)=3.0d0/128.0d0/(a**8)
      c(99)=5.0d0/128.0d0/(a**8)
      c(180)=1.0d0/512.0d0/(a**9)
    case (181)
      c(3)=45.0d0/32.0d0/(a**5)
      c(12)=45.0d0/32.0d0/(a**6)
      c(17)=15.0d0/32.0d0/(a**6)
      c(37)=15.0d0/128.0d0/(a**7)
      c(42)=15.0d0/32.0d0/(a**7)
      c(51)=3.0d0/128.0d0/(a**7)
      c(91)=5.0d0/128.0d0/(a**8)
      c(100)=3.0d0/128.0d0/(a**8)
      c(181)=1.0d0/512.0d0/(a**9)
    case (182)
      c(4)=9.0d0/32.0d0/(a**5)
      c(13)=9.0d0/32.0d0/(a**6)
      c(18)=9.0d0/32.0d0/(a**6)
      c(38)=3.0d0/128.0d0/(a**7)
      c(43)=9.0d0/32.0d0/(a**7)
      c(52)=3.0d0/128.0d0/(a**7)
      c(92)=3.0d0/128.0d0/(a**8)
      c(101)=3.0d0/128.0d0/(a**8)
      c(182)=1.0d0/512.0d0/(a**9)
    case (183)
      c(3)=9.0d0/32.0d0/(a**5)
      c(12)=9.0d0/32.0d0/(a**6)
      c(17)=3.0d0/64.0d0/(a**6)
      c(19)=9.0d0/64.0d0/(a**6)
      c(37)=3.0d0/128.0d0/(a**7)
      c(42)=3.0d0/64.0d0/(a**7)
      c(44)=9.0d0/64.0d0/(a**7)
      c(53)=3.0d0/128.0d0/(a**7)
      c(91)=1.0d0/256.0d0/(a**8)
      c(93)=3.0d0/256.0d0/(a**8)
      c(102)=3.0d0/128.0d0/(a**8)
      c(183)=1.0d0/512.0d0/(a**9)
    case (184)
      c(4)=9.0d0/32.0d0/(a**5)
      c(13)=9.0d0/32.0d0/(a**6)
      c(18)=9.0d0/64.0d0/(a**6)
      c(20)=3.0d0/64.0d0/(a**6)
      c(38)=3.0d0/128.0d0/(a**7)
      c(43)=9.0d0/64.0d0/(a**7)
      c(45)=3.0d0/64.0d0/(a**7)
      c(54)=3.0d0/128.0d0/(a**7)
      c(92)=3.0d0/256.0d0/(a**8)
      c(94)=1.0d0/256.0d0/(a**8)
      c(103)=3.0d0/128.0d0/(a**8)
      c(184)=1.0d0/512.0d0/(a**9)
    case (185)
      c(3)=9.0d0/32.0d0/(a**5)
      c(12)=9.0d0/32.0d0/(a**6)
      c(19)=9.0d0/32.0d0/(a**6)
      c(37)=3.0d0/128.0d0/(a**7)
      c(44)=9.0d0/32.0d0/(a**7)
      c(55)=3.0d0/128.0d0/(a**7)
      c(93)=3.0d0/128.0d0/(a**8)
      c(104)=3.0d0/128.0d0/(a**8)
      c(185)=1.0d0/512.0d0/(a**9)
    case (186)
      c(4)=45.0d0/32.0d0/(a**5)
      c(13)=45.0d0/32.0d0/(a**6)
      c(20)=15.0d0/32.0d0/(a**6)
      c(38)=15.0d0/128.0d0/(a**7)
      c(45)=15.0d0/32.0d0/(a**7)
      c(56)=3.0d0/128.0d0/(a**7)
      c(94)=5.0d0/128.0d0/(a**8)
      c(105)=3.0d0/128.0d0/(a**8)
      c(186)=1.0d0/512.0d0/(a**9)
    case (187)
      c(2)=45.0d0/32.0d0/(a**5)
      c(11)=15.0d0/64.0d0/(a**6)
      c(14)=135.0d0/64.0d0/(a**6)
      c(39)=45.0d0/128.0d0/(a**7)
      c(46)=45.0d0/128.0d0/(a**7)
      c(95)=15.0d0/256.0d0/(a**8)
      c(106)=3.0d0/256.0d0/(a**8)
      c(187)=1.0d0/512.0d0/(a**9)
    case (188)
      c(15)=45.0d0/64.0d0/(a**6)
      c(40)=15.0d0/128.0d0/(a**7)
      c(47)=15.0d0/64.0d0/(a**7)
      c(96)=5.0d0/128.0d0/(a**8)
      c(107)=3.0d0/256.0d0/(a**8)
      c(188)=1.0d0/512.0d0/(a**9)
    case (189)
      c(2)=9.0d0/32.0d0/(a**5)
      c(11)=3.0d0/64.0d0/(a**6)
      c(14)=9.0d0/32.0d0/(a**6)
      c(16)=9.0d0/64.0d0/(a**6)
      c(39)=3.0d0/64.0d0/(a**7)
      c(41)=3.0d0/128.0d0/(a**7)
      c(46)=3.0d0/128.0d0/(a**7)
      c(48)=9.0d0/64.0d0/(a**7)
      c(95)=1.0d0/256.0d0/(a**8)
      c(97)=3.0d0/128.0d0/(a**8)
      c(108)=3.0d0/256.0d0/(a**8)
      c(189)=1.0d0/512.0d0/(a**9)
    case (190)
      c(15)=27.0d0/64.0d0/(a**6)
      c(40)=9.0d0/128.0d0/(a**7)
      c(47)=9.0d0/128.0d0/(a**7)
      c(49)=9.0d0/128.0d0/(a**7)
      c(96)=3.0d0/256.0d0/(a**8)
      c(98)=3.0d0/256.0d0/(a**8)
      c(109)=3.0d0/256.0d0/(a**8)
      c(190)=1.0d0/512.0d0/(a**9)
    case (191)
      c(2)=9.0d0/32.0d0/(a**5)
      c(11)=3.0d0/64.0d0/(a**6)
      c(14)=9.0d0/64.0d0/(a**6)
      c(16)=9.0d0/32.0d0/(a**6)
      c(39)=3.0d0/128.0d0/(a**7)
      c(41)=3.0d0/64.0d0/(a**7)
      c(48)=9.0d0/64.0d0/(a**7)
      c(50)=3.0d0/128.0d0/(a**7)
      c(97)=3.0d0/128.0d0/(a**8)
      c(99)=1.0d0/256.0d0/(a**8)
      c(110)=3.0d0/256.0d0/(a**8)
      c(191)=1.0d0/512.0d0/(a**9)
    case (192)
      c(15)=45.0d0/64.0d0/(a**6)
      c(40)=15.0d0/128.0d0/(a**7)
      c(49)=15.0d0/64.0d0/(a**7)
      c(98)=5.0d0/128.0d0/(a**8)
      c(111)=3.0d0/256.0d0/(a**8)
      c(192)=1.0d0/512.0d0/(a**9)
    case (193)
      c(2)=45.0d0/32.0d0/(a**5)
      c(11)=15.0d0/64.0d0/(a**6)
      c(16)=135.0d0/64.0d0/(a**6)
      c(41)=45.0d0/128.0d0/(a**7)
      c(50)=45.0d0/128.0d0/(a**7)
      c(99)=15.0d0/256.0d0/(a**8)
      c(112)=3.0d0/256.0d0/(a**8)
      c(193)=1.0d0/512.0d0/(a**9)
    case (194)
      c(3)=105.0d0/32.0d0/(a**5)
      c(12)=105.0d0/64.0d0/(a**6)
      c(17)=105.0d0/64.0d0/(a**6)
      c(42)=105.0d0/128.0d0/(a**7)
      c(51)=21.0d0/128.0d0/(a**7)
      c(100)=21.0d0/256.0d0/(a**8)
      c(113)=1.0d0/256.0d0/(a**8)
      c(194)=1.0d0/512.0d0/(a**9)
    case (195)
      c(4)=15.0d0/32.0d0/(a**5)
      c(13)=15.0d0/64.0d0/(a**6)
      c(18)=45.0d0/64.0d0/(a**6)
      c(43)=45.0d0/128.0d0/(a**7)
      c(52)=15.0d0/128.0d0/(a**7)
      c(101)=15.0d0/256.0d0/(a**8)
      c(114)=1.0d0/256.0d0/(a**8)
      c(195)=1.0d0/512.0d0/(a**9)
    case (196)
      c(3)=15.0d0/32.0d0/(a**5)
      c(12)=15.0d0/64.0d0/(a**6)
      c(17)=5.0d0/32.0d0/(a**6)
      c(19)=15.0d0/64.0d0/(a**6)
      c(42)=5.0d0/64.0d0/(a**7)
      c(44)=15.0d0/128.0d0/(a**7)
      c(51)=1.0d0/128.0d0/(a**7)
      c(53)=5.0d0/64.0d0/(a**7)
      c(100)=1.0d0/256.0d0/(a**8)
      c(102)=5.0d0/128.0d0/(a**8)
      c(115)=1.0d0/256.0d0/(a**8)
      c(196)=1.0d0/512.0d0/(a**9)
    case (197)
      c(4)=9.0d0/32.0d0/(a**5)
      c(13)=9.0d0/64.0d0/(a**6)
      c(18)=9.0d0/32.0d0/(a**6)
      c(20)=3.0d0/64.0d0/(a**6)
      c(43)=9.0d0/64.0d0/(a**7)
      c(45)=3.0d0/128.0d0/(a**7)
      c(52)=3.0d0/128.0d0/(a**7)
      c(54)=3.0d0/64.0d0/(a**7)
      c(101)=3.0d0/256.0d0/(a**8)
      c(103)=3.0d0/128.0d0/(a**8)
      c(116)=1.0d0/256.0d0/(a**8)
      c(197)=1.0d0/512.0d0/(a**9)
    case (198)
      c(3)=9.0d0/32.0d0/(a**5)
      c(12)=9.0d0/64.0d0/(a**6)
      c(17)=3.0d0/64.0d0/(a**6)
      c(19)=9.0d0/32.0d0/(a**6)
      c(42)=3.0d0/128.0d0/(a**7)
      c(44)=9.0d0/64.0d0/(a**7)
      c(53)=3.0d0/64.0d0/(a**7)
      c(55)=3.0d0/128.0d0/(a**7)
      c(102)=3.0d0/128.0d0/(a**8)
      c(104)=3.0d0/256.0d0/(a**8)
      c(117)=1.0d0/256.0d0/(a**8)
      c(198)=1.0d0/512.0d0/(a**9)
    case (199)
      c(4)=15.0d0/32.0d0/(a**5)
      c(13)=15.0d0/64.0d0/(a**6)
      c(18)=15.0d0/64.0d0/(a**6)
      c(20)=5.0d0/32.0d0/(a**6)
      c(43)=15.0d0/128.0d0/(a**7)
      c(45)=5.0d0/64.0d0/(a**7)
      c(54)=5.0d0/64.0d0/(a**7)
      c(56)=1.0d0/128.0d0/(a**7)
      c(103)=5.0d0/128.0d0/(a**8)
      c(105)=1.0d0/256.0d0/(a**8)
      c(118)=1.0d0/256.0d0/(a**8)
      c(199)=1.0d0/512.0d0/(a**9)
    case (200)
      c(3)=15.0d0/32.0d0/(a**5)
      c(12)=15.0d0/64.0d0/(a**6)
      c(19)=45.0d0/64.0d0/(a**6)
      c(44)=45.0d0/128.0d0/(a**7)
      c(55)=15.0d0/128.0d0/(a**7)
      c(104)=15.0d0/256.0d0/(a**8)
      c(119)=1.0d0/256.0d0/(a**8)
      c(200)=1.0d0/512.0d0/(a**9)
    case (201)
      c(4)=105.0d0/32.0d0/(a**5)
      c(13)=105.0d0/64.0d0/(a**6)
      c(20)=105.0d0/64.0d0/(a**6)
      c(45)=105.0d0/128.0d0/(a**7)
      c(56)=21.0d0/128.0d0/(a**7)
      c(105)=21.0d0/256.0d0/(a**8)
      c(120)=1.0d0/256.0d0/(a**8)
      c(201)=1.0d0/512.0d0/(a**9)
    case (202)
      c(2)=105.0d0/32.0d0/(a**5)
      c(14)=105.0d0/16.0d0/(a**6)
      c(46)=105.0d0/64.0d0/(a**7)
      c(106)=7.0d0/64.0d0/(a**8)
      c(202)=1.0d0/512.0d0/(a**9)
    case (203)
      c(15)=105.0d0/64.0d0/(a**6)
      c(47)=105.0d0/128.0d0/(a**7)
      c(107)=21.0d0/256.0d0/(a**8)
      c(203)=1.0d0/512.0d0/(a**9)
    case (204)
      c(2)=15.0d0/32.0d0/(a**5)
      c(14)=45.0d0/64.0d0/(a**6)
      c(16)=15.0d0/64.0d0/(a**6)
      c(46)=15.0d0/128.0d0/(a**7)
      c(48)=45.0d0/128.0d0/(a**7)
      c(106)=1.0d0/256.0d0/(a**8)
      c(108)=15.0d0/256.0d0/(a**8)
      c(204)=1.0d0/512.0d0/(a**9)
    case (205)
      c(15)=45.0d0/64.0d0/(a**6)
      c(47)=15.0d0/64.0d0/(a**7)
      c(49)=15.0d0/128.0d0/(a**7)
      c(107)=3.0d0/256.0d0/(a**8)
      c(109)=5.0d0/128.0d0/(a**8)
      c(205)=1.0d0/512.0d0/(a**9)
    case (206)
      c(2)=9.0d0/32.0d0/(a**5)
      c(14)=9.0d0/32.0d0/(a**6)
      c(16)=9.0d0/32.0d0/(a**6)
      c(46)=3.0d0/128.0d0/(a**7)
      c(48)=9.0d0/32.0d0/(a**7)
      c(50)=3.0d0/128.0d0/(a**7)
      c(108)=3.0d0/128.0d0/(a**8)
      c(110)=3.0d0/128.0d0/(a**8)
      c(206)=1.0d0/512.0d0/(a**9)
    case (207)
      c(15)=45.0d0/64.0d0/(a**6)
      c(47)=15.0d0/128.0d0/(a**7)
      c(49)=15.0d0/64.0d0/(a**7)
      c(109)=5.0d0/128.0d0/(a**8)
      c(111)=3.0d0/256.0d0/(a**8)
      c(207)=1.0d0/512.0d0/(a**9)
    case (208)
      c(2)=15.0d0/32.0d0/(a**5)
      c(14)=15.0d0/64.0d0/(a**6)
      c(16)=45.0d0/64.0d0/(a**6)
      c(48)=45.0d0/128.0d0/(a**7)
      c(50)=15.0d0/128.0d0/(a**7)
      c(110)=15.0d0/256.0d0/(a**8)
      c(112)=1.0d0/256.0d0/(a**8)
      c(208)=1.0d0/512.0d0/(a**9)
    case (209)
      c(15)=105.0d0/64.0d0/(a**6)
      c(49)=105.0d0/128.0d0/(a**7)
      c(111)=21.0d0/256.0d0/(a**8)
      c(209)=1.0d0/512.0d0/(a**9)
    case (210)
      c(2)=105.0d0/32.0d0/(a**5)
      c(16)=105.0d0/16.0d0/(a**6)
      c(50)=105.0d0/64.0d0/(a**7)
      c(112)=7.0d0/64.0d0/(a**8)
      c(210)=1.0d0/512.0d0/(a**9)
    case (211)
      c(3)=945.0d0/32.0d0/(a**5)
      c(17)=315.0d0/16.0d0/(a**6)
      c(51)=189.0d0/64.0d0/(a**7)
      c(113)=9.0d0/64.0d0/(a**8)
      c(211)=1.0d0/512.0d0/(a**9)
    case (212)
      c(4)=105.0d0/32.0d0/(a**5)
      c(18)=105.0d0/16.0d0/(a**6)
      c(52)=105.0d0/64.0d0/(a**7)
      c(114)=7.0d0/64.0d0/(a**8)
      c(212)=1.0d0/512.0d0/(a**9)
    case (213)
      c(3)=105.0d0/32.0d0/(a**5)
      c(17)=105.0d0/64.0d0/(a**6)
      c(19)=105.0d0/64.0d0/(a**6)
      c(51)=21.0d0/128.0d0/(a**7)
      c(53)=105.0d0/128.0d0/(a**7)
      c(113)=1.0d0/256.0d0/(a**8)
      c(115)=21.0d0/256.0d0/(a**8)
      c(213)=1.0d0/512.0d0/(a**9)
    case (214)
      c(4)=45.0d0/32.0d0/(a**5)
      c(18)=135.0d0/64.0d0/(a**6)
      c(20)=15.0d0/64.0d0/(a**6)
      c(52)=45.0d0/128.0d0/(a**7)
      c(54)=45.0d0/128.0d0/(a**7)
      c(114)=3.0d0/256.0d0/(a**8)
      c(116)=15.0d0/256.0d0/(a**8)
      c(214)=1.0d0/512.0d0/(a**9)
    case (215)
      c(3)=45.0d0/32.0d0/(a**5)
      c(17)=15.0d0/32.0d0/(a**6)
      c(19)=45.0d0/32.0d0/(a**6)
      c(51)=3.0d0/128.0d0/(a**7)
      c(53)=15.0d0/32.0d0/(a**7)
      c(55)=15.0d0/128.0d0/(a**7)
      c(115)=3.0d0/128.0d0/(a**8)
      c(117)=5.0d0/128.0d0/(a**8)
      c(215)=1.0d0/512.0d0/(a**9)
    case (216)
      c(4)=45.0d0/32.0d0/(a**5)
      c(18)=45.0d0/32.0d0/(a**6)
      c(20)=15.0d0/32.0d0/(a**6)
      c(52)=15.0d0/128.0d0/(a**7)
      c(54)=15.0d0/32.0d0/(a**7)
      c(56)=3.0d0/128.0d0/(a**7)
      c(116)=5.0d0/128.0d0/(a**8)
      c(118)=3.0d0/128.0d0/(a**8)
      c(216)=1.0d0/512.0d0/(a**9)
    case (217)
      c(3)=45.0d0/32.0d0/(a**5)
      c(17)=15.0d0/64.0d0/(a**6)
      c(19)=135.0d0/64.0d0/(a**6)
      c(53)=45.0d0/128.0d0/(a**7)
      c(55)=45.0d0/128.0d0/(a**7)
      c(117)=15.0d0/256.0d0/(a**8)
      c(119)=3.0d0/256.0d0/(a**8)
      c(217)=1.0d0/512.0d0/(a**9)
    case (218)
      c(4)=105.0d0/32.0d0/(a**5)
      c(18)=105.0d0/64.0d0/(a**6)
      c(20)=105.0d0/64.0d0/(a**6)
      c(54)=105.0d0/128.0d0/(a**7)
      c(56)=21.0d0/128.0d0/(a**7)
      c(118)=21.0d0/256.0d0/(a**8)
      c(120)=1.0d0/256.0d0/(a**8)
      c(218)=1.0d0/512.0d0/(a**9)
    case (219)
      c(3)=105.0d0/32.0d0/(a**5)
      c(19)=105.0d0/16.0d0/(a**6)
      c(55)=105.0d0/64.0d0/(a**7)
      c(119)=7.0d0/64.0d0/(a**8)
      c(219)=1.0d0/512.0d0/(a**9)
    case (220)
      c(4)=945.0d0/32.0d0/(a**5)
      c(20)=315.0d0/16.0d0/(a**6)
      c(56)=189.0d0/64.0d0/(a**7)
      c(120)=9.0d0/64.0d0/(a**8)
      c(220)=1.0d0/512.0d0/(a**9)
    case (221)
      c(1)=945.0d0/32.0d0/(a**5)
      c(5)=4725.0d0/64.0d0/(a**6)
      c(21)=1575.0d0/64.0d0/(a**7)
      c(57)=315.0d0/128.0d0/(a**8)
      c(121)=45.0d0/512.0d0/(a**9)
      c(221)=1.0d0/1024.0d0/(a**10)
    case (222)
      c(6)=945.0d0/64.0d0/(a**6)
      c(22)=315.0d0/32.0d0/(a**7)
      c(58)=189.0d0/128.0d0/(a**8)
      c(122)=9.0d0/128.0d0/(a**9)
      c(222)=1.0d0/1024.0d0/(a**10)
    case (223)
      c(7)=945.0d0/64.0d0/(a**6)
      c(23)=315.0d0/32.0d0/(a**7)
      c(59)=189.0d0/128.0d0/(a**8)
      c(123)=9.0d0/128.0d0/(a**9)
      c(223)=1.0d0/1024.0d0/(a**10)
    case (224)
      c(1)=105.0d0/32.0d0/(a**5)
      c(5)=105.0d0/16.0d0/(a**6)
      c(8)=105.0d0/64.0d0/(a**6)
      c(21)=105.0d0/64.0d0/(a**7)
      c(24)=105.0d0/32.0d0/(a**7)
      c(57)=7.0d0/64.0d0/(a**8)
      c(60)=105.0d0/128.0d0/(a**8)
      c(121)=1.0d0/512.0d0/(a**9)
      c(124)=7.0d0/128.0d0/(a**9)
      c(224)=1.0d0/1024.0d0/(a**10)
    case (225)
      c(9)=105.0d0/64.0d0/(a**6)
      c(25)=105.0d0/32.0d0/(a**7)
      c(61)=105.0d0/128.0d0/(a**8)
      c(125)=7.0d0/128.0d0/(a**9)
      c(225)=1.0d0/1024.0d0/(a**10)
    case (226)
      c(1)=105.0d0/32.0d0/(a**5)
      c(5)=105.0d0/16.0d0/(a**6)
      c(10)=105.0d0/64.0d0/(a**6)
      c(21)=105.0d0/64.0d0/(a**7)
      c(26)=105.0d0/32.0d0/(a**7)
      c(57)=7.0d0/64.0d0/(a**8)
      c(62)=105.0d0/128.0d0/(a**8)
      c(121)=1.0d0/512.0d0/(a**9)
      c(126)=7.0d0/128.0d0/(a**9)
      c(226)=1.0d0/1024.0d0/(a**10)
    case (227)
      c(6)=315.0d0/64.0d0/(a**6)
      c(22)=315.0d0/128.0d0/(a**7)
      c(27)=105.0d0/128.0d0/(a**7)
      c(58)=63.0d0/256.0d0/(a**8)
      c(63)=105.0d0/256.0d0/(a**8)
      c(122)=3.0d0/512.0d0/(a**9)
      c(127)=21.0d0/512.0d0/(a**9)
      c(227)=1.0d0/1024.0d0/(a**10)
    case (228)
      c(7)=105.0d0/64.0d0/(a**6)
      c(23)=105.0d0/128.0d0/(a**7)
      c(28)=105.0d0/128.0d0/(a**7)
      c(59)=21.0d0/256.0d0/(a**8)
      c(64)=105.0d0/256.0d0/(a**8)
      c(123)=1.0d0/512.0d0/(a**9)
      c(128)=21.0d0/512.0d0/(a**9)
      c(228)=1.0d0/1024.0d0/(a**10)
    case (229)
      c(6)=105.0d0/64.0d0/(a**6)
      c(22)=105.0d0/128.0d0/(a**7)
      c(29)=105.0d0/128.0d0/(a**7)
      c(58)=21.0d0/256.0d0/(a**8)
      c(65)=105.0d0/256.0d0/(a**8)
      c(122)=1.0d0/512.0d0/(a**9)
      c(129)=21.0d0/512.0d0/(a**9)
      c(229)=1.0d0/1024.0d0/(a**10)
    case (230)
      c(7)=315.0d0/64.0d0/(a**6)
      c(23)=315.0d0/128.0d0/(a**7)
      c(30)=105.0d0/128.0d0/(a**7)
      c(59)=63.0d0/256.0d0/(a**8)
      c(66)=105.0d0/256.0d0/(a**8)
      c(123)=3.0d0/512.0d0/(a**9)
      c(130)=21.0d0/512.0d0/(a**9)
      c(230)=1.0d0/1024.0d0/(a**10)
    case (231)
      c(1)=45.0d0/32.0d0/(a**5)
      c(5)=135.0d0/64.0d0/(a**6)
      c(8)=45.0d0/32.0d0/(a**6)
      c(21)=45.0d0/128.0d0/(a**7)
      c(24)=135.0d0/64.0d0/(a**7)
      c(31)=15.0d0/128.0d0/(a**7)
      c(57)=3.0d0/256.0d0/(a**8)
      c(60)=45.0d0/128.0d0/(a**8)
      c(67)=45.0d0/256.0d0/(a**8)
      c(124)=3.0d0/256.0d0/(a**9)
      c(131)=15.0d0/512.0d0/(a**9)
      c(231)=1.0d0/1024.0d0/(a**10)
    case (232)
      c(9)=45.0d0/64.0d0/(a**6)
      c(25)=135.0d0/128.0d0/(a**7)
      c(32)=15.0d0/128.0d0/(a**7)
      c(61)=45.0d0/256.0d0/(a**8)
      c(68)=45.0d0/256.0d0/(a**8)
      c(125)=3.0d0/512.0d0/(a**9)
      c(132)=15.0d0/512.0d0/(a**9)
      c(232)=1.0d0/1024.0d0/(a**10)
    case (233)
      c(1)=15.0d0/32.0d0/(a**5)
      c(5)=45.0d0/64.0d0/(a**6)
      c(8)=15.0d0/64.0d0/(a**6)
      c(10)=15.0d0/64.0d0/(a**6)
      c(21)=15.0d0/128.0d0/(a**7)
      c(24)=45.0d0/128.0d0/(a**7)
      c(26)=45.0d0/128.0d0/(a**7)
      c(33)=15.0d0/128.0d0/(a**7)
      c(57)=1.0d0/256.0d0/(a**8)
      c(60)=15.0d0/256.0d0/(a**8)
      c(62)=15.0d0/256.0d0/(a**8)
      c(69)=45.0d0/256.0d0/(a**8)
      c(124)=1.0d0/512.0d0/(a**9)
      c(126)=1.0d0/512.0d0/(a**9)
      c(133)=15.0d0/512.0d0/(a**9)
      c(233)=1.0d0/1024.0d0/(a**10)
    case (234)
      c(9)=45.0d0/64.0d0/(a**6)
      c(25)=135.0d0/128.0d0/(a**7)
      c(34)=15.0d0/128.0d0/(a**7)
      c(61)=45.0d0/256.0d0/(a**8)
      c(70)=45.0d0/256.0d0/(a**8)
      c(125)=3.0d0/512.0d0/(a**9)
      c(134)=15.0d0/512.0d0/(a**9)
      c(234)=1.0d0/1024.0d0/(a**10)
    case (235)
      c(1)=45.0d0/32.0d0/(a**5)
      c(5)=135.0d0/64.0d0/(a**6)
      c(10)=45.0d0/32.0d0/(a**6)
      c(21)=45.0d0/128.0d0/(a**7)
      c(26)=135.0d0/64.0d0/(a**7)
      c(35)=15.0d0/128.0d0/(a**7)
      c(57)=3.0d0/256.0d0/(a**8)
      c(62)=45.0d0/128.0d0/(a**8)
      c(71)=45.0d0/256.0d0/(a**8)
      c(126)=3.0d0/256.0d0/(a**9)
      c(135)=15.0d0/512.0d0/(a**9)
      c(235)=1.0d0/1024.0d0/(a**10)
    case (236)
      c(6)=225.0d0/64.0d0/(a**6)
      c(22)=75.0d0/64.0d0/(a**7)
      c(27)=75.0d0/64.0d0/(a**7)
      c(58)=15.0d0/256.0d0/(a**8)
      c(63)=25.0d0/64.0d0/(a**8)
      c(72)=15.0d0/256.0d0/(a**8)
      c(127)=5.0d0/256.0d0/(a**9)
      c(136)=5.0d0/256.0d0/(a**9)
      c(236)=1.0d0/1024.0d0/(a**10)
    case (237)
      c(7)=45.0d0/64.0d0/(a**6)
      c(23)=15.0d0/64.0d0/(a**7)
      c(28)=45.0d0/64.0d0/(a**7)
      c(59)=3.0d0/256.0d0/(a**8)
      c(64)=15.0d0/64.0d0/(a**8)
      c(73)=15.0d0/256.0d0/(a**8)
      c(128)=3.0d0/256.0d0/(a**9)
      c(137)=5.0d0/256.0d0/(a**9)
      c(237)=1.0d0/1024.0d0/(a**10)
    case (238)
      c(6)=45.0d0/64.0d0/(a**6)
      c(22)=15.0d0/64.0d0/(a**7)
      c(27)=15.0d0/128.0d0/(a**7)
      c(29)=45.0d0/128.0d0/(a**7)
      c(58)=3.0d0/256.0d0/(a**8)
      c(63)=5.0d0/128.0d0/(a**8)
      c(65)=15.0d0/128.0d0/(a**8)
      c(74)=15.0d0/256.0d0/(a**8)
      c(127)=1.0d0/512.0d0/(a**9)
      c(129)=3.0d0/512.0d0/(a**9)
      c(138)=5.0d0/256.0d0/(a**9)
      c(238)=1.0d0/1024.0d0/(a**10)
    case (239)
      c(7)=45.0d0/64.0d0/(a**6)
      c(23)=15.0d0/64.0d0/(a**7)
      c(28)=45.0d0/128.0d0/(a**7)
      c(30)=15.0d0/128.0d0/(a**7)
      c(59)=3.0d0/256.0d0/(a**8)
      c(64)=15.0d0/128.0d0/(a**8)
      c(66)=5.0d0/128.0d0/(a**8)
      c(75)=15.0d0/256.0d0/(a**8)
      c(128)=3.0d0/512.0d0/(a**9)
      c(130)=1.0d0/512.0d0/(a**9)
      c(139)=5.0d0/256.0d0/(a**9)
      c(239)=1.0d0/1024.0d0/(a**10)
    case (240)
      c(6)=45.0d0/64.0d0/(a**6)
      c(22)=15.0d0/64.0d0/(a**7)
      c(29)=45.0d0/64.0d0/(a**7)
      c(58)=3.0d0/256.0d0/(a**8)
      c(65)=15.0d0/64.0d0/(a**8)
      c(76)=15.0d0/256.0d0/(a**8)
      c(129)=3.0d0/256.0d0/(a**9)
      c(140)=5.0d0/256.0d0/(a**9)
      c(240)=1.0d0/1024.0d0/(a**10)
    case (241)
      c(7)=225.0d0/64.0d0/(a**6)
      c(23)=75.0d0/64.0d0/(a**7)
      c(30)=75.0d0/64.0d0/(a**7)
      c(59)=15.0d0/256.0d0/(a**8)
      c(66)=25.0d0/64.0d0/(a**8)
      c(77)=15.0d0/256.0d0/(a**8)
      c(130)=5.0d0/256.0d0/(a**9)
      c(141)=5.0d0/256.0d0/(a**9)
      c(241)=1.0d0/1024.0d0/(a**10)
    case (242)
      c(1)=45.0d0/32.0d0/(a**5)
      c(5)=45.0d0/32.0d0/(a**6)
      c(8)=135.0d0/64.0d0/(a**6)
      c(21)=15.0d0/128.0d0/(a**7)
      c(24)=135.0d0/64.0d0/(a**7)
      c(31)=45.0d0/128.0d0/(a**7)
      c(60)=45.0d0/256.0d0/(a**8)
      c(67)=45.0d0/128.0d0/(a**8)
      c(78)=3.0d0/256.0d0/(a**8)
      c(131)=15.0d0/512.0d0/(a**9)
      c(142)=3.0d0/256.0d0/(a**9)
      c(242)=1.0d0/1024.0d0/(a**10)
    case (243)
      c(9)=45.0d0/64.0d0/(a**6)
      c(25)=45.0d0/64.0d0/(a**7)
      c(32)=15.0d0/64.0d0/(a**7)
      c(61)=15.0d0/256.0d0/(a**8)
      c(68)=15.0d0/64.0d0/(a**8)
      c(79)=3.0d0/256.0d0/(a**8)
      c(132)=5.0d0/256.0d0/(a**9)
      c(143)=3.0d0/256.0d0/(a**9)
      c(243)=1.0d0/1024.0d0/(a**10)
    case (244)
      c(1)=9.0d0/32.0d0/(a**5)
      c(5)=9.0d0/32.0d0/(a**6)
      c(8)=9.0d0/32.0d0/(a**6)
      c(10)=9.0d0/64.0d0/(a**6)
      c(21)=3.0d0/128.0d0/(a**7)
      c(24)=9.0d0/32.0d0/(a**7)
      c(26)=9.0d0/64.0d0/(a**7)
      c(31)=3.0d0/128.0d0/(a**7)
      c(33)=9.0d0/64.0d0/(a**7)
      c(60)=3.0d0/128.0d0/(a**8)
      c(62)=3.0d0/256.0d0/(a**8)
      c(67)=3.0d0/128.0d0/(a**8)
      c(69)=9.0d0/64.0d0/(a**8)
      c(80)=3.0d0/256.0d0/(a**8)
      c(131)=1.0d0/512.0d0/(a**9)
      c(133)=3.0d0/256.0d0/(a**9)
      c(144)=3.0d0/256.0d0/(a**9)
      c(244)=1.0d0/1024.0d0/(a**10)
    case (245)
      c(9)=27.0d0/64.0d0/(a**6)
      c(25)=27.0d0/64.0d0/(a**7)
      c(32)=9.0d0/128.0d0/(a**7)
      c(34)=9.0d0/128.0d0/(a**7)
      c(61)=9.0d0/256.0d0/(a**8)
      c(68)=9.0d0/128.0d0/(a**8)
      c(70)=9.0d0/128.0d0/(a**8)
      c(81)=3.0d0/256.0d0/(a**8)
      c(132)=3.0d0/512.0d0/(a**9)
      c(134)=3.0d0/512.0d0/(a**9)
      c(145)=3.0d0/256.0d0/(a**9)
      c(245)=1.0d0/1024.0d0/(a**10)
    case (246)
      c(1)=9.0d0/32.0d0/(a**5)
      c(5)=9.0d0/32.0d0/(a**6)
      c(8)=9.0d0/64.0d0/(a**6)
      c(10)=9.0d0/32.0d0/(a**6)
      c(21)=3.0d0/128.0d0/(a**7)
      c(24)=9.0d0/64.0d0/(a**7)
      c(26)=9.0d0/32.0d0/(a**7)
      c(33)=9.0d0/64.0d0/(a**7)
      c(35)=3.0d0/128.0d0/(a**7)
      c(60)=3.0d0/256.0d0/(a**8)
      c(62)=3.0d0/128.0d0/(a**8)
      c(69)=9.0d0/64.0d0/(a**8)
      c(71)=3.0d0/128.0d0/(a**8)
      c(82)=3.0d0/256.0d0/(a**8)
      c(133)=3.0d0/256.0d0/(a**9)
      c(135)=1.0d0/512.0d0/(a**9)
      c(146)=3.0d0/256.0d0/(a**9)
      c(246)=1.0d0/1024.0d0/(a**10)
    case (247)
      c(9)=45.0d0/64.0d0/(a**6)
      c(25)=45.0d0/64.0d0/(a**7)
      c(34)=15.0d0/64.0d0/(a**7)
      c(61)=15.0d0/256.0d0/(a**8)
      c(70)=15.0d0/64.0d0/(a**8)
      c(83)=3.0d0/256.0d0/(a**8)
      c(134)=5.0d0/256.0d0/(a**9)
      c(147)=3.0d0/256.0d0/(a**9)
      c(247)=1.0d0/1024.0d0/(a**10)
    case (248)
      c(1)=45.0d0/32.0d0/(a**5)
      c(5)=45.0d0/32.0d0/(a**6)
      c(10)=135.0d0/64.0d0/(a**6)
      c(21)=15.0d0/128.0d0/(a**7)
      c(26)=135.0d0/64.0d0/(a**7)
      c(35)=45.0d0/128.0d0/(a**7)
      c(62)=45.0d0/256.0d0/(a**8)
      c(71)=45.0d0/128.0d0/(a**8)
      c(84)=3.0d0/256.0d0/(a**8)
      c(135)=15.0d0/512.0d0/(a**9)
      c(148)=3.0d0/256.0d0/(a**9)
      c(248)=1.0d0/1024.0d0/(a**10)
    case (249)
      c(6)=315.0d0/64.0d0/(a**6)
      c(22)=105.0d0/128.0d0/(a**7)
      c(27)=315.0d0/128.0d0/(a**7)
      c(63)=105.0d0/256.0d0/(a**8)
      c(72)=63.0d0/256.0d0/(a**8)
      c(136)=21.0d0/512.0d0/(a**9)
      c(149)=3.0d0/512.0d0/(a**9)
      c(249)=1.0d0/1024.0d0/(a**10)
    case (250)
      c(7)=45.0d0/64.0d0/(a**6)
      c(23)=15.0d0/128.0d0/(a**7)
      c(28)=135.0d0/128.0d0/(a**7)
      c(64)=45.0d0/256.0d0/(a**8)
      c(73)=45.0d0/256.0d0/(a**8)
      c(137)=15.0d0/512.0d0/(a**9)
      c(150)=3.0d0/512.0d0/(a**9)
      c(250)=1.0d0/1024.0d0/(a**10)
    case (251)
      c(6)=45.0d0/64.0d0/(a**6)
      c(22)=15.0d0/128.0d0/(a**7)
      c(27)=15.0d0/64.0d0/(a**7)
      c(29)=45.0d0/128.0d0/(a**7)
      c(63)=5.0d0/128.0d0/(a**8)
      c(65)=15.0d0/256.0d0/(a**8)
      c(72)=3.0d0/256.0d0/(a**8)
      c(74)=15.0d0/128.0d0/(a**8)
      c(136)=1.0d0/512.0d0/(a**9)
      c(138)=5.0d0/256.0d0/(a**9)
      c(151)=3.0d0/512.0d0/(a**9)
      c(251)=1.0d0/1024.0d0/(a**10)
    case (252)
      c(7)=27.0d0/64.0d0/(a**6)
      c(23)=9.0d0/128.0d0/(a**7)
      c(28)=27.0d0/64.0d0/(a**7)
      c(30)=9.0d0/128.0d0/(a**7)
      c(64)=9.0d0/128.0d0/(a**8)
      c(66)=3.0d0/256.0d0/(a**8)
      c(73)=9.0d0/256.0d0/(a**8)
      c(75)=9.0d0/128.0d0/(a**8)
      c(137)=3.0d0/512.0d0/(a**9)
      c(139)=3.0d0/256.0d0/(a**9)
      c(152)=3.0d0/512.0d0/(a**9)
      c(252)=1.0d0/1024.0d0/(a**10)
    case (253)
      c(6)=27.0d0/64.0d0/(a**6)
      c(22)=9.0d0/128.0d0/(a**7)
      c(27)=9.0d0/128.0d0/(a**7)
      c(29)=27.0d0/64.0d0/(a**7)
      c(63)=3.0d0/256.0d0/(a**8)
      c(65)=9.0d0/128.0d0/(a**8)
      c(74)=9.0d0/128.0d0/(a**8)
      c(76)=9.0d0/256.0d0/(a**8)
      c(138)=3.0d0/256.0d0/(a**9)
      c(140)=3.0d0/512.0d0/(a**9)
      c(153)=3.0d0/512.0d0/(a**9)
      c(253)=1.0d0/1024.0d0/(a**10)
    case (254)
      c(7)=45.0d0/64.0d0/(a**6)
      c(23)=15.0d0/128.0d0/(a**7)
      c(28)=45.0d0/128.0d0/(a**7)
      c(30)=15.0d0/64.0d0/(a**7)
      c(64)=15.0d0/256.0d0/(a**8)
      c(66)=5.0d0/128.0d0/(a**8)
      c(75)=15.0d0/128.0d0/(a**8)
      c(77)=3.0d0/256.0d0/(a**8)
      c(139)=5.0d0/256.0d0/(a**9)
      c(141)=1.0d0/512.0d0/(a**9)
      c(154)=3.0d0/512.0d0/(a**9)
      c(254)=1.0d0/1024.0d0/(a**10)
    case (255)
      c(6)=45.0d0/64.0d0/(a**6)
      c(22)=15.0d0/128.0d0/(a**7)
      c(29)=135.0d0/128.0d0/(a**7)
      c(65)=45.0d0/256.0d0/(a**8)
      c(76)=45.0d0/256.0d0/(a**8)
      c(140)=15.0d0/512.0d0/(a**9)
      c(155)=3.0d0/512.0d0/(a**9)
      c(255)=1.0d0/1024.0d0/(a**10)
    case (256)
      c(7)=315.0d0/64.0d0/(a**6)
      c(23)=105.0d0/128.0d0/(a**7)
      c(30)=315.0d0/128.0d0/(a**7)
      c(66)=105.0d0/256.0d0/(a**8)
      c(77)=63.0d0/256.0d0/(a**8)
      c(141)=21.0d0/512.0d0/(a**9)
      c(156)=3.0d0/512.0d0/(a**9)
      c(256)=1.0d0/1024.0d0/(a**10)
    case (257)
      c(1)=105.0d0/32.0d0/(a**5)
      c(5)=105.0d0/64.0d0/(a**6)
      c(8)=105.0d0/16.0d0/(a**6)
      c(24)=105.0d0/32.0d0/(a**7)
      c(31)=105.0d0/64.0d0/(a**7)
      c(67)=105.0d0/128.0d0/(a**8)
      c(78)=7.0d0/64.0d0/(a**8)
      c(142)=7.0d0/128.0d0/(a**9)
      c(157)=1.0d0/512.0d0/(a**9)
      c(257)=1.0d0/1024.0d0/(a**10)
    case (258)
      c(9)=105.0d0/64.0d0/(a**6)
      c(25)=105.0d0/128.0d0/(a**7)
      c(32)=105.0d0/128.0d0/(a**7)
      c(68)=105.0d0/256.0d0/(a**8)
      c(79)=21.0d0/256.0d0/(a**8)
      c(143)=21.0d0/512.0d0/(a**9)
      c(158)=1.0d0/512.0d0/(a**9)
      c(258)=1.0d0/1024.0d0/(a**10)
    case (259)
      c(1)=15.0d0/32.0d0/(a**5)
      c(5)=15.0d0/64.0d0/(a**6)
      c(8)=45.0d0/64.0d0/(a**6)
      c(10)=15.0d0/64.0d0/(a**6)
      c(24)=45.0d0/128.0d0/(a**7)
      c(26)=15.0d0/128.0d0/(a**7)
      c(31)=15.0d0/128.0d0/(a**7)
      c(33)=45.0d0/128.0d0/(a**7)
      c(67)=15.0d0/256.0d0/(a**8)
      c(69)=45.0d0/256.0d0/(a**8)
      c(78)=1.0d0/256.0d0/(a**8)
      c(80)=15.0d0/256.0d0/(a**8)
      c(142)=1.0d0/512.0d0/(a**9)
      c(144)=15.0d0/512.0d0/(a**9)
      c(159)=1.0d0/512.0d0/(a**9)
      c(259)=1.0d0/1024.0d0/(a**10)
    case (260)
      c(9)=45.0d0/64.0d0/(a**6)
      c(25)=45.0d0/128.0d0/(a**7)
      c(32)=15.0d0/64.0d0/(a**7)
      c(34)=15.0d0/128.0d0/(a**7)
      c(68)=15.0d0/128.0d0/(a**8)
      c(70)=15.0d0/256.0d0/(a**8)
      c(79)=3.0d0/256.0d0/(a**8)
      c(81)=5.0d0/128.0d0/(a**8)
      c(143)=3.0d0/512.0d0/(a**9)
      c(145)=5.0d0/256.0d0/(a**9)
      c(160)=1.0d0/512.0d0/(a**9)
      c(260)=1.0d0/1024.0d0/(a**10)
    case (261)
      c(1)=9.0d0/32.0d0/(a**5)
      c(5)=9.0d0/64.0d0/(a**6)
      c(8)=9.0d0/32.0d0/(a**6)
      c(10)=9.0d0/32.0d0/(a**6)
      c(24)=9.0d0/64.0d0/(a**7)
      c(26)=9.0d0/64.0d0/(a**7)
      c(31)=3.0d0/128.0d0/(a**7)
      c(33)=9.0d0/32.0d0/(a**7)
      c(35)=3.0d0/128.0d0/(a**7)
      c(67)=3.0d0/256.0d0/(a**8)
      c(69)=9.0d0/64.0d0/(a**8)
      c(71)=3.0d0/256.0d0/(a**8)
      c(80)=3.0d0/128.0d0/(a**8)
      c(82)=3.0d0/128.0d0/(a**8)
      c(144)=3.0d0/256.0d0/(a**9)
      c(146)=3.0d0/256.0d0/(a**9)
      c(161)=1.0d0/512.0d0/(a**9)
      c(261)=1.0d0/1024.0d0/(a**10)
    case (262)
      c(9)=45.0d0/64.0d0/(a**6)
      c(25)=45.0d0/128.0d0/(a**7)
      c(32)=15.0d0/128.0d0/(a**7)
      c(34)=15.0d0/64.0d0/(a**7)
      c(68)=15.0d0/256.0d0/(a**8)
      c(70)=15.0d0/128.0d0/(a**8)
      c(81)=5.0d0/128.0d0/(a**8)
      c(83)=3.0d0/256.0d0/(a**8)
      c(145)=5.0d0/256.0d0/(a**9)
      c(147)=3.0d0/512.0d0/(a**9)
      c(162)=1.0d0/512.0d0/(a**9)
      c(262)=1.0d0/1024.0d0/(a**10)
    case (263)
      c(1)=15.0d0/32.0d0/(a**5)
      c(5)=15.0d0/64.0d0/(a**6)
      c(8)=15.0d0/64.0d0/(a**6)
      c(10)=45.0d0/64.0d0/(a**6)
      c(24)=15.0d0/128.0d0/(a**7)
      c(26)=45.0d0/128.0d0/(a**7)
      c(33)=45.0d0/128.0d0/(a**7)
      c(35)=15.0d0/128.0d0/(a**7)
      c(69)=45.0d0/256.0d0/(a**8)
      c(71)=15.0d0/256.0d0/(a**8)
      c(82)=15.0d0/256.0d0/(a**8)
      c(84)=1.0d0/256.0d0/(a**8)
      c(146)=15.0d0/512.0d0/(a**9)
      c(148)=1.0d0/512.0d0/(a**9)
      c(163)=1.0d0/512.0d0/(a**9)
      c(263)=1.0d0/1024.0d0/(a**10)
    case (264)
      c(9)=105.0d0/64.0d0/(a**6)
      c(25)=105.0d0/128.0d0/(a**7)
      c(34)=105.0d0/128.0d0/(a**7)
      c(70)=105.0d0/256.0d0/(a**8)
      c(83)=21.0d0/256.0d0/(a**8)
      c(147)=21.0d0/512.0d0/(a**9)
      c(164)=1.0d0/512.0d0/(a**9)
      c(264)=1.0d0/1024.0d0/(a**10)
    case (265)
      c(1)=105.0d0/32.0d0/(a**5)
      c(5)=105.0d0/64.0d0/(a**6)
      c(10)=105.0d0/16.0d0/(a**6)
      c(26)=105.0d0/32.0d0/(a**7)
      c(35)=105.0d0/64.0d0/(a**7)
      c(71)=105.0d0/128.0d0/(a**8)
      c(84)=7.0d0/64.0d0/(a**8)
      c(148)=7.0d0/128.0d0/(a**9)
      c(165)=1.0d0/512.0d0/(a**9)
      c(265)=1.0d0/1024.0d0/(a**10)
    case (266)
      c(6)=945.0d0/64.0d0/(a**6)
      c(27)=315.0d0/32.0d0/(a**7)
      c(72)=189.0d0/128.0d0/(a**8)
      c(149)=9.0d0/128.0d0/(a**9)
      c(266)=1.0d0/1024.0d0/(a**10)
    case (267)
      c(7)=105.0d0/64.0d0/(a**6)
      c(28)=105.0d0/32.0d0/(a**7)
      c(73)=105.0d0/128.0d0/(a**8)
      c(150)=7.0d0/128.0d0/(a**9)
      c(267)=1.0d0/1024.0d0/(a**10)
    case (268)
      c(6)=105.0d0/64.0d0/(a**6)
      c(27)=105.0d0/128.0d0/(a**7)
      c(29)=105.0d0/128.0d0/(a**7)
      c(72)=21.0d0/256.0d0/(a**8)
      c(74)=105.0d0/256.0d0/(a**8)
      c(149)=1.0d0/512.0d0/(a**9)
      c(151)=21.0d0/512.0d0/(a**9)
      c(268)=1.0d0/1024.0d0/(a**10)
    case (269)
      c(7)=45.0d0/64.0d0/(a**6)
      c(28)=135.0d0/128.0d0/(a**7)
      c(30)=15.0d0/128.0d0/(a**7)
      c(73)=45.0d0/256.0d0/(a**8)
      c(75)=45.0d0/256.0d0/(a**8)
      c(150)=3.0d0/512.0d0/(a**9)
      c(152)=15.0d0/512.0d0/(a**9)
      c(269)=1.0d0/1024.0d0/(a**10)
    case (270)
      c(6)=45.0d0/64.0d0/(a**6)
      c(27)=15.0d0/64.0d0/(a**7)
      c(29)=45.0d0/64.0d0/(a**7)
      c(72)=3.0d0/256.0d0/(a**8)
      c(74)=15.0d0/64.0d0/(a**8)
      c(76)=15.0d0/256.0d0/(a**8)
      c(151)=3.0d0/256.0d0/(a**9)
      c(153)=5.0d0/256.0d0/(a**9)
      c(270)=1.0d0/1024.0d0/(a**10)
    case (271)
      c(7)=45.0d0/64.0d0/(a**6)
      c(28)=45.0d0/64.0d0/(a**7)
      c(30)=15.0d0/64.0d0/(a**7)
      c(73)=15.0d0/256.0d0/(a**8)
      c(75)=15.0d0/64.0d0/(a**8)
      c(77)=3.0d0/256.0d0/(a**8)
      c(152)=5.0d0/256.0d0/(a**9)
      c(154)=3.0d0/256.0d0/(a**9)
      c(271)=1.0d0/1024.0d0/(a**10)
    case (272)
      c(6)=45.0d0/64.0d0/(a**6)
      c(27)=15.0d0/128.0d0/(a**7)
      c(29)=135.0d0/128.0d0/(a**7)
      c(74)=45.0d0/256.0d0/(a**8)
      c(76)=45.0d0/256.0d0/(a**8)
      c(153)=15.0d0/512.0d0/(a**9)
      c(155)=3.0d0/512.0d0/(a**9)
      c(272)=1.0d0/1024.0d0/(a**10)
    case (273)
      c(7)=105.0d0/64.0d0/(a**6)
      c(28)=105.0d0/128.0d0/(a**7)
      c(30)=105.0d0/128.0d0/(a**7)
      c(75)=105.0d0/256.0d0/(a**8)
      c(77)=21.0d0/256.0d0/(a**8)
      c(154)=21.0d0/512.0d0/(a**9)
      c(156)=1.0d0/512.0d0/(a**9)
      c(273)=1.0d0/1024.0d0/(a**10)
    case (274)
      c(6)=105.0d0/64.0d0/(a**6)
      c(29)=105.0d0/32.0d0/(a**7)
      c(76)=105.0d0/128.0d0/(a**8)
      c(155)=7.0d0/128.0d0/(a**9)
      c(274)=1.0d0/1024.0d0/(a**10)
    case (275)
      c(7)=945.0d0/64.0d0/(a**6)
      c(30)=315.0d0/32.0d0/(a**7)
      c(77)=189.0d0/128.0d0/(a**8)
      c(156)=9.0d0/128.0d0/(a**9)
      c(275)=1.0d0/1024.0d0/(a**10)
    case (276)
      c(1)=945.0d0/32.0d0/(a**5)
      c(8)=4725.0d0/64.0d0/(a**6)
      c(31)=1575.0d0/64.0d0/(a**7)
      c(78)=315.0d0/128.0d0/(a**8)
      c(157)=45.0d0/512.0d0/(a**9)
      c(276)=1.0d0/1024.0d0/(a**10)
    case (277)
      c(9)=945.0d0/64.0d0/(a**6)
      c(32)=315.0d0/32.0d0/(a**7)
      c(79)=189.0d0/128.0d0/(a**8)
      c(158)=9.0d0/128.0d0/(a**9)
      c(277)=1.0d0/1024.0d0/(a**10)
    case (278)
      c(1)=105.0d0/32.0d0/(a**5)
      c(8)=105.0d0/16.0d0/(a**6)
      c(10)=105.0d0/64.0d0/(a**6)
      c(31)=105.0d0/64.0d0/(a**7)
      c(33)=105.0d0/32.0d0/(a**7)
      c(78)=7.0d0/64.0d0/(a**8)
      c(80)=105.0d0/128.0d0/(a**8)
      c(157)=1.0d0/512.0d0/(a**9)
      c(159)=7.0d0/128.0d0/(a**9)
      c(278)=1.0d0/1024.0d0/(a**10)
    case (279)
      c(9)=315.0d0/64.0d0/(a**6)
      c(32)=315.0d0/128.0d0/(a**7)
      c(34)=105.0d0/128.0d0/(a**7)
      c(79)=63.0d0/256.0d0/(a**8)
      c(81)=105.0d0/256.0d0/(a**8)
      c(158)=3.0d0/512.0d0/(a**9)
      c(160)=21.0d0/512.0d0/(a**9)
      c(279)=1.0d0/1024.0d0/(a**10)
    case (280)
      c(1)=45.0d0/32.0d0/(a**5)
      c(8)=135.0d0/64.0d0/(a**6)
      c(10)=45.0d0/32.0d0/(a**6)
      c(31)=45.0d0/128.0d0/(a**7)
      c(33)=135.0d0/64.0d0/(a**7)
      c(35)=15.0d0/128.0d0/(a**7)
      c(78)=3.0d0/256.0d0/(a**8)
      c(80)=45.0d0/128.0d0/(a**8)
      c(82)=45.0d0/256.0d0/(a**8)
      c(159)=3.0d0/256.0d0/(a**9)
      c(161)=15.0d0/512.0d0/(a**9)
      c(280)=1.0d0/1024.0d0/(a**10)
    case (281)
      c(9)=225.0d0/64.0d0/(a**6)
      c(32)=75.0d0/64.0d0/(a**7)
      c(34)=75.0d0/64.0d0/(a**7)
      c(79)=15.0d0/256.0d0/(a**8)
      c(81)=25.0d0/64.0d0/(a**8)
      c(83)=15.0d0/256.0d0/(a**8)
      c(160)=5.0d0/256.0d0/(a**9)
      c(162)=5.0d0/256.0d0/(a**9)
      c(281)=1.0d0/1024.0d0/(a**10)
    case (282)
      c(1)=45.0d0/32.0d0/(a**5)
      c(8)=45.0d0/32.0d0/(a**6)
      c(10)=135.0d0/64.0d0/(a**6)
      c(31)=15.0d0/128.0d0/(a**7)
      c(33)=135.0d0/64.0d0/(a**7)
      c(35)=45.0d0/128.0d0/(a**7)
      c(80)=45.0d0/256.0d0/(a**8)
      c(82)=45.0d0/128.0d0/(a**8)
      c(84)=3.0d0/256.0d0/(a**8)
      c(161)=15.0d0/512.0d0/(a**9)
      c(163)=3.0d0/256.0d0/(a**9)
      c(282)=1.0d0/1024.0d0/(a**10)
    case (283)
      c(9)=315.0d0/64.0d0/(a**6)
      c(32)=105.0d0/128.0d0/(a**7)
      c(34)=315.0d0/128.0d0/(a**7)
      c(81)=105.0d0/256.0d0/(a**8)
      c(83)=63.0d0/256.0d0/(a**8)
      c(162)=21.0d0/512.0d0/(a**9)
      c(164)=3.0d0/512.0d0/(a**9)
      c(283)=1.0d0/1024.0d0/(a**10)
    case (284)
      c(1)=105.0d0/32.0d0/(a**5)
      c(8)=105.0d0/64.0d0/(a**6)
      c(10)=105.0d0/16.0d0/(a**6)
      c(33)=105.0d0/32.0d0/(a**7)
      c(35)=105.0d0/64.0d0/(a**7)
      c(82)=105.0d0/128.0d0/(a**8)
      c(84)=7.0d0/64.0d0/(a**8)
      c(163)=7.0d0/128.0d0/(a**9)
      c(165)=1.0d0/512.0d0/(a**9)
      c(284)=1.0d0/1024.0d0/(a**10)
    case (285)
      c(9)=945.0d0/64.0d0/(a**6)
      c(34)=315.0d0/32.0d0/(a**7)
      c(83)=189.0d0/128.0d0/(a**8)
      c(164)=9.0d0/128.0d0/(a**9)
      c(285)=1.0d0/1024.0d0/(a**10)
    case (286)
      c(1)=945.0d0/32.0d0/(a**5)
      c(10)=4725.0d0/64.0d0/(a**6)
      c(35)=1575.0d0/64.0d0/(a**7)
      c(84)=315.0d0/128.0d0/(a**8)
      c(165)=45.0d0/512.0d0/(a**9)
      c(286)=1.0d0/1024.0d0/(a**10)
    case (287)
      c(2)=10395.0d0/64.0d0/(a**6)
      c(11)=17325.0d0/128.0d0/(a**7)
      c(36)=3465.0d0/128.0d0/(a**8)
      c(85)=495.0d0/256.0d0/(a**9)
      c(166)=55.0d0/1024.0d0/(a**10)
      c(287)=1.0d0/2048.0d0/(a**11)
    case (288)
      c(3)=945.0d0/64.0d0/(a**6)
      c(12)=4725.0d0/128.0d0/(a**7)
      c(37)=1575.0d0/128.0d0/(a**8)
      c(86)=315.0d0/256.0d0/(a**9)
      c(167)=45.0d0/1024.0d0/(a**10)
      c(288)=1.0d0/2048.0d0/(a**11)
    case (289)
      c(4)=945.0d0/64.0d0/(a**6)
      c(13)=4725.0d0/128.0d0/(a**7)
      c(38)=1575.0d0/128.0d0/(a**8)
      c(87)=315.0d0/256.0d0/(a**9)
      c(168)=45.0d0/1024.0d0/(a**10)
      c(289)=1.0d0/2048.0d0/(a**11)
    case (290)
      c(2)=945.0d0/64.0d0/(a**6)
      c(11)=315.0d0/32.0d0/(a**7)
      c(14)=945.0d0/128.0d0/(a**7)
      c(36)=189.0d0/128.0d0/(a**8)
      c(39)=315.0d0/64.0d0/(a**8)
      c(85)=9.0d0/128.0d0/(a**9)
      c(88)=189.0d0/256.0d0/(a**9)
      c(166)=1.0d0/1024.0d0/(a**10)
      c(169)=9.0d0/256.0d0/(a**10)
      c(290)=1.0d0/2048.0d0/(a**11)
    case (291)
      c(15)=945.0d0/128.0d0/(a**7)
      c(40)=315.0d0/64.0d0/(a**8)
      c(89)=189.0d0/256.0d0/(a**9)
      c(170)=9.0d0/256.0d0/(a**10)
      c(291)=1.0d0/2048.0d0/(a**11)
    case (292)
      c(2)=945.0d0/64.0d0/(a**6)
      c(11)=315.0d0/32.0d0/(a**7)
      c(16)=945.0d0/128.0d0/(a**7)
      c(36)=189.0d0/128.0d0/(a**8)
      c(41)=315.0d0/64.0d0/(a**8)
      c(85)=9.0d0/128.0d0/(a**9)
      c(90)=189.0d0/256.0d0/(a**9)
      c(166)=1.0d0/1024.0d0/(a**10)
      c(171)=9.0d0/256.0d0/(a**10)
      c(292)=1.0d0/2048.0d0/(a**11)
    case (293)
      c(3)=315.0d0/64.0d0/(a**6)
      c(12)=315.0d0/32.0d0/(a**7)
      c(17)=105.0d0/128.0d0/(a**7)
      c(37)=315.0d0/128.0d0/(a**8)
      c(42)=105.0d0/64.0d0/(a**8)
      c(86)=21.0d0/128.0d0/(a**9)
      c(91)=105.0d0/256.0d0/(a**9)
      c(167)=3.0d0/1024.0d0/(a**10)
      c(172)=7.0d0/256.0d0/(a**10)
      c(293)=1.0d0/2048.0d0/(a**11)
    case (294)
      c(4)=105.0d0/64.0d0/(a**6)
      c(13)=105.0d0/32.0d0/(a**7)
      c(18)=105.0d0/128.0d0/(a**7)
      c(38)=105.0d0/128.0d0/(a**8)
      c(43)=105.0d0/64.0d0/(a**8)
      c(87)=7.0d0/128.0d0/(a**9)
      c(92)=105.0d0/256.0d0/(a**9)
      c(168)=1.0d0/1024.0d0/(a**10)
      c(173)=7.0d0/256.0d0/(a**10)
      c(294)=1.0d0/2048.0d0/(a**11)
    case (295)
      c(3)=105.0d0/64.0d0/(a**6)
      c(12)=105.0d0/32.0d0/(a**7)
      c(19)=105.0d0/128.0d0/(a**7)
      c(37)=105.0d0/128.0d0/(a**8)
      c(44)=105.0d0/64.0d0/(a**8)
      c(86)=7.0d0/128.0d0/(a**9)
      c(93)=105.0d0/256.0d0/(a**9)
      c(167)=1.0d0/1024.0d0/(a**10)
      c(174)=7.0d0/256.0d0/(a**10)
      c(295)=1.0d0/2048.0d0/(a**11)
    case (296)
      c(4)=315.0d0/64.0d0/(a**6)
      c(13)=315.0d0/32.0d0/(a**7)
      c(20)=105.0d0/128.0d0/(a**7)
      c(38)=315.0d0/128.0d0/(a**8)
      c(45)=105.0d0/64.0d0/(a**8)
      c(87)=21.0d0/128.0d0/(a**9)
      c(94)=105.0d0/256.0d0/(a**9)
      c(168)=3.0d0/1024.0d0/(a**10)
      c(175)=7.0d0/256.0d0/(a**10)
      c(296)=1.0d0/2048.0d0/(a**11)
    case (297)
      c(2)=315.0d0/64.0d0/(a**6)
      c(11)=315.0d0/128.0d0/(a**7)
      c(14)=315.0d0/64.0d0/(a**7)
      c(36)=63.0d0/256.0d0/(a**8)
      c(39)=315.0d0/128.0d0/(a**8)
      c(46)=105.0d0/256.0d0/(a**8)
      c(85)=3.0d0/512.0d0/(a**9)
      c(88)=63.0d0/256.0d0/(a**9)
      c(95)=105.0d0/512.0d0/(a**9)
      c(169)=3.0d0/512.0d0/(a**10)
      c(176)=21.0d0/1024.0d0/(a**10)
      c(297)=1.0d0/2048.0d0/(a**11)
    case (298)
      c(15)=315.0d0/128.0d0/(a**7)
      c(40)=315.0d0/256.0d0/(a**8)
      c(47)=105.0d0/256.0d0/(a**8)
      c(89)=63.0d0/512.0d0/(a**9)
      c(96)=105.0d0/512.0d0/(a**9)
      c(170)=3.0d0/1024.0d0/(a**10)
      c(177)=21.0d0/1024.0d0/(a**10)
      c(298)=1.0d0/2048.0d0/(a**11)
    case (299)
      c(2)=105.0d0/64.0d0/(a**6)
      c(11)=105.0d0/128.0d0/(a**7)
      c(14)=105.0d0/128.0d0/(a**7)
      c(16)=105.0d0/128.0d0/(a**7)
      c(36)=21.0d0/256.0d0/(a**8)
      c(39)=105.0d0/256.0d0/(a**8)
      c(41)=105.0d0/256.0d0/(a**8)
      c(48)=105.0d0/256.0d0/(a**8)
      c(85)=1.0d0/512.0d0/(a**9)
      c(88)=21.0d0/512.0d0/(a**9)
      c(90)=21.0d0/512.0d0/(a**9)
      c(97)=105.0d0/512.0d0/(a**9)
      c(169)=1.0d0/1024.0d0/(a**10)
      c(171)=1.0d0/1024.0d0/(a**10)
      c(178)=21.0d0/1024.0d0/(a**10)
      c(299)=1.0d0/2048.0d0/(a**11)
    case (300)
      c(15)=315.0d0/128.0d0/(a**7)
      c(40)=315.0d0/256.0d0/(a**8)
      c(49)=105.0d0/256.0d0/(a**8)
      c(89)=63.0d0/512.0d0/(a**9)
      c(98)=105.0d0/512.0d0/(a**9)
      c(170)=3.0d0/1024.0d0/(a**10)
      c(179)=21.0d0/1024.0d0/(a**10)
      c(300)=1.0d0/2048.0d0/(a**11)
    case (301)
      c(2)=315.0d0/64.0d0/(a**6)
      c(11)=315.0d0/128.0d0/(a**7)
      c(16)=315.0d0/64.0d0/(a**7)
      c(36)=63.0d0/256.0d0/(a**8)
      c(41)=315.0d0/128.0d0/(a**8)
      c(50)=105.0d0/256.0d0/(a**8)
      c(85)=3.0d0/512.0d0/(a**9)
      c(90)=63.0d0/256.0d0/(a**9)
      c(99)=105.0d0/512.0d0/(a**9)
      c(171)=3.0d0/512.0d0/(a**10)
      c(180)=21.0d0/1024.0d0/(a**10)
      c(301)=1.0d0/2048.0d0/(a**11)
    case (302)
      c(3)=225.0d0/64.0d0/(a**6)
      c(12)=675.0d0/128.0d0/(a**7)
      c(17)=75.0d0/64.0d0/(a**7)
      c(37)=225.0d0/256.0d0/(a**8)
      c(42)=225.0d0/128.0d0/(a**8)
      c(51)=15.0d0/256.0d0/(a**8)
      c(86)=15.0d0/512.0d0/(a**9)
      c(91)=75.0d0/256.0d0/(a**9)
      c(100)=45.0d0/512.0d0/(a**9)
      c(172)=5.0d0/512.0d0/(a**10)
      c(181)=15.0d0/1024.0d0/(a**10)
      c(302)=1.0d0/2048.0d0/(a**11)
    case (303)
      c(4)=45.0d0/64.0d0/(a**6)
      c(13)=135.0d0/128.0d0/(a**7)
      c(18)=45.0d0/64.0d0/(a**7)
      c(38)=45.0d0/256.0d0/(a**8)
      c(43)=135.0d0/128.0d0/(a**8)
      c(52)=15.0d0/256.0d0/(a**8)
      c(87)=3.0d0/512.0d0/(a**9)
      c(92)=45.0d0/256.0d0/(a**9)
      c(101)=45.0d0/512.0d0/(a**9)
      c(173)=3.0d0/512.0d0/(a**10)
      c(182)=15.0d0/1024.0d0/(a**10)
      c(303)=1.0d0/2048.0d0/(a**11)
    case (304)
      c(3)=45.0d0/64.0d0/(a**6)
      c(12)=135.0d0/128.0d0/(a**7)
      c(17)=15.0d0/128.0d0/(a**7)
      c(19)=45.0d0/128.0d0/(a**7)
      c(37)=45.0d0/256.0d0/(a**8)
      c(42)=45.0d0/256.0d0/(a**8)
      c(44)=135.0d0/256.0d0/(a**8)
      c(53)=15.0d0/256.0d0/(a**8)
      c(86)=3.0d0/512.0d0/(a**9)
      c(91)=15.0d0/512.0d0/(a**9)
      c(93)=45.0d0/512.0d0/(a**9)
      c(102)=45.0d0/512.0d0/(a**9)
      c(172)=1.0d0/1024.0d0/(a**10)
      c(174)=3.0d0/1024.0d0/(a**10)
      c(183)=15.0d0/1024.0d0/(a**10)
      c(304)=1.0d0/2048.0d0/(a**11)
    case (305)
      c(4)=45.0d0/64.0d0/(a**6)
      c(13)=135.0d0/128.0d0/(a**7)
      c(18)=45.0d0/128.0d0/(a**7)
      c(20)=15.0d0/128.0d0/(a**7)
      c(38)=45.0d0/256.0d0/(a**8)
      c(43)=135.0d0/256.0d0/(a**8)
      c(45)=45.0d0/256.0d0/(a**8)
      c(54)=15.0d0/256.0d0/(a**8)
      c(87)=3.0d0/512.0d0/(a**9)
      c(92)=45.0d0/512.0d0/(a**9)
      c(94)=15.0d0/512.0d0/(a**9)
      c(103)=45.0d0/512.0d0/(a**9)
      c(173)=3.0d0/1024.0d0/(a**10)
      c(175)=1.0d0/1024.0d0/(a**10)
      c(184)=15.0d0/1024.0d0/(a**10)
      c(305)=1.0d0/2048.0d0/(a**11)
    case (306)
      c(3)=45.0d0/64.0d0/(a**6)
      c(12)=135.0d0/128.0d0/(a**7)
      c(19)=45.0d0/64.0d0/(a**7)
      c(37)=45.0d0/256.0d0/(a**8)
      c(44)=135.0d0/128.0d0/(a**8)
      c(55)=15.0d0/256.0d0/(a**8)
      c(86)=3.0d0/512.0d0/(a**9)
      c(93)=45.0d0/256.0d0/(a**9)
      c(104)=45.0d0/512.0d0/(a**9)
      c(174)=3.0d0/512.0d0/(a**10)
      c(185)=15.0d0/1024.0d0/(a**10)
      c(306)=1.0d0/2048.0d0/(a**11)
    case (307)
      c(4)=225.0d0/64.0d0/(a**6)
      c(13)=675.0d0/128.0d0/(a**7)
      c(20)=75.0d0/64.0d0/(a**7)
      c(38)=225.0d0/256.0d0/(a**8)
      c(45)=225.0d0/128.0d0/(a**8)
      c(56)=15.0d0/256.0d0/(a**8)
      c(87)=15.0d0/512.0d0/(a**9)
      c(94)=75.0d0/256.0d0/(a**9)
      c(105)=45.0d0/512.0d0/(a**9)
      c(175)=5.0d0/512.0d0/(a**10)
      c(186)=15.0d0/1024.0d0/(a**10)
      c(307)=1.0d0/2048.0d0/(a**11)
    case (308)
      c(2)=225.0d0/64.0d0/(a**6)
      c(11)=75.0d0/64.0d0/(a**7)
      c(14)=675.0d0/128.0d0/(a**7)
      c(36)=15.0d0/256.0d0/(a**8)
      c(39)=225.0d0/128.0d0/(a**8)
      c(46)=225.0d0/256.0d0/(a**8)
      c(88)=45.0d0/512.0d0/(a**9)
      c(95)=75.0d0/256.0d0/(a**9)
      c(106)=15.0d0/512.0d0/(a**9)
      c(176)=15.0d0/1024.0d0/(a**10)
      c(187)=5.0d0/512.0d0/(a**10)
      c(308)=1.0d0/2048.0d0/(a**11)
    case (309)
      c(15)=225.0d0/128.0d0/(a**7)
      c(40)=75.0d0/128.0d0/(a**8)
      c(47)=75.0d0/128.0d0/(a**8)
      c(89)=15.0d0/512.0d0/(a**9)
      c(96)=25.0d0/128.0d0/(a**9)
      c(107)=15.0d0/512.0d0/(a**9)
      c(177)=5.0d0/512.0d0/(a**10)
      c(188)=5.0d0/512.0d0/(a**10)
      c(309)=1.0d0/2048.0d0/(a**11)
    case (310)
      c(2)=45.0d0/64.0d0/(a**6)
      c(11)=15.0d0/64.0d0/(a**7)
      c(14)=45.0d0/64.0d0/(a**7)
      c(16)=45.0d0/128.0d0/(a**7)
      c(36)=3.0d0/256.0d0/(a**8)
      c(39)=15.0d0/64.0d0/(a**8)
      c(41)=15.0d0/128.0d0/(a**8)
      c(46)=15.0d0/256.0d0/(a**8)
      c(48)=45.0d0/128.0d0/(a**8)
      c(88)=3.0d0/256.0d0/(a**9)
      c(90)=3.0d0/512.0d0/(a**9)
      c(95)=5.0d0/256.0d0/(a**9)
      c(97)=15.0d0/128.0d0/(a**9)
      c(108)=15.0d0/512.0d0/(a**9)
      c(176)=1.0d0/1024.0d0/(a**10)
      c(178)=3.0d0/512.0d0/(a**10)
      c(189)=5.0d0/512.0d0/(a**10)
      c(310)=1.0d0/2048.0d0/(a**11)
    case (311)
      c(15)=135.0d0/128.0d0/(a**7)
      c(40)=45.0d0/128.0d0/(a**8)
      c(47)=45.0d0/256.0d0/(a**8)
      c(49)=45.0d0/256.0d0/(a**8)
      c(89)=9.0d0/512.0d0/(a**9)
      c(96)=15.0d0/256.0d0/(a**9)
      c(98)=15.0d0/256.0d0/(a**9)
      c(109)=15.0d0/512.0d0/(a**9)
      c(177)=3.0d0/1024.0d0/(a**10)
      c(179)=3.0d0/1024.0d0/(a**10)
      c(190)=5.0d0/512.0d0/(a**10)
      c(311)=1.0d0/2048.0d0/(a**11)
    case (312)
      c(2)=45.0d0/64.0d0/(a**6)
      c(11)=15.0d0/64.0d0/(a**7)
      c(14)=45.0d0/128.0d0/(a**7)
      c(16)=45.0d0/64.0d0/(a**7)
      c(36)=3.0d0/256.0d0/(a**8)
      c(39)=15.0d0/128.0d0/(a**8)
      c(41)=15.0d0/64.0d0/(a**8)
      c(48)=45.0d0/128.0d0/(a**8)
      c(50)=15.0d0/256.0d0/(a**8)
      c(88)=3.0d0/512.0d0/(a**9)
      c(90)=3.0d0/256.0d0/(a**9)
      c(97)=15.0d0/128.0d0/(a**9)
      c(99)=5.0d0/256.0d0/(a**9)
      c(110)=15.0d0/512.0d0/(a**9)
      c(178)=3.0d0/512.0d0/(a**10)
      c(180)=1.0d0/1024.0d0/(a**10)
      c(191)=5.0d0/512.0d0/(a**10)
      c(312)=1.0d0/2048.0d0/(a**11)
    case (313)
      c(15)=225.0d0/128.0d0/(a**7)
      c(40)=75.0d0/128.0d0/(a**8)
      c(49)=75.0d0/128.0d0/(a**8)
      c(89)=15.0d0/512.0d0/(a**9)
      c(98)=25.0d0/128.0d0/(a**9)
      c(111)=15.0d0/512.0d0/(a**9)
      c(179)=5.0d0/512.0d0/(a**10)
      c(192)=5.0d0/512.0d0/(a**10)
      c(313)=1.0d0/2048.0d0/(a**11)
    case (314)
      c(2)=225.0d0/64.0d0/(a**6)
      c(11)=75.0d0/64.0d0/(a**7)
      c(16)=675.0d0/128.0d0/(a**7)
      c(36)=15.0d0/256.0d0/(a**8)
      c(41)=225.0d0/128.0d0/(a**8)
      c(50)=225.0d0/256.0d0/(a**8)
      c(90)=45.0d0/512.0d0/(a**9)
      c(99)=75.0d0/256.0d0/(a**9)
      c(112)=15.0d0/512.0d0/(a**9)
      c(180)=15.0d0/1024.0d0/(a**10)
      c(193)=5.0d0/512.0d0/(a**10)
      c(314)=1.0d0/2048.0d0/(a**11)
    case (315)
      c(3)=315.0d0/64.0d0/(a**6)
      c(12)=315.0d0/64.0d0/(a**7)
      c(17)=315.0d0/128.0d0/(a**7)
      c(37)=105.0d0/256.0d0/(a**8)
      c(42)=315.0d0/128.0d0/(a**8)
      c(51)=63.0d0/256.0d0/(a**8)
      c(91)=105.0d0/512.0d0/(a**9)
      c(100)=63.0d0/256.0d0/(a**9)
      c(113)=3.0d0/512.0d0/(a**9)
      c(181)=21.0d0/1024.0d0/(a**10)
      c(194)=3.0d0/512.0d0/(a**10)
      c(315)=1.0d0/2048.0d0/(a**11)
    case (316)
      c(4)=45.0d0/64.0d0/(a**6)
      c(13)=45.0d0/64.0d0/(a**7)
      c(18)=135.0d0/128.0d0/(a**7)
      c(38)=15.0d0/256.0d0/(a**8)
      c(43)=135.0d0/128.0d0/(a**8)
      c(52)=45.0d0/256.0d0/(a**8)
      c(92)=45.0d0/512.0d0/(a**9)
      c(101)=45.0d0/256.0d0/(a**9)
      c(114)=3.0d0/512.0d0/(a**9)
      c(182)=15.0d0/1024.0d0/(a**10)
      c(195)=3.0d0/512.0d0/(a**10)
      c(316)=1.0d0/2048.0d0/(a**11)
    case (317)
      c(3)=45.0d0/64.0d0/(a**6)
      c(12)=45.0d0/64.0d0/(a**7)
      c(17)=15.0d0/64.0d0/(a**7)
      c(19)=45.0d0/128.0d0/(a**7)
      c(37)=15.0d0/256.0d0/(a**8)
      c(42)=15.0d0/64.0d0/(a**8)
      c(44)=45.0d0/128.0d0/(a**8)
      c(51)=3.0d0/256.0d0/(a**8)
      c(53)=15.0d0/128.0d0/(a**8)
      c(91)=5.0d0/256.0d0/(a**9)
      c(93)=15.0d0/512.0d0/(a**9)
      c(100)=3.0d0/256.0d0/(a**9)
      c(102)=15.0d0/128.0d0/(a**9)
      c(115)=3.0d0/512.0d0/(a**9)
      c(181)=1.0d0/1024.0d0/(a**10)
      c(183)=5.0d0/512.0d0/(a**10)
      c(196)=3.0d0/512.0d0/(a**10)
      c(317)=1.0d0/2048.0d0/(a**11)
    case (318)
      c(4)=27.0d0/64.0d0/(a**6)
      c(13)=27.0d0/64.0d0/(a**7)
      c(18)=27.0d0/64.0d0/(a**7)
      c(20)=9.0d0/128.0d0/(a**7)
      c(38)=9.0d0/256.0d0/(a**8)
      c(43)=27.0d0/64.0d0/(a**8)
      c(45)=9.0d0/128.0d0/(a**8)
      c(52)=9.0d0/256.0d0/(a**8)
      c(54)=9.0d0/128.0d0/(a**8)
      c(92)=9.0d0/256.0d0/(a**9)
      c(94)=3.0d0/512.0d0/(a**9)
      c(101)=9.0d0/256.0d0/(a**9)
      c(103)=9.0d0/128.0d0/(a**9)
      c(116)=3.0d0/512.0d0/(a**9)
      c(182)=3.0d0/1024.0d0/(a**10)
      c(184)=3.0d0/512.0d0/(a**10)
      c(197)=3.0d0/512.0d0/(a**10)
      c(318)=1.0d0/2048.0d0/(a**11)
    case (319)
      c(3)=27.0d0/64.0d0/(a**6)
      c(12)=27.0d0/64.0d0/(a**7)
      c(17)=9.0d0/128.0d0/(a**7)
      c(19)=27.0d0/64.0d0/(a**7)
      c(37)=9.0d0/256.0d0/(a**8)
      c(42)=9.0d0/128.0d0/(a**8)
      c(44)=27.0d0/64.0d0/(a**8)
      c(53)=9.0d0/128.0d0/(a**8)
      c(55)=9.0d0/256.0d0/(a**8)
      c(91)=3.0d0/512.0d0/(a**9)
      c(93)=9.0d0/256.0d0/(a**9)
      c(102)=9.0d0/128.0d0/(a**9)
      c(104)=9.0d0/256.0d0/(a**9)
      c(117)=3.0d0/512.0d0/(a**9)
      c(183)=3.0d0/512.0d0/(a**10)
      c(185)=3.0d0/1024.0d0/(a**10)
      c(198)=3.0d0/512.0d0/(a**10)
      c(319)=1.0d0/2048.0d0/(a**11)
    case (320)
      c(4)=45.0d0/64.0d0/(a**6)
      c(13)=45.0d0/64.0d0/(a**7)
      c(18)=45.0d0/128.0d0/(a**7)
      c(20)=15.0d0/64.0d0/(a**7)
      c(38)=15.0d0/256.0d0/(a**8)
      c(43)=45.0d0/128.0d0/(a**8)
      c(45)=15.0d0/64.0d0/(a**8)
      c(54)=15.0d0/128.0d0/(a**8)
      c(56)=3.0d0/256.0d0/(a**8)
      c(92)=15.0d0/512.0d0/(a**9)
      c(94)=5.0d0/256.0d0/(a**9)
      c(103)=15.0d0/128.0d0/(a**9)
      c(105)=3.0d0/256.0d0/(a**9)
      c(118)=3.0d0/512.0d0/(a**9)
      c(184)=5.0d0/512.0d0/(a**10)
      c(186)=1.0d0/1024.0d0/(a**10)
      c(199)=3.0d0/512.0d0/(a**10)
      c(320)=1.0d0/2048.0d0/(a**11)
    case (321)
      c(3)=45.0d0/64.0d0/(a**6)
      c(12)=45.0d0/64.0d0/(a**7)
      c(19)=135.0d0/128.0d0/(a**7)
      c(37)=15.0d0/256.0d0/(a**8)
      c(44)=135.0d0/128.0d0/(a**8)
      c(55)=45.0d0/256.0d0/(a**8)
      c(93)=45.0d0/512.0d0/(a**9)
      c(104)=45.0d0/256.0d0/(a**9)
      c(119)=3.0d0/512.0d0/(a**9)
      c(185)=15.0d0/1024.0d0/(a**10)
      c(200)=3.0d0/512.0d0/(a**10)
      c(321)=1.0d0/2048.0d0/(a**11)
    case (322)
      c(4)=315.0d0/64.0d0/(a**6)
      c(13)=315.0d0/64.0d0/(a**7)
      c(20)=315.0d0/128.0d0/(a**7)
      c(38)=105.0d0/256.0d0/(a**8)
      c(45)=315.0d0/128.0d0/(a**8)
      c(56)=63.0d0/256.0d0/(a**8)
      c(94)=105.0d0/512.0d0/(a**9)
      c(105)=63.0d0/256.0d0/(a**9)
      c(120)=3.0d0/512.0d0/(a**9)
      c(186)=21.0d0/1024.0d0/(a**10)
      c(201)=3.0d0/512.0d0/(a**10)
      c(322)=1.0d0/2048.0d0/(a**11)
    case (323)
      c(2)=315.0d0/64.0d0/(a**6)
      c(11)=105.0d0/128.0d0/(a**7)
      c(14)=315.0d0/32.0d0/(a**7)
      c(39)=105.0d0/64.0d0/(a**8)
      c(46)=315.0d0/128.0d0/(a**8)
      c(95)=105.0d0/256.0d0/(a**9)
      c(106)=21.0d0/128.0d0/(a**9)
      c(187)=7.0d0/256.0d0/(a**10)
      c(202)=3.0d0/1024.0d0/(a**10)
      c(323)=1.0d0/2048.0d0/(a**11)
    case (324)
      c(15)=315.0d0/128.0d0/(a**7)
      c(40)=105.0d0/256.0d0/(a**8)
      c(47)=315.0d0/256.0d0/(a**8)
      c(96)=105.0d0/512.0d0/(a**9)
      c(107)=63.0d0/512.0d0/(a**9)
      c(188)=21.0d0/1024.0d0/(a**10)
      c(203)=3.0d0/1024.0d0/(a**10)
      c(324)=1.0d0/2048.0d0/(a**11)
    case (325)
      c(2)=45.0d0/64.0d0/(a**6)
      c(11)=15.0d0/128.0d0/(a**7)
      c(14)=135.0d0/128.0d0/(a**7)
      c(16)=45.0d0/128.0d0/(a**7)
      c(39)=45.0d0/256.0d0/(a**8)
      c(41)=15.0d0/256.0d0/(a**8)
      c(46)=45.0d0/256.0d0/(a**8)
      c(48)=135.0d0/256.0d0/(a**8)
      c(95)=15.0d0/512.0d0/(a**9)
      c(97)=45.0d0/512.0d0/(a**9)
      c(106)=3.0d0/512.0d0/(a**9)
      c(108)=45.0d0/512.0d0/(a**9)
      c(187)=1.0d0/1024.0d0/(a**10)
      c(189)=15.0d0/1024.0d0/(a**10)
      c(204)=3.0d0/1024.0d0/(a**10)
      c(325)=1.0d0/2048.0d0/(a**11)
    case (326)
      c(15)=135.0d0/128.0d0/(a**7)
      c(40)=45.0d0/256.0d0/(a**8)
      c(47)=45.0d0/128.0d0/(a**8)
      c(49)=45.0d0/256.0d0/(a**8)
      c(96)=15.0d0/256.0d0/(a**9)
      c(98)=15.0d0/512.0d0/(a**9)
      c(107)=9.0d0/512.0d0/(a**9)
      c(109)=15.0d0/256.0d0/(a**9)
      c(188)=3.0d0/1024.0d0/(a**10)
      c(190)=5.0d0/512.0d0/(a**10)
      c(205)=3.0d0/1024.0d0/(a**10)
      c(326)=1.0d0/2048.0d0/(a**11)
    case (327)
      c(2)=27.0d0/64.0d0/(a**6)
      c(11)=9.0d0/128.0d0/(a**7)
      c(14)=27.0d0/64.0d0/(a**7)
      c(16)=27.0d0/64.0d0/(a**7)
      c(39)=9.0d0/128.0d0/(a**8)
      c(41)=9.0d0/128.0d0/(a**8)
      c(46)=9.0d0/256.0d0/(a**8)
      c(48)=27.0d0/64.0d0/(a**8)
      c(50)=9.0d0/256.0d0/(a**8)
      c(95)=3.0d0/512.0d0/(a**9)
      c(97)=9.0d0/128.0d0/(a**9)
      c(99)=3.0d0/512.0d0/(a**9)
      c(108)=9.0d0/256.0d0/(a**9)
      c(110)=9.0d0/256.0d0/(a**9)
      c(189)=3.0d0/512.0d0/(a**10)
      c(191)=3.0d0/512.0d0/(a**10)
      c(206)=3.0d0/1024.0d0/(a**10)
      c(327)=1.0d0/2048.0d0/(a**11)
    case (328)
      c(15)=135.0d0/128.0d0/(a**7)
      c(40)=45.0d0/256.0d0/(a**8)
      c(47)=45.0d0/256.0d0/(a**8)
      c(49)=45.0d0/128.0d0/(a**8)
      c(96)=15.0d0/512.0d0/(a**9)
      c(98)=15.0d0/256.0d0/(a**9)
      c(109)=15.0d0/256.0d0/(a**9)
      c(111)=9.0d0/512.0d0/(a**9)
      c(190)=5.0d0/512.0d0/(a**10)
      c(192)=3.0d0/1024.0d0/(a**10)
      c(207)=3.0d0/1024.0d0/(a**10)
      c(328)=1.0d0/2048.0d0/(a**11)
    case (329)
      c(2)=45.0d0/64.0d0/(a**6)
      c(11)=15.0d0/128.0d0/(a**7)
      c(14)=45.0d0/128.0d0/(a**7)
      c(16)=135.0d0/128.0d0/(a**7)
      c(39)=15.0d0/256.0d0/(a**8)
      c(41)=45.0d0/256.0d0/(a**8)
      c(48)=135.0d0/256.0d0/(a**8)
      c(50)=45.0d0/256.0d0/(a**8)
      c(97)=45.0d0/512.0d0/(a**9)
      c(99)=15.0d0/512.0d0/(a**9)
      c(110)=45.0d0/512.0d0/(a**9)
      c(112)=3.0d0/512.0d0/(a**9)
      c(191)=15.0d0/1024.0d0/(a**10)
      c(193)=1.0d0/1024.0d0/(a**10)
      c(208)=3.0d0/1024.0d0/(a**10)
      c(329)=1.0d0/2048.0d0/(a**11)
    case (330)
      c(15)=315.0d0/128.0d0/(a**7)
      c(40)=105.0d0/256.0d0/(a**8)
      c(49)=315.0d0/256.0d0/(a**8)
      c(98)=105.0d0/512.0d0/(a**9)
      c(111)=63.0d0/512.0d0/(a**9)
      c(192)=21.0d0/1024.0d0/(a**10)
      c(209)=3.0d0/1024.0d0/(a**10)
      c(330)=1.0d0/2048.0d0/(a**11)
    case (331)
      c(2)=315.0d0/64.0d0/(a**6)
      c(11)=105.0d0/128.0d0/(a**7)
      c(16)=315.0d0/32.0d0/(a**7)
      c(41)=105.0d0/64.0d0/(a**8)
      c(50)=315.0d0/128.0d0/(a**8)
      c(99)=105.0d0/256.0d0/(a**9)
      c(112)=21.0d0/128.0d0/(a**9)
      c(193)=7.0d0/256.0d0/(a**10)
      c(210)=3.0d0/1024.0d0/(a**10)
      c(331)=1.0d0/2048.0d0/(a**11)
    case (332)
      c(3)=945.0d0/64.0d0/(a**6)
      c(12)=945.0d0/128.0d0/(a**7)
      c(17)=315.0d0/32.0d0/(a**7)
      c(42)=315.0d0/64.0d0/(a**8)
      c(51)=189.0d0/128.0d0/(a**8)
      c(100)=189.0d0/256.0d0/(a**9)
      c(113)=9.0d0/128.0d0/(a**9)
      c(194)=9.0d0/256.0d0/(a**10)
      c(211)=1.0d0/1024.0d0/(a**10)
      c(332)=1.0d0/2048.0d0/(a**11)
    case (333)
      c(4)=105.0d0/64.0d0/(a**6)
      c(13)=105.0d0/128.0d0/(a**7)
      c(18)=105.0d0/32.0d0/(a**7)
      c(43)=105.0d0/64.0d0/(a**8)
      c(52)=105.0d0/128.0d0/(a**8)
      c(101)=105.0d0/256.0d0/(a**9)
      c(114)=7.0d0/128.0d0/(a**9)
      c(195)=7.0d0/256.0d0/(a**10)
      c(212)=1.0d0/1024.0d0/(a**10)
      c(333)=1.0d0/2048.0d0/(a**11)
    case (334)
      c(3)=105.0d0/64.0d0/(a**6)
      c(12)=105.0d0/128.0d0/(a**7)
      c(17)=105.0d0/128.0d0/(a**7)
      c(19)=105.0d0/128.0d0/(a**7)
      c(42)=105.0d0/256.0d0/(a**8)
      c(44)=105.0d0/256.0d0/(a**8)
      c(51)=21.0d0/256.0d0/(a**8)
      c(53)=105.0d0/256.0d0/(a**8)
      c(100)=21.0d0/512.0d0/(a**9)
      c(102)=105.0d0/512.0d0/(a**9)
      c(113)=1.0d0/512.0d0/(a**9)
      c(115)=21.0d0/512.0d0/(a**9)
      c(194)=1.0d0/1024.0d0/(a**10)
      c(196)=21.0d0/1024.0d0/(a**10)
      c(213)=1.0d0/1024.0d0/(a**10)
      c(334)=1.0d0/2048.0d0/(a**11)
    case (335)
      c(4)=45.0d0/64.0d0/(a**6)
      c(13)=45.0d0/128.0d0/(a**7)
      c(18)=135.0d0/128.0d0/(a**7)
      c(20)=15.0d0/128.0d0/(a**7)
      c(43)=135.0d0/256.0d0/(a**8)
      c(45)=15.0d0/256.0d0/(a**8)
      c(52)=45.0d0/256.0d0/(a**8)
      c(54)=45.0d0/256.0d0/(a**8)
      c(101)=45.0d0/512.0d0/(a**9)
      c(103)=45.0d0/512.0d0/(a**9)
      c(114)=3.0d0/512.0d0/(a**9)
      c(116)=15.0d0/512.0d0/(a**9)
      c(195)=3.0d0/1024.0d0/(a**10)
      c(197)=15.0d0/1024.0d0/(a**10)
      c(214)=1.0d0/1024.0d0/(a**10)
      c(335)=1.0d0/2048.0d0/(a**11)
    case (336)
      c(3)=45.0d0/64.0d0/(a**6)
      c(12)=45.0d0/128.0d0/(a**7)
      c(17)=15.0d0/64.0d0/(a**7)
      c(19)=45.0d0/64.0d0/(a**7)
      c(42)=15.0d0/128.0d0/(a**8)
      c(44)=45.0d0/128.0d0/(a**8)
      c(51)=3.0d0/256.0d0/(a**8)
      c(53)=15.0d0/64.0d0/(a**8)
      c(55)=15.0d0/256.0d0/(a**8)
      c(100)=3.0d0/512.0d0/(a**9)
      c(102)=15.0d0/128.0d0/(a**9)
      c(104)=15.0d0/512.0d0/(a**9)
      c(115)=3.0d0/256.0d0/(a**9)
      c(117)=5.0d0/256.0d0/(a**9)
      c(196)=3.0d0/512.0d0/(a**10)
      c(198)=5.0d0/512.0d0/(a**10)
      c(215)=1.0d0/1024.0d0/(a**10)
      c(336)=1.0d0/2048.0d0/(a**11)
    case (337)
      c(4)=45.0d0/64.0d0/(a**6)
      c(13)=45.0d0/128.0d0/(a**7)
      c(18)=45.0d0/64.0d0/(a**7)
      c(20)=15.0d0/64.0d0/(a**7)
      c(43)=45.0d0/128.0d0/(a**8)
      c(45)=15.0d0/128.0d0/(a**8)
      c(52)=15.0d0/256.0d0/(a**8)
      c(54)=15.0d0/64.0d0/(a**8)
      c(56)=3.0d0/256.0d0/(a**8)
      c(101)=15.0d0/512.0d0/(a**9)
      c(103)=15.0d0/128.0d0/(a**9)
      c(105)=3.0d0/512.0d0/(a**9)
      c(116)=5.0d0/256.0d0/(a**9)
      c(118)=3.0d0/256.0d0/(a**9)
      c(197)=5.0d0/512.0d0/(a**10)
      c(199)=3.0d0/512.0d0/(a**10)
      c(216)=1.0d0/1024.0d0/(a**10)
      c(337)=1.0d0/2048.0d0/(a**11)
    case (338)
      c(3)=45.0d0/64.0d0/(a**6)
      c(12)=45.0d0/128.0d0/(a**7)
      c(17)=15.0d0/128.0d0/(a**7)
      c(19)=135.0d0/128.0d0/(a**7)
      c(42)=15.0d0/256.0d0/(a**8)
      c(44)=135.0d0/256.0d0/(a**8)
      c(53)=45.0d0/256.0d0/(a**8)
      c(55)=45.0d0/256.0d0/(a**8)
      c(102)=45.0d0/512.0d0/(a**9)
      c(104)=45.0d0/512.0d0/(a**9)
      c(117)=15.0d0/512.0d0/(a**9)
      c(119)=3.0d0/512.0d0/(a**9)
      c(198)=15.0d0/1024.0d0/(a**10)
      c(200)=3.0d0/1024.0d0/(a**10)
      c(217)=1.0d0/1024.0d0/(a**10)
      c(338)=1.0d0/2048.0d0/(a**11)
    case (339)
      c(4)=105.0d0/64.0d0/(a**6)
      c(13)=105.0d0/128.0d0/(a**7)
      c(18)=105.0d0/128.0d0/(a**7)
      c(20)=105.0d0/128.0d0/(a**7)
      c(43)=105.0d0/256.0d0/(a**8)
      c(45)=105.0d0/256.0d0/(a**8)
      c(54)=105.0d0/256.0d0/(a**8)
      c(56)=21.0d0/256.0d0/(a**8)
      c(103)=105.0d0/512.0d0/(a**9)
      c(105)=21.0d0/512.0d0/(a**9)
      c(118)=21.0d0/512.0d0/(a**9)
      c(120)=1.0d0/512.0d0/(a**9)
      c(199)=21.0d0/1024.0d0/(a**10)
      c(201)=1.0d0/1024.0d0/(a**10)
      c(218)=1.0d0/1024.0d0/(a**10)
      c(339)=1.0d0/2048.0d0/(a**11)
    case (340)
      c(3)=105.0d0/64.0d0/(a**6)
      c(12)=105.0d0/128.0d0/(a**7)
      c(19)=105.0d0/32.0d0/(a**7)
      c(44)=105.0d0/64.0d0/(a**8)
      c(55)=105.0d0/128.0d0/(a**8)
      c(104)=105.0d0/256.0d0/(a**9)
      c(119)=7.0d0/128.0d0/(a**9)
      c(200)=7.0d0/256.0d0/(a**10)
      c(219)=1.0d0/1024.0d0/(a**10)
      c(340)=1.0d0/2048.0d0/(a**11)
    case (341)
      c(4)=945.0d0/64.0d0/(a**6)
      c(13)=945.0d0/128.0d0/(a**7)
      c(20)=315.0d0/32.0d0/(a**7)
      c(45)=315.0d0/64.0d0/(a**8)
      c(56)=189.0d0/128.0d0/(a**8)
      c(105)=189.0d0/256.0d0/(a**9)
      c(120)=9.0d0/128.0d0/(a**9)
      c(201)=9.0d0/256.0d0/(a**10)
      c(220)=1.0d0/1024.0d0/(a**10)
      c(341)=1.0d0/2048.0d0/(a**11)
    case (342)
      c(2)=945.0d0/64.0d0/(a**6)
      c(14)=4725.0d0/128.0d0/(a**7)
      c(46)=1575.0d0/128.0d0/(a**8)
      c(106)=315.0d0/256.0d0/(a**9)
      c(202)=45.0d0/1024.0d0/(a**10)
      c(342)=1.0d0/2048.0d0/(a**11)
    case (343)
      c(15)=945.0d0/128.0d0/(a**7)
      c(47)=315.0d0/64.0d0/(a**8)
      c(107)=189.0d0/256.0d0/(a**9)
      c(203)=9.0d0/256.0d0/(a**10)
      c(343)=1.0d0/2048.0d0/(a**11)
    case (344)
      c(2)=105.0d0/64.0d0/(a**6)
      c(14)=105.0d0/32.0d0/(a**7)
      c(16)=105.0d0/128.0d0/(a**7)
      c(46)=105.0d0/128.0d0/(a**8)
      c(48)=105.0d0/64.0d0/(a**8)
      c(106)=7.0d0/128.0d0/(a**9)
      c(108)=105.0d0/256.0d0/(a**9)
      c(202)=1.0d0/1024.0d0/(a**10)
      c(204)=7.0d0/256.0d0/(a**10)
      c(344)=1.0d0/2048.0d0/(a**11)
    case (345)
      c(15)=315.0d0/128.0d0/(a**7)
      c(47)=315.0d0/256.0d0/(a**8)
      c(49)=105.0d0/256.0d0/(a**8)
      c(107)=63.0d0/512.0d0/(a**9)
      c(109)=105.0d0/512.0d0/(a**9)
      c(203)=3.0d0/1024.0d0/(a**10)
      c(205)=21.0d0/1024.0d0/(a**10)
      c(345)=1.0d0/2048.0d0/(a**11)
    case (346)
      c(2)=45.0d0/64.0d0/(a**6)
      c(14)=135.0d0/128.0d0/(a**7)
      c(16)=45.0d0/64.0d0/(a**7)
      c(46)=45.0d0/256.0d0/(a**8)
      c(48)=135.0d0/128.0d0/(a**8)
      c(50)=15.0d0/256.0d0/(a**8)
      c(106)=3.0d0/512.0d0/(a**9)
      c(108)=45.0d0/256.0d0/(a**9)
      c(110)=45.0d0/512.0d0/(a**9)
      c(204)=3.0d0/512.0d0/(a**10)
      c(206)=15.0d0/1024.0d0/(a**10)
      c(346)=1.0d0/2048.0d0/(a**11)
    case (347)
      c(15)=225.0d0/128.0d0/(a**7)
      c(47)=75.0d0/128.0d0/(a**8)
      c(49)=75.0d0/128.0d0/(a**8)
      c(107)=15.0d0/512.0d0/(a**9)
      c(109)=25.0d0/128.0d0/(a**9)
      c(111)=15.0d0/512.0d0/(a**9)
      c(205)=5.0d0/512.0d0/(a**10)
      c(207)=5.0d0/512.0d0/(a**10)
      c(347)=1.0d0/2048.0d0/(a**11)
    case (348)
      c(2)=45.0d0/64.0d0/(a**6)
      c(14)=45.0d0/64.0d0/(a**7)
      c(16)=135.0d0/128.0d0/(a**7)
      c(46)=15.0d0/256.0d0/(a**8)
      c(48)=135.0d0/128.0d0/(a**8)
      c(50)=45.0d0/256.0d0/(a**8)
      c(108)=45.0d0/512.0d0/(a**9)
      c(110)=45.0d0/256.0d0/(a**9)
      c(112)=3.0d0/512.0d0/(a**9)
      c(206)=15.0d0/1024.0d0/(a**10)
      c(208)=3.0d0/512.0d0/(a**10)
      c(348)=1.0d0/2048.0d0/(a**11)
    case (349)
      c(15)=315.0d0/128.0d0/(a**7)
      c(47)=105.0d0/256.0d0/(a**8)
      c(49)=315.0d0/256.0d0/(a**8)
      c(109)=105.0d0/512.0d0/(a**9)
      c(111)=63.0d0/512.0d0/(a**9)
      c(207)=21.0d0/1024.0d0/(a**10)
      c(209)=3.0d0/1024.0d0/(a**10)
      c(349)=1.0d0/2048.0d0/(a**11)
    case (350)
      c(2)=105.0d0/64.0d0/(a**6)
      c(14)=105.0d0/128.0d0/(a**7)
      c(16)=105.0d0/32.0d0/(a**7)
      c(48)=105.0d0/64.0d0/(a**8)
      c(50)=105.0d0/128.0d0/(a**8)
      c(110)=105.0d0/256.0d0/(a**9)
      c(112)=7.0d0/128.0d0/(a**9)
      c(208)=7.0d0/256.0d0/(a**10)
      c(210)=1.0d0/1024.0d0/(a**10)
      c(350)=1.0d0/2048.0d0/(a**11)
    case (351)
      c(15)=945.0d0/128.0d0/(a**7)
      c(49)=315.0d0/64.0d0/(a**8)
      c(111)=189.0d0/256.0d0/(a**9)
      c(209)=9.0d0/256.0d0/(a**10)
      c(351)=1.0d0/2048.0d0/(a**11)
    case (352)
      c(2)=945.0d0/64.0d0/(a**6)
      c(16)=4725.0d0/128.0d0/(a**7)
      c(50)=1575.0d0/128.0d0/(a**8)
      c(112)=315.0d0/256.0d0/(a**9)
      c(210)=45.0d0/1024.0d0/(a**10)
      c(352)=1.0d0/2048.0d0/(a**11)
    case (353)
      c(3)=10395.0d0/64.0d0/(a**6)
      c(17)=17325.0d0/128.0d0/(a**7)
      c(51)=3465.0d0/128.0d0/(a**8)
      c(113)=495.0d0/256.0d0/(a**9)
      c(211)=55.0d0/1024.0d0/(a**10)
      c(353)=1.0d0/2048.0d0/(a**11)
    case (354)
      c(4)=945.0d0/64.0d0/(a**6)
      c(18)=4725.0d0/128.0d0/(a**7)
      c(52)=1575.0d0/128.0d0/(a**8)
      c(114)=315.0d0/256.0d0/(a**9)
      c(212)=45.0d0/1024.0d0/(a**10)
      c(354)=1.0d0/2048.0d0/(a**11)
    case (355)
      c(3)=945.0d0/64.0d0/(a**6)
      c(17)=315.0d0/32.0d0/(a**7)
      c(19)=945.0d0/128.0d0/(a**7)
      c(51)=189.0d0/128.0d0/(a**8)
      c(53)=315.0d0/64.0d0/(a**8)
      c(113)=9.0d0/128.0d0/(a**9)
      c(115)=189.0d0/256.0d0/(a**9)
      c(211)=1.0d0/1024.0d0/(a**10)
      c(213)=9.0d0/256.0d0/(a**10)
      c(355)=1.0d0/2048.0d0/(a**11)
    case (356)
      c(4)=315.0d0/64.0d0/(a**6)
      c(18)=315.0d0/32.0d0/(a**7)
      c(20)=105.0d0/128.0d0/(a**7)
      c(52)=315.0d0/128.0d0/(a**8)
      c(54)=105.0d0/64.0d0/(a**8)
      c(114)=21.0d0/128.0d0/(a**9)
      c(116)=105.0d0/256.0d0/(a**9)
      c(212)=3.0d0/1024.0d0/(a**10)
      c(214)=7.0d0/256.0d0/(a**10)
      c(356)=1.0d0/2048.0d0/(a**11)
    case (357)
      c(3)=315.0d0/64.0d0/(a**6)
      c(17)=315.0d0/128.0d0/(a**7)
      c(19)=315.0d0/64.0d0/(a**7)
      c(51)=63.0d0/256.0d0/(a**8)
      c(53)=315.0d0/128.0d0/(a**8)
      c(55)=105.0d0/256.0d0/(a**8)
      c(113)=3.0d0/512.0d0/(a**9)
      c(115)=63.0d0/256.0d0/(a**9)
      c(117)=105.0d0/512.0d0/(a**9)
      c(213)=3.0d0/512.0d0/(a**10)
      c(215)=21.0d0/1024.0d0/(a**10)
      c(357)=1.0d0/2048.0d0/(a**11)
    case (358)
      c(4)=225.0d0/64.0d0/(a**6)
      c(18)=675.0d0/128.0d0/(a**7)
      c(20)=75.0d0/64.0d0/(a**7)
      c(52)=225.0d0/256.0d0/(a**8)
      c(54)=225.0d0/128.0d0/(a**8)
      c(56)=15.0d0/256.0d0/(a**8)
      c(114)=15.0d0/512.0d0/(a**9)
      c(116)=75.0d0/256.0d0/(a**9)
      c(118)=45.0d0/512.0d0/(a**9)
      c(214)=5.0d0/512.0d0/(a**10)
      c(216)=15.0d0/1024.0d0/(a**10)
      c(358)=1.0d0/2048.0d0/(a**11)
    case (359)
      c(3)=225.0d0/64.0d0/(a**6)
      c(17)=75.0d0/64.0d0/(a**7)
      c(19)=675.0d0/128.0d0/(a**7)
      c(51)=15.0d0/256.0d0/(a**8)
      c(53)=225.0d0/128.0d0/(a**8)
      c(55)=225.0d0/256.0d0/(a**8)
      c(115)=45.0d0/512.0d0/(a**9)
      c(117)=75.0d0/256.0d0/(a**9)
      c(119)=15.0d0/512.0d0/(a**9)
      c(215)=15.0d0/1024.0d0/(a**10)
      c(217)=5.0d0/512.0d0/(a**10)
      c(359)=1.0d0/2048.0d0/(a**11)
    case (360)
      c(4)=315.0d0/64.0d0/(a**6)
      c(18)=315.0d0/64.0d0/(a**7)
      c(20)=315.0d0/128.0d0/(a**7)
      c(52)=105.0d0/256.0d0/(a**8)
      c(54)=315.0d0/128.0d0/(a**8)
      c(56)=63.0d0/256.0d0/(a**8)
      c(116)=105.0d0/512.0d0/(a**9)
      c(118)=63.0d0/256.0d0/(a**9)
      c(120)=3.0d0/512.0d0/(a**9)
      c(216)=21.0d0/1024.0d0/(a**10)
      c(218)=3.0d0/512.0d0/(a**10)
      c(360)=1.0d0/2048.0d0/(a**11)
    case (361)
      c(3)=315.0d0/64.0d0/(a**6)
      c(17)=105.0d0/128.0d0/(a**7)
      c(19)=315.0d0/32.0d0/(a**7)
      c(53)=105.0d0/64.0d0/(a**8)
      c(55)=315.0d0/128.0d0/(a**8)
      c(117)=105.0d0/256.0d0/(a**9)
      c(119)=21.0d0/128.0d0/(a**9)
      c(217)=7.0d0/256.0d0/(a**10)
      c(219)=3.0d0/1024.0d0/(a**10)
      c(361)=1.0d0/2048.0d0/(a**11)
    case (362)
      c(4)=945.0d0/64.0d0/(a**6)
      c(18)=945.0d0/128.0d0/(a**7)
      c(20)=315.0d0/32.0d0/(a**7)
      c(54)=315.0d0/64.0d0/(a**8)
      c(56)=189.0d0/128.0d0/(a**8)
      c(118)=189.0d0/256.0d0/(a**9)
      c(120)=9.0d0/128.0d0/(a**9)
      c(218)=9.0d0/256.0d0/(a**10)
      c(220)=1.0d0/1024.0d0/(a**10)
      c(362)=1.0d0/2048.0d0/(a**11)
    case (363)
      c(3)=945.0d0/64.0d0/(a**6)
      c(19)=4725.0d0/128.0d0/(a**7)
      c(55)=1575.0d0/128.0d0/(a**8)
      c(119)=315.0d0/256.0d0/(a**9)
      c(219)=45.0d0/1024.0d0/(a**10)
      c(363)=1.0d0/2048.0d0/(a**11)
    case (364)
      c(4)=10395.0d0/64.0d0/(a**6)
      c(20)=17325.0d0/128.0d0/(a**7)
      c(56)=3465.0d0/128.0d0/(a**8)
      c(120)=495.0d0/256.0d0/(a**9)
      c(220)=55.0d0/1024.0d0/(a**10)
      c(364)=1.0d0/2048.0d0/(a**11)
    case (365)
      c(1)=10395.0d0/64.0d0/(a**6)
      c(5)=31185.0d0/64.0d0/(a**7)
      c(21)=51975.0d0/256.0d0/(a**8)
      c(57)=3465.0d0/128.0d0/(a**9)
      c(121)=1485.0d0/1024.0d0/(a**10)
      c(221)=33.0d0/1024.0d0/(a**11)
      c(365)=1.0d0/4096.0d0/(a**12)
    case (366)
      c(6)=10395.0d0/128.0d0/(a**7)
      c(22)=17325.0d0/256.0d0/(a**8)
      c(58)=3465.0d0/256.0d0/(a**9)
      c(122)=495.0d0/512.0d0/(a**10)
      c(222)=55.0d0/2048.0d0/(a**11)
      c(366)=1.0d0/4096.0d0/(a**12)
    case (367)
      c(7)=10395.0d0/128.0d0/(a**7)
      c(23)=17325.0d0/256.0d0/(a**8)
      c(59)=3465.0d0/256.0d0/(a**9)
      c(123)=495.0d0/512.0d0/(a**10)
      c(223)=55.0d0/2048.0d0/(a**11)
      c(367)=1.0d0/4096.0d0/(a**12)
    case (368)
      c(1)=945.0d0/64.0d0/(a**6)
      c(5)=4725.0d0/128.0d0/(a**7)
      c(8)=945.0d0/128.0d0/(a**7)
      c(21)=1575.0d0/128.0d0/(a**8)
      c(24)=4725.0d0/256.0d0/(a**8)
      c(57)=315.0d0/256.0d0/(a**9)
      c(60)=1575.0d0/256.0d0/(a**9)
      c(121)=45.0d0/1024.0d0/(a**10)
      c(124)=315.0d0/512.0d0/(a**10)
      c(221)=1.0d0/2048.0d0/(a**11)
      c(224)=45.0d0/2048.0d0/(a**11)
      c(368)=1.0d0/4096.0d0/(a**12)
    case (369)
      c(9)=945.0d0/128.0d0/(a**7)
      c(25)=4725.0d0/256.0d0/(a**8)
      c(61)=1575.0d0/256.0d0/(a**9)
      c(125)=315.0d0/512.0d0/(a**10)
      c(225)=45.0d0/2048.0d0/(a**11)
      c(369)=1.0d0/4096.0d0/(a**12)
    case (370)
      c(1)=945.0d0/64.0d0/(a**6)
      c(5)=4725.0d0/128.0d0/(a**7)
      c(10)=945.0d0/128.0d0/(a**7)
      c(21)=1575.0d0/128.0d0/(a**8)
      c(26)=4725.0d0/256.0d0/(a**8)
      c(57)=315.0d0/256.0d0/(a**9)
      c(62)=1575.0d0/256.0d0/(a**9)
      c(121)=45.0d0/1024.0d0/(a**10)
      c(126)=315.0d0/512.0d0/(a**10)
      c(221)=1.0d0/2048.0d0/(a**11)
      c(226)=45.0d0/2048.0d0/(a**11)
      c(370)=1.0d0/4096.0d0/(a**12)
    case (371)
      c(6)=2835.0d0/128.0d0/(a**7)
      c(22)=945.0d0/64.0d0/(a**8)
      c(27)=945.0d0/256.0d0/(a**8)
      c(58)=567.0d0/256.0d0/(a**9)
      c(63)=315.0d0/128.0d0/(a**9)
      c(122)=27.0d0/256.0d0/(a**10)
      c(127)=189.0d0/512.0d0/(a**10)
      c(222)=3.0d0/2048.0d0/(a**11)
      c(227)=9.0d0/512.0d0/(a**11)
      c(371)=1.0d0/4096.0d0/(a**12)
    case (372)
      c(7)=945.0d0/128.0d0/(a**7)
      c(23)=315.0d0/64.0d0/(a**8)
      c(28)=945.0d0/256.0d0/(a**8)
      c(59)=189.0d0/256.0d0/(a**9)
      c(64)=315.0d0/128.0d0/(a**9)
      c(123)=9.0d0/256.0d0/(a**10)
      c(128)=189.0d0/512.0d0/(a**10)
      c(223)=1.0d0/2048.0d0/(a**11)
      c(228)=9.0d0/512.0d0/(a**11)
      c(372)=1.0d0/4096.0d0/(a**12)
    case (373)
      c(6)=945.0d0/128.0d0/(a**7)
      c(22)=315.0d0/64.0d0/(a**8)
      c(29)=945.0d0/256.0d0/(a**8)
      c(58)=189.0d0/256.0d0/(a**9)
      c(65)=315.0d0/128.0d0/(a**9)
      c(122)=9.0d0/256.0d0/(a**10)
      c(129)=189.0d0/512.0d0/(a**10)
      c(222)=1.0d0/2048.0d0/(a**11)
      c(229)=9.0d0/512.0d0/(a**11)
      c(373)=1.0d0/4096.0d0/(a**12)
    case (374)
      c(7)=2835.0d0/128.0d0/(a**7)
      c(23)=945.0d0/64.0d0/(a**8)
      c(30)=945.0d0/256.0d0/(a**8)
      c(59)=567.0d0/256.0d0/(a**9)
      c(66)=315.0d0/128.0d0/(a**9)
      c(123)=27.0d0/256.0d0/(a**10)
      c(130)=189.0d0/512.0d0/(a**10)
      c(223)=3.0d0/2048.0d0/(a**11)
      c(230)=9.0d0/512.0d0/(a**11)
      c(374)=1.0d0/4096.0d0/(a**12)
    case (375)
      c(1)=315.0d0/64.0d0/(a**6)
      c(5)=315.0d0/32.0d0/(a**7)
      c(8)=315.0d0/64.0d0/(a**7)
      c(21)=315.0d0/128.0d0/(a**8)
      c(24)=315.0d0/32.0d0/(a**8)
      c(31)=105.0d0/256.0d0/(a**8)
      c(57)=21.0d0/128.0d0/(a**9)
      c(60)=315.0d0/128.0d0/(a**9)
      c(67)=105.0d0/128.0d0/(a**9)
      c(121)=3.0d0/1024.0d0/(a**10)
      c(124)=21.0d0/128.0d0/(a**10)
      c(131)=105.0d0/512.0d0/(a**10)
      c(224)=3.0d0/1024.0d0/(a**11)
      c(231)=7.0d0/512.0d0/(a**11)
      c(375)=1.0d0/4096.0d0/(a**12)
    case (376)
      c(9)=315.0d0/128.0d0/(a**7)
      c(25)=315.0d0/64.0d0/(a**8)
      c(32)=105.0d0/256.0d0/(a**8)
      c(61)=315.0d0/256.0d0/(a**9)
      c(68)=105.0d0/128.0d0/(a**9)
      c(125)=21.0d0/256.0d0/(a**10)
      c(132)=105.0d0/512.0d0/(a**10)
      c(225)=3.0d0/2048.0d0/(a**11)
      c(232)=7.0d0/512.0d0/(a**11)
      c(376)=1.0d0/4096.0d0/(a**12)
    case (377)
      c(1)=105.0d0/64.0d0/(a**6)
      c(5)=105.0d0/32.0d0/(a**7)
      c(8)=105.0d0/128.0d0/(a**7)
      c(10)=105.0d0/128.0d0/(a**7)
      c(21)=105.0d0/128.0d0/(a**8)
      c(24)=105.0d0/64.0d0/(a**8)
      c(26)=105.0d0/64.0d0/(a**8)
      c(33)=105.0d0/256.0d0/(a**8)
      c(57)=7.0d0/128.0d0/(a**9)
      c(60)=105.0d0/256.0d0/(a**9)
      c(62)=105.0d0/256.0d0/(a**9)
      c(69)=105.0d0/128.0d0/(a**9)
      c(121)=1.0d0/1024.0d0/(a**10)
      c(124)=7.0d0/256.0d0/(a**10)
      c(126)=7.0d0/256.0d0/(a**10)
      c(133)=105.0d0/512.0d0/(a**10)
      c(224)=1.0d0/2048.0d0/(a**11)
      c(226)=1.0d0/2048.0d0/(a**11)
      c(233)=7.0d0/512.0d0/(a**11)
      c(377)=1.0d0/4096.0d0/(a**12)
    case (378)
      c(9)=315.0d0/128.0d0/(a**7)
      c(25)=315.0d0/64.0d0/(a**8)
      c(34)=105.0d0/256.0d0/(a**8)
      c(61)=315.0d0/256.0d0/(a**9)
      c(70)=105.0d0/128.0d0/(a**9)
      c(125)=21.0d0/256.0d0/(a**10)
      c(134)=105.0d0/512.0d0/(a**10)
      c(225)=3.0d0/2048.0d0/(a**11)
      c(234)=7.0d0/512.0d0/(a**11)
      c(378)=1.0d0/4096.0d0/(a**12)
    case (379)
      c(1)=315.0d0/64.0d0/(a**6)
      c(5)=315.0d0/32.0d0/(a**7)
      c(10)=315.0d0/64.0d0/(a**7)
      c(21)=315.0d0/128.0d0/(a**8)
      c(26)=315.0d0/32.0d0/(a**8)
      c(35)=105.0d0/256.0d0/(a**8)
      c(57)=21.0d0/128.0d0/(a**9)
      c(62)=315.0d0/128.0d0/(a**9)
      c(71)=105.0d0/128.0d0/(a**9)
      c(121)=3.0d0/1024.0d0/(a**10)
      c(126)=21.0d0/128.0d0/(a**10)
      c(135)=105.0d0/512.0d0/(a**10)
      c(226)=3.0d0/1024.0d0/(a**11)
      c(235)=7.0d0/512.0d0/(a**11)
      c(379)=1.0d0/4096.0d0/(a**12)
    case (380)
      c(6)=1575.0d0/128.0d0/(a**7)
      c(22)=1575.0d0/256.0d0/(a**8)
      c(27)=525.0d0/128.0d0/(a**8)
      c(58)=315.0d0/512.0d0/(a**9)
      c(63)=525.0d0/256.0d0/(a**9)
      c(72)=105.0d0/512.0d0/(a**9)
      c(122)=15.0d0/1024.0d0/(a**10)
      c(127)=105.0d0/512.0d0/(a**10)
      c(136)=105.0d0/1024.0d0/(a**10)
      c(227)=5.0d0/1024.0d0/(a**11)
      c(236)=21.0d0/2048.0d0/(a**11)
      c(380)=1.0d0/4096.0d0/(a**12)
    case (381)
      c(7)=315.0d0/128.0d0/(a**7)
      c(23)=315.0d0/256.0d0/(a**8)
      c(28)=315.0d0/128.0d0/(a**8)
      c(59)=63.0d0/512.0d0/(a**9)
      c(64)=315.0d0/256.0d0/(a**9)
      c(73)=105.0d0/512.0d0/(a**9)
      c(123)=3.0d0/1024.0d0/(a**10)
      c(128)=63.0d0/512.0d0/(a**10)
      c(137)=105.0d0/1024.0d0/(a**10)
      c(228)=3.0d0/1024.0d0/(a**11)
      c(237)=21.0d0/2048.0d0/(a**11)
      c(381)=1.0d0/4096.0d0/(a**12)
    case (382)
      c(6)=315.0d0/128.0d0/(a**7)
      c(22)=315.0d0/256.0d0/(a**8)
      c(27)=105.0d0/256.0d0/(a**8)
      c(29)=315.0d0/256.0d0/(a**8)
      c(58)=63.0d0/512.0d0/(a**9)
      c(63)=105.0d0/512.0d0/(a**9)
      c(65)=315.0d0/512.0d0/(a**9)
      c(74)=105.0d0/512.0d0/(a**9)
      c(122)=3.0d0/1024.0d0/(a**10)
      c(127)=21.0d0/1024.0d0/(a**10)
      c(129)=63.0d0/1024.0d0/(a**10)
      c(138)=105.0d0/1024.0d0/(a**10)
      c(227)=1.0d0/2048.0d0/(a**11)
      c(229)=3.0d0/2048.0d0/(a**11)
      c(238)=21.0d0/2048.0d0/(a**11)
      c(382)=1.0d0/4096.0d0/(a**12)
    case (383)
      c(7)=315.0d0/128.0d0/(a**7)
      c(23)=315.0d0/256.0d0/(a**8)
      c(28)=315.0d0/256.0d0/(a**8)
      c(30)=105.0d0/256.0d0/(a**8)
      c(59)=63.0d0/512.0d0/(a**9)
      c(64)=315.0d0/512.0d0/(a**9)
      c(66)=105.0d0/512.0d0/(a**9)
      c(75)=105.0d0/512.0d0/(a**9)
      c(123)=3.0d0/1024.0d0/(a**10)
      c(128)=63.0d0/1024.0d0/(a**10)
      c(130)=21.0d0/1024.0d0/(a**10)
      c(139)=105.0d0/1024.0d0/(a**10)
      c(228)=3.0d0/2048.0d0/(a**11)
      c(230)=1.0d0/2048.0d0/(a**11)
      c(239)=21.0d0/2048.0d0/(a**11)
      c(383)=1.0d0/4096.0d0/(a**12)
    case (384)
      c(6)=315.0d0/128.0d0/(a**7)
      c(22)=315.0d0/256.0d0/(a**8)
      c(29)=315.0d0/128.0d0/(a**8)
      c(58)=63.0d0/512.0d0/(a**9)
      c(65)=315.0d0/256.0d0/(a**9)
      c(76)=105.0d0/512.0d0/(a**9)
      c(122)=3.0d0/1024.0d0/(a**10)
      c(129)=63.0d0/512.0d0/(a**10)
      c(140)=105.0d0/1024.0d0/(a**10)
      c(229)=3.0d0/1024.0d0/(a**11)
      c(240)=21.0d0/2048.0d0/(a**11)
      c(384)=1.0d0/4096.0d0/(a**12)
    case (385)
      c(7)=1575.0d0/128.0d0/(a**7)
      c(23)=1575.0d0/256.0d0/(a**8)
      c(30)=525.0d0/128.0d0/(a**8)
      c(59)=315.0d0/512.0d0/(a**9)
      c(66)=525.0d0/256.0d0/(a**9)
      c(77)=105.0d0/512.0d0/(a**9)
      c(123)=15.0d0/1024.0d0/(a**10)
      c(130)=105.0d0/512.0d0/(a**10)
      c(141)=105.0d0/1024.0d0/(a**10)
      c(230)=5.0d0/1024.0d0/(a**11)
      c(241)=21.0d0/2048.0d0/(a**11)
      c(385)=1.0d0/4096.0d0/(a**12)
    case (386)
      c(1)=225.0d0/64.0d0/(a**6)
      c(5)=675.0d0/128.0d0/(a**7)
      c(8)=675.0d0/128.0d0/(a**7)
      c(21)=225.0d0/256.0d0/(a**8)
      c(24)=2025.0d0/256.0d0/(a**8)
      c(31)=225.0d0/256.0d0/(a**8)
      c(57)=15.0d0/512.0d0/(a**9)
      c(60)=675.0d0/512.0d0/(a**9)
      c(67)=675.0d0/512.0d0/(a**9)
      c(78)=15.0d0/512.0d0/(a**9)
      c(124)=45.0d0/1024.0d0/(a**10)
      c(131)=225.0d0/1024.0d0/(a**10)
      c(142)=45.0d0/1024.0d0/(a**10)
      c(231)=15.0d0/2048.0d0/(a**11)
      c(242)=15.0d0/2048.0d0/(a**11)
      c(386)=1.0d0/4096.0d0/(a**12)
    case (387)
      c(9)=225.0d0/128.0d0/(a**7)
      c(25)=675.0d0/256.0d0/(a**8)
      c(32)=75.0d0/128.0d0/(a**8)
      c(61)=225.0d0/512.0d0/(a**9)
      c(68)=225.0d0/256.0d0/(a**9)
      c(79)=15.0d0/512.0d0/(a**9)
      c(125)=15.0d0/1024.0d0/(a**10)
      c(132)=75.0d0/512.0d0/(a**10)
      c(143)=45.0d0/1024.0d0/(a**10)
      c(232)=5.0d0/1024.0d0/(a**11)
      c(243)=15.0d0/2048.0d0/(a**11)
      c(387)=1.0d0/4096.0d0/(a**12)
    case (388)
      c(1)=45.0d0/64.0d0/(a**6)
      c(5)=135.0d0/128.0d0/(a**7)
      c(8)=45.0d0/64.0d0/(a**7)
      c(10)=45.0d0/128.0d0/(a**7)
      c(21)=45.0d0/256.0d0/(a**8)
      c(24)=135.0d0/128.0d0/(a**8)
      c(26)=135.0d0/256.0d0/(a**8)
      c(31)=15.0d0/256.0d0/(a**8)
      c(33)=45.0d0/128.0d0/(a**8)
      c(57)=3.0d0/512.0d0/(a**9)
      c(60)=45.0d0/256.0d0/(a**9)
      c(62)=45.0d0/512.0d0/(a**9)
      c(67)=45.0d0/512.0d0/(a**9)
      c(69)=135.0d0/256.0d0/(a**9)
      c(80)=15.0d0/512.0d0/(a**9)
      c(124)=3.0d0/512.0d0/(a**10)
      c(126)=3.0d0/1024.0d0/(a**10)
      c(131)=15.0d0/1024.0d0/(a**10)
      c(133)=45.0d0/512.0d0/(a**10)
      c(144)=45.0d0/1024.0d0/(a**10)
      c(231)=1.0d0/2048.0d0/(a**11)
      c(233)=3.0d0/1024.0d0/(a**11)
      c(244)=15.0d0/2048.0d0/(a**11)
      c(388)=1.0d0/4096.0d0/(a**12)
    case (389)
      c(9)=135.0d0/128.0d0/(a**7)
      c(25)=405.0d0/256.0d0/(a**8)
      c(32)=45.0d0/256.0d0/(a**8)
      c(34)=45.0d0/256.0d0/(a**8)
      c(61)=135.0d0/512.0d0/(a**9)
      c(68)=135.0d0/512.0d0/(a**9)
      c(70)=135.0d0/512.0d0/(a**9)
      c(81)=15.0d0/512.0d0/(a**9)
      c(125)=9.0d0/1024.0d0/(a**10)
      c(132)=45.0d0/1024.0d0/(a**10)
      c(134)=45.0d0/1024.0d0/(a**10)
      c(145)=45.0d0/1024.0d0/(a**10)
      c(232)=3.0d0/2048.0d0/(a**11)
      c(234)=3.0d0/2048.0d0/(a**11)
      c(245)=15.0d0/2048.0d0/(a**11)
      c(389)=1.0d0/4096.0d0/(a**12)
    case (390)
      c(1)=45.0d0/64.0d0/(a**6)
      c(5)=135.0d0/128.0d0/(a**7)
      c(8)=45.0d0/128.0d0/(a**7)
      c(10)=45.0d0/64.0d0/(a**7)
      c(21)=45.0d0/256.0d0/(a**8)
      c(24)=135.0d0/256.0d0/(a**8)
      c(26)=135.0d0/128.0d0/(a**8)
      c(33)=45.0d0/128.0d0/(a**8)
      c(35)=15.0d0/256.0d0/(a**8)
      c(57)=3.0d0/512.0d0/(a**9)
      c(60)=45.0d0/512.0d0/(a**9)
      c(62)=45.0d0/256.0d0/(a**9)
      c(69)=135.0d0/256.0d0/(a**9)
      c(71)=45.0d0/512.0d0/(a**9)
      c(82)=15.0d0/512.0d0/(a**9)
      c(124)=3.0d0/1024.0d0/(a**10)
      c(126)=3.0d0/512.0d0/(a**10)
      c(133)=45.0d0/512.0d0/(a**10)
      c(135)=15.0d0/1024.0d0/(a**10)
      c(146)=45.0d0/1024.0d0/(a**10)
      c(233)=3.0d0/1024.0d0/(a**11)
      c(235)=1.0d0/2048.0d0/(a**11)
      c(246)=15.0d0/2048.0d0/(a**11)
      c(390)=1.0d0/4096.0d0/(a**12)
    case (391)
      c(9)=225.0d0/128.0d0/(a**7)
      c(25)=675.0d0/256.0d0/(a**8)
      c(34)=75.0d0/128.0d0/(a**8)
      c(61)=225.0d0/512.0d0/(a**9)
      c(70)=225.0d0/256.0d0/(a**9)
      c(83)=15.0d0/512.0d0/(a**9)
      c(125)=15.0d0/1024.0d0/(a**10)
      c(134)=75.0d0/512.0d0/(a**10)
      c(147)=45.0d0/1024.0d0/(a**10)
      c(234)=5.0d0/1024.0d0/(a**11)
      c(247)=15.0d0/2048.0d0/(a**11)
      c(391)=1.0d0/4096.0d0/(a**12)
    case (392)
      c(1)=225.0d0/64.0d0/(a**6)
      c(5)=675.0d0/128.0d0/(a**7)
      c(10)=675.0d0/128.0d0/(a**7)
      c(21)=225.0d0/256.0d0/(a**8)
      c(26)=2025.0d0/256.0d0/(a**8)
      c(35)=225.0d0/256.0d0/(a**8)
      c(57)=15.0d0/512.0d0/(a**9)
      c(62)=675.0d0/512.0d0/(a**9)
      c(71)=675.0d0/512.0d0/(a**9)
      c(84)=15.0d0/512.0d0/(a**9)
      c(126)=45.0d0/1024.0d0/(a**10)
      c(135)=225.0d0/1024.0d0/(a**10)
      c(148)=45.0d0/1024.0d0/(a**10)
      c(235)=15.0d0/2048.0d0/(a**11)
      c(248)=15.0d0/2048.0d0/(a**11)
      c(392)=1.0d0/4096.0d0/(a**12)
    case (393)
      c(6)=1575.0d0/128.0d0/(a**7)
      c(22)=525.0d0/128.0d0/(a**8)
      c(27)=1575.0d0/256.0d0/(a**8)
      c(58)=105.0d0/512.0d0/(a**9)
      c(63)=525.0d0/256.0d0/(a**9)
      c(72)=315.0d0/512.0d0/(a**9)
      c(127)=105.0d0/1024.0d0/(a**10)
      c(136)=105.0d0/512.0d0/(a**10)
      c(149)=15.0d0/1024.0d0/(a**10)
      c(236)=21.0d0/2048.0d0/(a**11)
      c(249)=5.0d0/1024.0d0/(a**11)
      c(393)=1.0d0/4096.0d0/(a**12)
    case (394)
      c(7)=225.0d0/128.0d0/(a**7)
      c(23)=75.0d0/128.0d0/(a**8)
      c(28)=675.0d0/256.0d0/(a**8)
      c(59)=15.0d0/512.0d0/(a**9)
      c(64)=225.0d0/256.0d0/(a**9)
      c(73)=225.0d0/512.0d0/(a**9)
      c(128)=45.0d0/1024.0d0/(a**10)
      c(137)=75.0d0/512.0d0/(a**10)
      c(150)=15.0d0/1024.0d0/(a**10)
      c(237)=15.0d0/2048.0d0/(a**11)
      c(250)=5.0d0/1024.0d0/(a**11)
      c(394)=1.0d0/4096.0d0/(a**12)
    case (395)
      c(6)=225.0d0/128.0d0/(a**7)
      c(22)=75.0d0/128.0d0/(a**8)
      c(27)=75.0d0/128.0d0/(a**8)
      c(29)=225.0d0/256.0d0/(a**8)
      c(58)=15.0d0/512.0d0/(a**9)
      c(63)=25.0d0/128.0d0/(a**9)
      c(65)=75.0d0/256.0d0/(a**9)
      c(72)=15.0d0/512.0d0/(a**9)
      c(74)=75.0d0/256.0d0/(a**9)
      c(127)=5.0d0/512.0d0/(a**10)
      c(129)=15.0d0/1024.0d0/(a**10)
      c(136)=5.0d0/512.0d0/(a**10)
      c(138)=25.0d0/256.0d0/(a**10)
      c(151)=15.0d0/1024.0d0/(a**10)
      c(236)=1.0d0/2048.0d0/(a**11)
      c(238)=5.0d0/1024.0d0/(a**11)
      c(251)=5.0d0/1024.0d0/(a**11)
      c(395)=1.0d0/4096.0d0/(a**12)
    case (396)
      c(7)=135.0d0/128.0d0/(a**7)
      c(23)=45.0d0/128.0d0/(a**8)
      c(28)=135.0d0/128.0d0/(a**8)
      c(30)=45.0d0/256.0d0/(a**8)
      c(59)=9.0d0/512.0d0/(a**9)
      c(64)=45.0d0/128.0d0/(a**9)
      c(66)=15.0d0/256.0d0/(a**9)
      c(73)=45.0d0/512.0d0/(a**9)
      c(75)=45.0d0/256.0d0/(a**9)
      c(128)=9.0d0/512.0d0/(a**10)
      c(130)=3.0d0/1024.0d0/(a**10)
      c(137)=15.0d0/512.0d0/(a**10)
      c(139)=15.0d0/256.0d0/(a**10)
      c(152)=15.0d0/1024.0d0/(a**10)
      c(237)=3.0d0/2048.0d0/(a**11)
      c(239)=3.0d0/1024.0d0/(a**11)
      c(252)=5.0d0/1024.0d0/(a**11)
      c(396)=1.0d0/4096.0d0/(a**12)
    case (397)
      c(6)=135.0d0/128.0d0/(a**7)
      c(22)=45.0d0/128.0d0/(a**8)
      c(27)=45.0d0/256.0d0/(a**8)
      c(29)=135.0d0/128.0d0/(a**8)
      c(58)=9.0d0/512.0d0/(a**9)
      c(63)=15.0d0/256.0d0/(a**9)
      c(65)=45.0d0/128.0d0/(a**9)
      c(74)=45.0d0/256.0d0/(a**9)
      c(76)=45.0d0/512.0d0/(a**9)
      c(127)=3.0d0/1024.0d0/(a**10)
      c(129)=9.0d0/512.0d0/(a**10)
      c(138)=15.0d0/256.0d0/(a**10)
      c(140)=15.0d0/512.0d0/(a**10)
      c(153)=15.0d0/1024.0d0/(a**10)
      c(238)=3.0d0/1024.0d0/(a**11)
      c(240)=3.0d0/2048.0d0/(a**11)
      c(253)=5.0d0/1024.0d0/(a**11)
      c(397)=1.0d0/4096.0d0/(a**12)
    case (398)
      c(7)=225.0d0/128.0d0/(a**7)
      c(23)=75.0d0/128.0d0/(a**8)
      c(28)=225.0d0/256.0d0/(a**8)
      c(30)=75.0d0/128.0d0/(a**8)
      c(59)=15.0d0/512.0d0/(a**9)
      c(64)=75.0d0/256.0d0/(a**9)
      c(66)=25.0d0/128.0d0/(a**9)
      c(75)=75.0d0/256.0d0/(a**9)
      c(77)=15.0d0/512.0d0/(a**9)
      c(128)=15.0d0/1024.0d0/(a**10)
      c(130)=5.0d0/512.0d0/(a**10)
      c(139)=25.0d0/256.0d0/(a**10)
      c(141)=5.0d0/512.0d0/(a**10)
      c(154)=15.0d0/1024.0d0/(a**10)
      c(239)=5.0d0/1024.0d0/(a**11)
      c(241)=1.0d0/2048.0d0/(a**11)
      c(254)=5.0d0/1024.0d0/(a**11)
      c(398)=1.0d0/4096.0d0/(a**12)
    case (399)
      c(6)=225.0d0/128.0d0/(a**7)
      c(22)=75.0d0/128.0d0/(a**8)
      c(29)=675.0d0/256.0d0/(a**8)
      c(58)=15.0d0/512.0d0/(a**9)
      c(65)=225.0d0/256.0d0/(a**9)
      c(76)=225.0d0/512.0d0/(a**9)
      c(129)=45.0d0/1024.0d0/(a**10)
      c(140)=75.0d0/512.0d0/(a**10)
      c(155)=15.0d0/1024.0d0/(a**10)
      c(240)=15.0d0/2048.0d0/(a**11)
      c(255)=5.0d0/1024.0d0/(a**11)
      c(399)=1.0d0/4096.0d0/(a**12)
    case (400)
      c(7)=1575.0d0/128.0d0/(a**7)
      c(23)=525.0d0/128.0d0/(a**8)
      c(30)=1575.0d0/256.0d0/(a**8)
      c(59)=105.0d0/512.0d0/(a**9)
      c(66)=525.0d0/256.0d0/(a**9)
      c(77)=315.0d0/512.0d0/(a**9)
      c(130)=105.0d0/1024.0d0/(a**10)
      c(141)=105.0d0/512.0d0/(a**10)
      c(156)=15.0d0/1024.0d0/(a**10)
      c(241)=21.0d0/2048.0d0/(a**11)
      c(256)=5.0d0/1024.0d0/(a**11)
      c(400)=1.0d0/4096.0d0/(a**12)
    case (401)
      c(1)=315.0d0/64.0d0/(a**6)
      c(5)=315.0d0/64.0d0/(a**7)
      c(8)=315.0d0/32.0d0/(a**7)
      c(21)=105.0d0/256.0d0/(a**8)
      c(24)=315.0d0/32.0d0/(a**8)
      c(31)=315.0d0/128.0d0/(a**8)
      c(60)=105.0d0/128.0d0/(a**9)
      c(67)=315.0d0/128.0d0/(a**9)
      c(78)=21.0d0/128.0d0/(a**9)
      c(131)=105.0d0/512.0d0/(a**10)
      c(142)=21.0d0/128.0d0/(a**10)
      c(157)=3.0d0/1024.0d0/(a**10)
      c(242)=7.0d0/512.0d0/(a**11)
      c(257)=3.0d0/1024.0d0/(a**11)
      c(401)=1.0d0/4096.0d0/(a**12)
    case (402)
      c(9)=315.0d0/128.0d0/(a**7)
      c(25)=315.0d0/128.0d0/(a**8)
      c(32)=315.0d0/256.0d0/(a**8)
      c(61)=105.0d0/512.0d0/(a**9)
      c(68)=315.0d0/256.0d0/(a**9)
      c(79)=63.0d0/512.0d0/(a**9)
      c(132)=105.0d0/1024.0d0/(a**10)
      c(143)=63.0d0/512.0d0/(a**10)
      c(158)=3.0d0/1024.0d0/(a**10)
      c(243)=21.0d0/2048.0d0/(a**11)
      c(258)=3.0d0/1024.0d0/(a**11)
      c(402)=1.0d0/4096.0d0/(a**12)
    case (403)
      c(1)=45.0d0/64.0d0/(a**6)
      c(5)=45.0d0/64.0d0/(a**7)
      c(8)=135.0d0/128.0d0/(a**7)
      c(10)=45.0d0/128.0d0/(a**7)
      c(21)=15.0d0/256.0d0/(a**8)
      c(24)=135.0d0/128.0d0/(a**8)
      c(26)=45.0d0/128.0d0/(a**8)
      c(31)=45.0d0/256.0d0/(a**8)
      c(33)=135.0d0/256.0d0/(a**8)
      c(60)=45.0d0/512.0d0/(a**9)
      c(62)=15.0d0/512.0d0/(a**9)
      c(67)=45.0d0/256.0d0/(a**9)
      c(69)=135.0d0/256.0d0/(a**9)
      c(78)=3.0d0/512.0d0/(a**9)
      c(80)=45.0d0/512.0d0/(a**9)
      c(131)=15.0d0/1024.0d0/(a**10)
      c(133)=45.0d0/1024.0d0/(a**10)
      c(142)=3.0d0/512.0d0/(a**10)
      c(144)=45.0d0/512.0d0/(a**10)
      c(159)=3.0d0/1024.0d0/(a**10)
      c(242)=1.0d0/2048.0d0/(a**11)
      c(244)=15.0d0/2048.0d0/(a**11)
      c(259)=3.0d0/1024.0d0/(a**11)
      c(403)=1.0d0/4096.0d0/(a**12)
    case (404)
      c(9)=135.0d0/128.0d0/(a**7)
      c(25)=135.0d0/128.0d0/(a**8)
      c(32)=45.0d0/128.0d0/(a**8)
      c(34)=45.0d0/256.0d0/(a**8)
      c(61)=45.0d0/512.0d0/(a**9)
      c(68)=45.0d0/128.0d0/(a**9)
      c(70)=45.0d0/256.0d0/(a**9)
      c(79)=9.0d0/512.0d0/(a**9)
      c(81)=15.0d0/256.0d0/(a**9)
      c(132)=15.0d0/512.0d0/(a**10)
      c(134)=15.0d0/1024.0d0/(a**10)
      c(143)=9.0d0/512.0d0/(a**10)
      c(145)=15.0d0/256.0d0/(a**10)
      c(160)=3.0d0/1024.0d0/(a**10)
      c(243)=3.0d0/2048.0d0/(a**11)
      c(245)=5.0d0/1024.0d0/(a**11)
      c(260)=3.0d0/1024.0d0/(a**11)
      c(404)=1.0d0/4096.0d0/(a**12)
    case (405)
      c(1)=27.0d0/64.0d0/(a**6)
      c(5)=27.0d0/64.0d0/(a**7)
      c(8)=27.0d0/64.0d0/(a**7)
      c(10)=27.0d0/64.0d0/(a**7)
      c(21)=9.0d0/256.0d0/(a**8)
      c(24)=27.0d0/64.0d0/(a**8)
      c(26)=27.0d0/64.0d0/(a**8)
      c(31)=9.0d0/256.0d0/(a**8)
      c(33)=27.0d0/64.0d0/(a**8)
      c(35)=9.0d0/256.0d0/(a**8)
      c(60)=9.0d0/256.0d0/(a**9)
      c(62)=9.0d0/256.0d0/(a**9)
      c(67)=9.0d0/256.0d0/(a**9)
      c(69)=27.0d0/64.0d0/(a**9)
      c(71)=9.0d0/256.0d0/(a**9)
      c(80)=9.0d0/256.0d0/(a**9)
      c(82)=9.0d0/256.0d0/(a**9)
      c(131)=3.0d0/1024.0d0/(a**10)
      c(133)=9.0d0/256.0d0/(a**10)
      c(135)=3.0d0/1024.0d0/(a**10)
      c(144)=9.0d0/256.0d0/(a**10)
      c(146)=9.0d0/256.0d0/(a**10)
      c(161)=3.0d0/1024.0d0/(a**10)
      c(244)=3.0d0/1024.0d0/(a**11)
      c(246)=3.0d0/1024.0d0/(a**11)
      c(261)=3.0d0/1024.0d0/(a**11)
      c(405)=1.0d0/4096.0d0/(a**12)
    case (406)
      c(9)=135.0d0/128.0d0/(a**7)
      c(25)=135.0d0/128.0d0/(a**8)
      c(32)=45.0d0/256.0d0/(a**8)
      c(34)=45.0d0/128.0d0/(a**8)
      c(61)=45.0d0/512.0d0/(a**9)
      c(68)=45.0d0/256.0d0/(a**9)
      c(70)=45.0d0/128.0d0/(a**9)
      c(81)=15.0d0/256.0d0/(a**9)
      c(83)=9.0d0/512.0d0/(a**9)
      c(132)=15.0d0/1024.0d0/(a**10)
      c(134)=15.0d0/512.0d0/(a**10)
      c(145)=15.0d0/256.0d0/(a**10)
      c(147)=9.0d0/512.0d0/(a**10)
      c(162)=3.0d0/1024.0d0/(a**10)
      c(245)=5.0d0/1024.0d0/(a**11)
      c(247)=3.0d0/2048.0d0/(a**11)
      c(262)=3.0d0/1024.0d0/(a**11)
      c(406)=1.0d0/4096.0d0/(a**12)
    case (407)
      c(1)=45.0d0/64.0d0/(a**6)
      c(5)=45.0d0/64.0d0/(a**7)
      c(8)=45.0d0/128.0d0/(a**7)
      c(10)=135.0d0/128.0d0/(a**7)
      c(21)=15.0d0/256.0d0/(a**8)
      c(24)=45.0d0/128.0d0/(a**8)
      c(26)=135.0d0/128.0d0/(a**8)
      c(33)=135.0d0/256.0d0/(a**8)
      c(35)=45.0d0/256.0d0/(a**8)
      c(60)=15.0d0/512.0d0/(a**9)
      c(62)=45.0d0/512.0d0/(a**9)
      c(69)=135.0d0/256.0d0/(a**9)
      c(71)=45.0d0/256.0d0/(a**9)
      c(82)=45.0d0/512.0d0/(a**9)
      c(84)=3.0d0/512.0d0/(a**9)
      c(133)=45.0d0/1024.0d0/(a**10)
      c(135)=15.0d0/1024.0d0/(a**10)
      c(146)=45.0d0/512.0d0/(a**10)
      c(148)=3.0d0/512.0d0/(a**10)
      c(163)=3.0d0/1024.0d0/(a**10)
      c(246)=15.0d0/2048.0d0/(a**11)
      c(248)=1.0d0/2048.0d0/(a**11)
      c(263)=3.0d0/1024.0d0/(a**11)
      c(407)=1.0d0/4096.0d0/(a**12)
    case (408)
      c(9)=315.0d0/128.0d0/(a**7)
      c(25)=315.0d0/128.0d0/(a**8)
      c(34)=315.0d0/256.0d0/(a**8)
      c(61)=105.0d0/512.0d0/(a**9)
      c(70)=315.0d0/256.0d0/(a**9)
      c(83)=63.0d0/512.0d0/(a**9)
      c(134)=105.0d0/1024.0d0/(a**10)
      c(147)=63.0d0/512.0d0/(a**10)
      c(164)=3.0d0/1024.0d0/(a**10)
      c(247)=21.0d0/2048.0d0/(a**11)
      c(264)=3.0d0/1024.0d0/(a**11)
      c(408)=1.0d0/4096.0d0/(a**12)
    case (409)
      c(1)=315.0d0/64.0d0/(a**6)
      c(5)=315.0d0/64.0d0/(a**7)
      c(10)=315.0d0/32.0d0/(a**7)
      c(21)=105.0d0/256.0d0/(a**8)
      c(26)=315.0d0/32.0d0/(a**8)
      c(35)=315.0d0/128.0d0/(a**8)
      c(62)=105.0d0/128.0d0/(a**9)
      c(71)=315.0d0/128.0d0/(a**9)
      c(84)=21.0d0/128.0d0/(a**9)
      c(135)=105.0d0/512.0d0/(a**10)
      c(148)=21.0d0/128.0d0/(a**10)
      c(165)=3.0d0/1024.0d0/(a**10)
      c(248)=7.0d0/512.0d0/(a**11)
      c(265)=3.0d0/1024.0d0/(a**11)
      c(409)=1.0d0/4096.0d0/(a**12)
    case (410)
      c(6)=2835.0d0/128.0d0/(a**7)
      c(22)=945.0d0/256.0d0/(a**8)
      c(27)=945.0d0/64.0d0/(a**8)
      c(63)=315.0d0/128.0d0/(a**9)
      c(72)=567.0d0/256.0d0/(a**9)
      c(136)=189.0d0/512.0d0/(a**10)
      c(149)=27.0d0/256.0d0/(a**10)
      c(249)=9.0d0/512.0d0/(a**11)
      c(266)=3.0d0/2048.0d0/(a**11)
      c(410)=1.0d0/4096.0d0/(a**12)
    case (411)
      c(7)=315.0d0/128.0d0/(a**7)
      c(23)=105.0d0/256.0d0/(a**8)
      c(28)=315.0d0/64.0d0/(a**8)
      c(64)=105.0d0/128.0d0/(a**9)
      c(73)=315.0d0/256.0d0/(a**9)
      c(137)=105.0d0/512.0d0/(a**10)
      c(150)=21.0d0/256.0d0/(a**10)
      c(250)=7.0d0/512.0d0/(a**11)
      c(267)=3.0d0/2048.0d0/(a**11)
      c(411)=1.0d0/4096.0d0/(a**12)
    case (412)
      c(6)=315.0d0/128.0d0/(a**7)
      c(22)=105.0d0/256.0d0/(a**8)
      c(27)=315.0d0/256.0d0/(a**8)
      c(29)=315.0d0/256.0d0/(a**8)
      c(63)=105.0d0/512.0d0/(a**9)
      c(65)=105.0d0/512.0d0/(a**9)
      c(72)=63.0d0/512.0d0/(a**9)
      c(74)=315.0d0/512.0d0/(a**9)
      c(136)=21.0d0/1024.0d0/(a**10)
      c(138)=105.0d0/1024.0d0/(a**10)
      c(149)=3.0d0/1024.0d0/(a**10)
      c(151)=63.0d0/1024.0d0/(a**10)
      c(249)=1.0d0/2048.0d0/(a**11)
      c(251)=21.0d0/2048.0d0/(a**11)
      c(268)=3.0d0/2048.0d0/(a**11)
      c(412)=1.0d0/4096.0d0/(a**12)
    case (413)
      c(7)=135.0d0/128.0d0/(a**7)
      c(23)=45.0d0/256.0d0/(a**8)
      c(28)=405.0d0/256.0d0/(a**8)
      c(30)=45.0d0/256.0d0/(a**8)
      c(64)=135.0d0/512.0d0/(a**9)
      c(66)=15.0d0/512.0d0/(a**9)
      c(73)=135.0d0/512.0d0/(a**9)
      c(75)=135.0d0/512.0d0/(a**9)
      c(137)=45.0d0/1024.0d0/(a**10)
      c(139)=45.0d0/1024.0d0/(a**10)
      c(150)=9.0d0/1024.0d0/(a**10)
      c(152)=45.0d0/1024.0d0/(a**10)
      c(250)=3.0d0/2048.0d0/(a**11)
      c(252)=15.0d0/2048.0d0/(a**11)
      c(269)=3.0d0/2048.0d0/(a**11)
      c(413)=1.0d0/4096.0d0/(a**12)
    case (414)
      c(6)=135.0d0/128.0d0/(a**7)
      c(22)=45.0d0/256.0d0/(a**8)
      c(27)=45.0d0/128.0d0/(a**8)
      c(29)=135.0d0/128.0d0/(a**8)
      c(63)=15.0d0/256.0d0/(a**9)
      c(65)=45.0d0/256.0d0/(a**9)
      c(72)=9.0d0/512.0d0/(a**9)
      c(74)=45.0d0/128.0d0/(a**9)
      c(76)=45.0d0/512.0d0/(a**9)
      c(136)=3.0d0/1024.0d0/(a**10)
      c(138)=15.0d0/256.0d0/(a**10)
      c(140)=15.0d0/1024.0d0/(a**10)
      c(151)=9.0d0/512.0d0/(a**10)
      c(153)=15.0d0/512.0d0/(a**10)
      c(251)=3.0d0/1024.0d0/(a**11)
      c(253)=5.0d0/1024.0d0/(a**11)
      c(270)=3.0d0/2048.0d0/(a**11)
      c(414)=1.0d0/4096.0d0/(a**12)
    case (415)
      c(7)=135.0d0/128.0d0/(a**7)
      c(23)=45.0d0/256.0d0/(a**8)
      c(28)=135.0d0/128.0d0/(a**8)
      c(30)=45.0d0/128.0d0/(a**8)
      c(64)=45.0d0/256.0d0/(a**9)
      c(66)=15.0d0/256.0d0/(a**9)
      c(73)=45.0d0/512.0d0/(a**9)
      c(75)=45.0d0/128.0d0/(a**9)
      c(77)=9.0d0/512.0d0/(a**9)
      c(137)=15.0d0/1024.0d0/(a**10)
      c(139)=15.0d0/256.0d0/(a**10)
      c(141)=3.0d0/1024.0d0/(a**10)
      c(152)=15.0d0/512.0d0/(a**10)
      c(154)=9.0d0/512.0d0/(a**10)
      c(252)=5.0d0/1024.0d0/(a**11)
      c(254)=3.0d0/1024.0d0/(a**11)
      c(271)=3.0d0/2048.0d0/(a**11)
      c(415)=1.0d0/4096.0d0/(a**12)
    case (416)
      c(6)=135.0d0/128.0d0/(a**7)
      c(22)=45.0d0/256.0d0/(a**8)
      c(27)=45.0d0/256.0d0/(a**8)
      c(29)=405.0d0/256.0d0/(a**8)
      c(63)=15.0d0/512.0d0/(a**9)
      c(65)=135.0d0/512.0d0/(a**9)
      c(74)=135.0d0/512.0d0/(a**9)
      c(76)=135.0d0/512.0d0/(a**9)
      c(138)=45.0d0/1024.0d0/(a**10)
      c(140)=45.0d0/1024.0d0/(a**10)
      c(153)=45.0d0/1024.0d0/(a**10)
      c(155)=9.0d0/1024.0d0/(a**10)
      c(253)=15.0d0/2048.0d0/(a**11)
      c(255)=3.0d0/2048.0d0/(a**11)
      c(272)=3.0d0/2048.0d0/(a**11)
      c(416)=1.0d0/4096.0d0/(a**12)
    case (417)
      c(7)=315.0d0/128.0d0/(a**7)
      c(23)=105.0d0/256.0d0/(a**8)
      c(28)=315.0d0/256.0d0/(a**8)
      c(30)=315.0d0/256.0d0/(a**8)
      c(64)=105.0d0/512.0d0/(a**9)
      c(66)=105.0d0/512.0d0/(a**9)
      c(75)=315.0d0/512.0d0/(a**9)
      c(77)=63.0d0/512.0d0/(a**9)
      c(139)=105.0d0/1024.0d0/(a**10)
      c(141)=21.0d0/1024.0d0/(a**10)
      c(154)=63.0d0/1024.0d0/(a**10)
      c(156)=3.0d0/1024.0d0/(a**10)
      c(254)=21.0d0/2048.0d0/(a**11)
      c(256)=1.0d0/2048.0d0/(a**11)
      c(273)=3.0d0/2048.0d0/(a**11)
      c(417)=1.0d0/4096.0d0/(a**12)
    case (418)
      c(6)=315.0d0/128.0d0/(a**7)
      c(22)=105.0d0/256.0d0/(a**8)
      c(29)=315.0d0/64.0d0/(a**8)
      c(65)=105.0d0/128.0d0/(a**9)
      c(76)=315.0d0/256.0d0/(a**9)
      c(140)=105.0d0/512.0d0/(a**10)
      c(155)=21.0d0/256.0d0/(a**10)
      c(255)=7.0d0/512.0d0/(a**11)
      c(274)=3.0d0/2048.0d0/(a**11)
      c(418)=1.0d0/4096.0d0/(a**12)
    case (419)
      c(7)=2835.0d0/128.0d0/(a**7)
      c(23)=945.0d0/256.0d0/(a**8)
      c(30)=945.0d0/64.0d0/(a**8)
      c(66)=315.0d0/128.0d0/(a**9)
      c(77)=567.0d0/256.0d0/(a**9)
      c(141)=189.0d0/512.0d0/(a**10)
      c(156)=27.0d0/256.0d0/(a**10)
      c(256)=9.0d0/512.0d0/(a**11)
      c(275)=3.0d0/2048.0d0/(a**11)
      c(419)=1.0d0/4096.0d0/(a**12)
    case (420)
      c(1)=945.0d0/64.0d0/(a**6)
      c(5)=945.0d0/128.0d0/(a**7)
      c(8)=4725.0d0/128.0d0/(a**7)
      c(24)=4725.0d0/256.0d0/(a**8)
      c(31)=1575.0d0/128.0d0/(a**8)
      c(67)=1575.0d0/256.0d0/(a**9)
      c(78)=315.0d0/256.0d0/(a**9)
      c(142)=315.0d0/512.0d0/(a**10)
      c(157)=45.0d0/1024.0d0/(a**10)
      c(257)=45.0d0/2048.0d0/(a**11)
      c(276)=1.0d0/2048.0d0/(a**11)
      c(420)=1.0d0/4096.0d0/(a**12)
    case (421)
      c(9)=945.0d0/128.0d0/(a**7)
      c(25)=945.0d0/256.0d0/(a**8)
      c(32)=315.0d0/64.0d0/(a**8)
      c(68)=315.0d0/128.0d0/(a**9)
      c(79)=189.0d0/256.0d0/(a**9)
      c(143)=189.0d0/512.0d0/(a**10)
      c(158)=9.0d0/256.0d0/(a**10)
      c(258)=9.0d0/512.0d0/(a**11)
      c(277)=1.0d0/2048.0d0/(a**11)
      c(421)=1.0d0/4096.0d0/(a**12)
    case (422)
      c(1)=105.0d0/64.0d0/(a**6)
      c(5)=105.0d0/128.0d0/(a**7)
      c(8)=105.0d0/32.0d0/(a**7)
      c(10)=105.0d0/128.0d0/(a**7)
      c(24)=105.0d0/64.0d0/(a**8)
      c(26)=105.0d0/256.0d0/(a**8)
      c(31)=105.0d0/128.0d0/(a**8)
      c(33)=105.0d0/64.0d0/(a**8)
      c(67)=105.0d0/256.0d0/(a**9)
      c(69)=105.0d0/128.0d0/(a**9)
      c(78)=7.0d0/128.0d0/(a**9)
      c(80)=105.0d0/256.0d0/(a**9)
      c(142)=7.0d0/256.0d0/(a**10)
      c(144)=105.0d0/512.0d0/(a**10)
      c(157)=1.0d0/1024.0d0/(a**10)
      c(159)=7.0d0/256.0d0/(a**10)
      c(257)=1.0d0/2048.0d0/(a**11)
      c(259)=7.0d0/512.0d0/(a**11)
      c(278)=1.0d0/2048.0d0/(a**11)
      c(422)=1.0d0/4096.0d0/(a**12)
    case (423)
      c(9)=315.0d0/128.0d0/(a**7)
      c(25)=315.0d0/256.0d0/(a**8)
      c(32)=315.0d0/256.0d0/(a**8)
      c(34)=105.0d0/256.0d0/(a**8)
      c(68)=315.0d0/512.0d0/(a**9)
      c(70)=105.0d0/512.0d0/(a**9)
      c(79)=63.0d0/512.0d0/(a**9)
      c(81)=105.0d0/512.0d0/(a**9)
      c(143)=63.0d0/1024.0d0/(a**10)
      c(145)=105.0d0/1024.0d0/(a**10)
      c(158)=3.0d0/1024.0d0/(a**10)
      c(160)=21.0d0/1024.0d0/(a**10)
      c(258)=3.0d0/2048.0d0/(a**11)
      c(260)=21.0d0/2048.0d0/(a**11)
      c(279)=1.0d0/2048.0d0/(a**11)
      c(423)=1.0d0/4096.0d0/(a**12)
    case (424)
      c(1)=45.0d0/64.0d0/(a**6)
      c(5)=45.0d0/128.0d0/(a**7)
      c(8)=135.0d0/128.0d0/(a**7)
      c(10)=45.0d0/64.0d0/(a**7)
      c(24)=135.0d0/256.0d0/(a**8)
      c(26)=45.0d0/128.0d0/(a**8)
      c(31)=45.0d0/256.0d0/(a**8)
      c(33)=135.0d0/128.0d0/(a**8)
      c(35)=15.0d0/256.0d0/(a**8)
      c(67)=45.0d0/512.0d0/(a**9)
      c(69)=135.0d0/256.0d0/(a**9)
      c(71)=15.0d0/512.0d0/(a**9)
      c(78)=3.0d0/512.0d0/(a**9)
      c(80)=45.0d0/256.0d0/(a**9)
      c(82)=45.0d0/512.0d0/(a**9)
      c(142)=3.0d0/1024.0d0/(a**10)
      c(144)=45.0d0/512.0d0/(a**10)
      c(146)=45.0d0/1024.0d0/(a**10)
      c(159)=3.0d0/512.0d0/(a**10)
      c(161)=15.0d0/1024.0d0/(a**10)
      c(259)=3.0d0/1024.0d0/(a**11)
      c(261)=15.0d0/2048.0d0/(a**11)
      c(280)=1.0d0/2048.0d0/(a**11)
      c(424)=1.0d0/4096.0d0/(a**12)
    case (425)
      c(9)=225.0d0/128.0d0/(a**7)
      c(25)=225.0d0/256.0d0/(a**8)
      c(32)=75.0d0/128.0d0/(a**8)
      c(34)=75.0d0/128.0d0/(a**8)
      c(68)=75.0d0/256.0d0/(a**9)
      c(70)=75.0d0/256.0d0/(a**9)
      c(79)=15.0d0/512.0d0/(a**9)
      c(81)=25.0d0/128.0d0/(a**9)
      c(83)=15.0d0/512.0d0/(a**9)
      c(143)=15.0d0/1024.0d0/(a**10)
      c(145)=25.0d0/256.0d0/(a**10)
      c(147)=15.0d0/1024.0d0/(a**10)
      c(160)=5.0d0/512.0d0/(a**10)
      c(162)=5.0d0/512.0d0/(a**10)
      c(260)=5.0d0/1024.0d0/(a**11)
      c(262)=5.0d0/1024.0d0/(a**11)
      c(281)=1.0d0/2048.0d0/(a**11)
      c(425)=1.0d0/4096.0d0/(a**12)
    case (426)
      c(1)=45.0d0/64.0d0/(a**6)
      c(5)=45.0d0/128.0d0/(a**7)
      c(8)=45.0d0/64.0d0/(a**7)
      c(10)=135.0d0/128.0d0/(a**7)
      c(24)=45.0d0/128.0d0/(a**8)
      c(26)=135.0d0/256.0d0/(a**8)
      c(31)=15.0d0/256.0d0/(a**8)
      c(33)=135.0d0/128.0d0/(a**8)
      c(35)=45.0d0/256.0d0/(a**8)
      c(67)=15.0d0/512.0d0/(a**9)
      c(69)=135.0d0/256.0d0/(a**9)
      c(71)=45.0d0/512.0d0/(a**9)
      c(80)=45.0d0/512.0d0/(a**9)
      c(82)=45.0d0/256.0d0/(a**9)
      c(84)=3.0d0/512.0d0/(a**9)
      c(144)=45.0d0/1024.0d0/(a**10)
      c(146)=45.0d0/512.0d0/(a**10)
      c(148)=3.0d0/1024.0d0/(a**10)
      c(161)=15.0d0/1024.0d0/(a**10)
      c(163)=3.0d0/512.0d0/(a**10)
      c(261)=15.0d0/2048.0d0/(a**11)
      c(263)=3.0d0/1024.0d0/(a**11)
      c(282)=1.0d0/2048.0d0/(a**11)
      c(426)=1.0d0/4096.0d0/(a**12)
    case (427)
      c(9)=315.0d0/128.0d0/(a**7)
      c(25)=315.0d0/256.0d0/(a**8)
      c(32)=105.0d0/256.0d0/(a**8)
      c(34)=315.0d0/256.0d0/(a**8)
      c(68)=105.0d0/512.0d0/(a**9)
      c(70)=315.0d0/512.0d0/(a**9)
      c(81)=105.0d0/512.0d0/(a**9)
      c(83)=63.0d0/512.0d0/(a**9)
      c(145)=105.0d0/1024.0d0/(a**10)
      c(147)=63.0d0/1024.0d0/(a**10)
      c(162)=21.0d0/1024.0d0/(a**10)
      c(164)=3.0d0/1024.0d0/(a**10)
      c(262)=21.0d0/2048.0d0/(a**11)
      c(264)=3.0d0/2048.0d0/(a**11)
      c(283)=1.0d0/2048.0d0/(a**11)
      c(427)=1.0d0/4096.0d0/(a**12)
    case (428)
      c(1)=105.0d0/64.0d0/(a**6)
      c(5)=105.0d0/128.0d0/(a**7)
      c(8)=105.0d0/128.0d0/(a**7)
      c(10)=105.0d0/32.0d0/(a**7)
      c(24)=105.0d0/256.0d0/(a**8)
      c(26)=105.0d0/64.0d0/(a**8)
      c(33)=105.0d0/64.0d0/(a**8)
      c(35)=105.0d0/128.0d0/(a**8)
      c(69)=105.0d0/128.0d0/(a**9)
      c(71)=105.0d0/256.0d0/(a**9)
      c(82)=105.0d0/256.0d0/(a**9)
      c(84)=7.0d0/128.0d0/(a**9)
      c(146)=105.0d0/512.0d0/(a**10)
      c(148)=7.0d0/256.0d0/(a**10)
      c(163)=7.0d0/256.0d0/(a**10)
      c(165)=1.0d0/1024.0d0/(a**10)
      c(263)=7.0d0/512.0d0/(a**11)
      c(265)=1.0d0/2048.0d0/(a**11)
      c(284)=1.0d0/2048.0d0/(a**11)
      c(428)=1.0d0/4096.0d0/(a**12)
    case (429)
      c(9)=945.0d0/128.0d0/(a**7)
      c(25)=945.0d0/256.0d0/(a**8)
      c(34)=315.0d0/64.0d0/(a**8)
      c(70)=315.0d0/128.0d0/(a**9)
      c(83)=189.0d0/256.0d0/(a**9)
      c(147)=189.0d0/512.0d0/(a**10)
      c(164)=9.0d0/256.0d0/(a**10)
      c(264)=9.0d0/512.0d0/(a**11)
      c(285)=1.0d0/2048.0d0/(a**11)
      c(429)=1.0d0/4096.0d0/(a**12)
    case (430)
      c(1)=945.0d0/64.0d0/(a**6)
      c(5)=945.0d0/128.0d0/(a**7)
      c(10)=4725.0d0/128.0d0/(a**7)
      c(26)=4725.0d0/256.0d0/(a**8)
      c(35)=1575.0d0/128.0d0/(a**8)
      c(71)=1575.0d0/256.0d0/(a**9)
      c(84)=315.0d0/256.0d0/(a**9)
      c(148)=315.0d0/512.0d0/(a**10)
      c(165)=45.0d0/1024.0d0/(a**10)
      c(265)=45.0d0/2048.0d0/(a**11)
      c(286)=1.0d0/2048.0d0/(a**11)
      c(430)=1.0d0/4096.0d0/(a**12)
    case (431)
      c(6)=10395.0d0/128.0d0/(a**7)
      c(27)=17325.0d0/256.0d0/(a**8)
      c(72)=3465.0d0/256.0d0/(a**9)
      c(149)=495.0d0/512.0d0/(a**10)
      c(266)=55.0d0/2048.0d0/(a**11)
      c(431)=1.0d0/4096.0d0/(a**12)
    case (432)
      c(7)=945.0d0/128.0d0/(a**7)
      c(28)=4725.0d0/256.0d0/(a**8)
      c(73)=1575.0d0/256.0d0/(a**9)
      c(150)=315.0d0/512.0d0/(a**10)
      c(267)=45.0d0/2048.0d0/(a**11)
      c(432)=1.0d0/4096.0d0/(a**12)
    case (433)
      c(6)=945.0d0/128.0d0/(a**7)
      c(27)=315.0d0/64.0d0/(a**8)
      c(29)=945.0d0/256.0d0/(a**8)
      c(72)=189.0d0/256.0d0/(a**9)
      c(74)=315.0d0/128.0d0/(a**9)
      c(149)=9.0d0/256.0d0/(a**10)
      c(151)=189.0d0/512.0d0/(a**10)
      c(266)=1.0d0/2048.0d0/(a**11)
      c(268)=9.0d0/512.0d0/(a**11)
      c(433)=1.0d0/4096.0d0/(a**12)
    case (434)
      c(7)=315.0d0/128.0d0/(a**7)
      c(28)=315.0d0/64.0d0/(a**8)
      c(30)=105.0d0/256.0d0/(a**8)
      c(73)=315.0d0/256.0d0/(a**9)
      c(75)=105.0d0/128.0d0/(a**9)
      c(150)=21.0d0/256.0d0/(a**10)
      c(152)=105.0d0/512.0d0/(a**10)
      c(267)=3.0d0/2048.0d0/(a**11)
      c(269)=7.0d0/512.0d0/(a**11)
      c(434)=1.0d0/4096.0d0/(a**12)
    case (435)
      c(6)=315.0d0/128.0d0/(a**7)
      c(27)=315.0d0/256.0d0/(a**8)
      c(29)=315.0d0/128.0d0/(a**8)
      c(72)=63.0d0/512.0d0/(a**9)
      c(74)=315.0d0/256.0d0/(a**9)
      c(76)=105.0d0/512.0d0/(a**9)
      c(149)=3.0d0/1024.0d0/(a**10)
      c(151)=63.0d0/512.0d0/(a**10)
      c(153)=105.0d0/1024.0d0/(a**10)
      c(268)=3.0d0/1024.0d0/(a**11)
      c(270)=21.0d0/2048.0d0/(a**11)
      c(435)=1.0d0/4096.0d0/(a**12)
    case (436)
      c(7)=225.0d0/128.0d0/(a**7)
      c(28)=675.0d0/256.0d0/(a**8)
      c(30)=75.0d0/128.0d0/(a**8)
      c(73)=225.0d0/512.0d0/(a**9)
      c(75)=225.0d0/256.0d0/(a**9)
      c(77)=15.0d0/512.0d0/(a**9)
      c(150)=15.0d0/1024.0d0/(a**10)
      c(152)=75.0d0/512.0d0/(a**10)
      c(154)=45.0d0/1024.0d0/(a**10)
      c(269)=5.0d0/1024.0d0/(a**11)
      c(271)=15.0d0/2048.0d0/(a**11)
      c(436)=1.0d0/4096.0d0/(a**12)
    case (437)
      c(6)=225.0d0/128.0d0/(a**7)
      c(27)=75.0d0/128.0d0/(a**8)
      c(29)=675.0d0/256.0d0/(a**8)
      c(72)=15.0d0/512.0d0/(a**9)
      c(74)=225.0d0/256.0d0/(a**9)
      c(76)=225.0d0/512.0d0/(a**9)
      c(151)=45.0d0/1024.0d0/(a**10)
      c(153)=75.0d0/512.0d0/(a**10)
      c(155)=15.0d0/1024.0d0/(a**10)
      c(270)=15.0d0/2048.0d0/(a**11)
      c(272)=5.0d0/1024.0d0/(a**11)
      c(437)=1.0d0/4096.0d0/(a**12)
    case (438)
      c(7)=315.0d0/128.0d0/(a**7)
      c(28)=315.0d0/128.0d0/(a**8)
      c(30)=315.0d0/256.0d0/(a**8)
      c(73)=105.0d0/512.0d0/(a**9)
      c(75)=315.0d0/256.0d0/(a**9)
      c(77)=63.0d0/512.0d0/(a**9)
      c(152)=105.0d0/1024.0d0/(a**10)
      c(154)=63.0d0/512.0d0/(a**10)
      c(156)=3.0d0/1024.0d0/(a**10)
      c(271)=21.0d0/2048.0d0/(a**11)
      c(273)=3.0d0/1024.0d0/(a**11)
      c(438)=1.0d0/4096.0d0/(a**12)
    case (439)
      c(6)=315.0d0/128.0d0/(a**7)
      c(27)=105.0d0/256.0d0/(a**8)
      c(29)=315.0d0/64.0d0/(a**8)
      c(74)=105.0d0/128.0d0/(a**9)
      c(76)=315.0d0/256.0d0/(a**9)
      c(153)=105.0d0/512.0d0/(a**10)
      c(155)=21.0d0/256.0d0/(a**10)
      c(272)=7.0d0/512.0d0/(a**11)
      c(274)=3.0d0/2048.0d0/(a**11)
      c(439)=1.0d0/4096.0d0/(a**12)
    case (440)
      c(7)=945.0d0/128.0d0/(a**7)
      c(28)=945.0d0/256.0d0/(a**8)
      c(30)=315.0d0/64.0d0/(a**8)
      c(75)=315.0d0/128.0d0/(a**9)
      c(77)=189.0d0/256.0d0/(a**9)
      c(154)=189.0d0/512.0d0/(a**10)
      c(156)=9.0d0/256.0d0/(a**10)
      c(273)=9.0d0/512.0d0/(a**11)
      c(275)=1.0d0/2048.0d0/(a**11)
      c(440)=1.0d0/4096.0d0/(a**12)
    case (441)
      c(6)=945.0d0/128.0d0/(a**7)
      c(29)=4725.0d0/256.0d0/(a**8)
      c(76)=1575.0d0/256.0d0/(a**9)
      c(155)=315.0d0/512.0d0/(a**10)
      c(274)=45.0d0/2048.0d0/(a**11)
      c(441)=1.0d0/4096.0d0/(a**12)
    case (442)
      c(7)=10395.0d0/128.0d0/(a**7)
      c(30)=17325.0d0/256.0d0/(a**8)
      c(77)=3465.0d0/256.0d0/(a**9)
      c(156)=495.0d0/512.0d0/(a**10)
      c(275)=55.0d0/2048.0d0/(a**11)
      c(442)=1.0d0/4096.0d0/(a**12)
    case (443)
      c(1)=10395.0d0/64.0d0/(a**6)
      c(8)=31185.0d0/64.0d0/(a**7)
      c(31)=51975.0d0/256.0d0/(a**8)
      c(78)=3465.0d0/128.0d0/(a**9)
      c(157)=1485.0d0/1024.0d0/(a**10)
      c(276)=33.0d0/1024.0d0/(a**11)
      c(443)=1.0d0/4096.0d0/(a**12)
    case (444)
      c(9)=10395.0d0/128.0d0/(a**7)
      c(32)=17325.0d0/256.0d0/(a**8)
      c(79)=3465.0d0/256.0d0/(a**9)
      c(158)=495.0d0/512.0d0/(a**10)
      c(277)=55.0d0/2048.0d0/(a**11)
      c(444)=1.0d0/4096.0d0/(a**12)
    case (445)
      c(1)=945.0d0/64.0d0/(a**6)
      c(8)=4725.0d0/128.0d0/(a**7)
      c(10)=945.0d0/128.0d0/(a**7)
      c(31)=1575.0d0/128.0d0/(a**8)
      c(33)=4725.0d0/256.0d0/(a**8)
      c(78)=315.0d0/256.0d0/(a**9)
      c(80)=1575.0d0/256.0d0/(a**9)
      c(157)=45.0d0/1024.0d0/(a**10)
      c(159)=315.0d0/512.0d0/(a**10)
      c(276)=1.0d0/2048.0d0/(a**11)
      c(278)=45.0d0/2048.0d0/(a**11)
      c(445)=1.0d0/4096.0d0/(a**12)
    case (446)
      c(9)=2835.0d0/128.0d0/(a**7)
      c(32)=945.0d0/64.0d0/(a**8)
      c(34)=945.0d0/256.0d0/(a**8)
      c(79)=567.0d0/256.0d0/(a**9)
      c(81)=315.0d0/128.0d0/(a**9)
      c(158)=27.0d0/256.0d0/(a**10)
      c(160)=189.0d0/512.0d0/(a**10)
      c(277)=3.0d0/2048.0d0/(a**11)
      c(279)=9.0d0/512.0d0/(a**11)
      c(446)=1.0d0/4096.0d0/(a**12)
    case (447)
      c(1)=315.0d0/64.0d0/(a**6)
      c(8)=315.0d0/32.0d0/(a**7)
      c(10)=315.0d0/64.0d0/(a**7)
      c(31)=315.0d0/128.0d0/(a**8)
      c(33)=315.0d0/32.0d0/(a**8)
      c(35)=105.0d0/256.0d0/(a**8)
      c(78)=21.0d0/128.0d0/(a**9)
      c(80)=315.0d0/128.0d0/(a**9)
      c(82)=105.0d0/128.0d0/(a**9)
      c(157)=3.0d0/1024.0d0/(a**10)
      c(159)=21.0d0/128.0d0/(a**10)
      c(161)=105.0d0/512.0d0/(a**10)
      c(278)=3.0d0/1024.0d0/(a**11)
      c(280)=7.0d0/512.0d0/(a**11)
      c(447)=1.0d0/4096.0d0/(a**12)
    case (448)
      c(9)=1575.0d0/128.0d0/(a**7)
      c(32)=1575.0d0/256.0d0/(a**8)
      c(34)=525.0d0/128.0d0/(a**8)
      c(79)=315.0d0/512.0d0/(a**9)
      c(81)=525.0d0/256.0d0/(a**9)
      c(83)=105.0d0/512.0d0/(a**9)
      c(158)=15.0d0/1024.0d0/(a**10)
      c(160)=105.0d0/512.0d0/(a**10)
      c(162)=105.0d0/1024.0d0/(a**10)
      c(279)=5.0d0/1024.0d0/(a**11)
      c(281)=21.0d0/2048.0d0/(a**11)
      c(448)=1.0d0/4096.0d0/(a**12)
    case (449)
      c(1)=225.0d0/64.0d0/(a**6)
      c(8)=675.0d0/128.0d0/(a**7)
      c(10)=675.0d0/128.0d0/(a**7)
      c(31)=225.0d0/256.0d0/(a**8)
      c(33)=2025.0d0/256.0d0/(a**8)
      c(35)=225.0d0/256.0d0/(a**8)
      c(78)=15.0d0/512.0d0/(a**9)
      c(80)=675.0d0/512.0d0/(a**9)
      c(82)=675.0d0/512.0d0/(a**9)
      c(84)=15.0d0/512.0d0/(a**9)
      c(159)=45.0d0/1024.0d0/(a**10)
      c(161)=225.0d0/1024.0d0/(a**10)
      c(163)=45.0d0/1024.0d0/(a**10)
      c(280)=15.0d0/2048.0d0/(a**11)
      c(282)=15.0d0/2048.0d0/(a**11)
      c(449)=1.0d0/4096.0d0/(a**12)
    case (450)
      c(9)=1575.0d0/128.0d0/(a**7)
      c(32)=525.0d0/128.0d0/(a**8)
      c(34)=1575.0d0/256.0d0/(a**8)
      c(79)=105.0d0/512.0d0/(a**9)
      c(81)=525.0d0/256.0d0/(a**9)
      c(83)=315.0d0/512.0d0/(a**9)
      c(160)=105.0d0/1024.0d0/(a**10)
      c(162)=105.0d0/512.0d0/(a**10)
      c(164)=15.0d0/1024.0d0/(a**10)
      c(281)=21.0d0/2048.0d0/(a**11)
      c(283)=5.0d0/1024.0d0/(a**11)
      c(450)=1.0d0/4096.0d0/(a**12)
    case (451)
      c(1)=315.0d0/64.0d0/(a**6)
      c(8)=315.0d0/64.0d0/(a**7)
      c(10)=315.0d0/32.0d0/(a**7)
      c(31)=105.0d0/256.0d0/(a**8)
      c(33)=315.0d0/32.0d0/(a**8)
      c(35)=315.0d0/128.0d0/(a**8)
      c(80)=105.0d0/128.0d0/(a**9)
      c(82)=315.0d0/128.0d0/(a**9)
      c(84)=21.0d0/128.0d0/(a**9)
      c(161)=105.0d0/512.0d0/(a**10)
      c(163)=21.0d0/128.0d0/(a**10)
      c(165)=3.0d0/1024.0d0/(a**10)
      c(282)=7.0d0/512.0d0/(a**11)
      c(284)=3.0d0/1024.0d0/(a**11)
      c(451)=1.0d0/4096.0d0/(a**12)
    case (452)
      c(9)=2835.0d0/128.0d0/(a**7)
      c(32)=945.0d0/256.0d0/(a**8)
      c(34)=945.0d0/64.0d0/(a**8)
      c(81)=315.0d0/128.0d0/(a**9)
      c(83)=567.0d0/256.0d0/(a**9)
      c(162)=189.0d0/512.0d0/(a**10)
      c(164)=27.0d0/256.0d0/(a**10)
      c(283)=9.0d0/512.0d0/(a**11)
      c(285)=3.0d0/2048.0d0/(a**11)
      c(452)=1.0d0/4096.0d0/(a**12)
    case (453)
      c(1)=945.0d0/64.0d0/(a**6)
      c(8)=945.0d0/128.0d0/(a**7)
      c(10)=4725.0d0/128.0d0/(a**7)
      c(33)=4725.0d0/256.0d0/(a**8)
      c(35)=1575.0d0/128.0d0/(a**8)
      c(82)=1575.0d0/256.0d0/(a**9)
      c(84)=315.0d0/256.0d0/(a**9)
      c(163)=315.0d0/512.0d0/(a**10)
      c(165)=45.0d0/1024.0d0/(a**10)
      c(284)=45.0d0/2048.0d0/(a**11)
      c(286)=1.0d0/2048.0d0/(a**11)
      c(453)=1.0d0/4096.0d0/(a**12)
    case (454)
      c(9)=10395.0d0/128.0d0/(a**7)
      c(34)=17325.0d0/256.0d0/(a**8)
      c(83)=3465.0d0/256.0d0/(a**9)
      c(164)=495.0d0/512.0d0/(a**10)
      c(285)=55.0d0/2048.0d0/(a**11)
      c(454)=1.0d0/4096.0d0/(a**12)
    case (455)
      c(1)=10395.0d0/64.0d0/(a**6)
      c(10)=31185.0d0/64.0d0/(a**7)
      c(35)=51975.0d0/256.0d0/(a**8)
      c(84)=3465.0d0/128.0d0/(a**9)
      c(165)=1485.0d0/1024.0d0/(a**10)
      c(286)=33.0d0/1024.0d0/(a**11)
      c(455)=1.0d0/4096.0d0/(a**12)
    case default
      print*,'Error: R_to_D not implemented for ir=',ir
      stop
  end select
 
end subroutine

end module

